searchData={"items":[{"type":"module","title":"Reactor","doc":"Reactor is a dynamic, concurrent, dependency resolving saga orchestrator.","ref":"Reactor.html"},{"type":"module","title":"Usage - Reactor","doc":"You can construct a reactor using the `Reactor` Spark DSL:\n\n```elixir\ndefmodule HelloWorldReactor do\n  @moduledoc false\n  use Reactor\n\n  input :whom\n\n  step :greet, Greeter do\n    argument :whom, input(:whom)\n  end\n\n  return :greet\nend\n```\n\n    iex> Reactor.run(HelloWorldReactor, %{whom: \"Dear Reader\"})\n    {:ok, \"Hello, Dear Reader!\"}\n\nor you can build it programmatically:\n\n    iex> reactor = Builder.new()\n    ...> {:ok, reactor} = Builder.add_input(reactor, :whom)\n    ...> {:ok, reactor} = Builder.add_step(reactor, :greet, Greeter, whom: {:input, :whom})\n    ...> {:ok, reactor} = Builder.return(reactor, :greet)\n    ...> Reactor.run(reactor, %{whom: nil})\n    {:ok, \"Hello, World!\"}","ref":"Reactor.html#module-usage"},{"type":"module","title":"Options - Reactor","doc":"* `:extensions` (list of module that adopts `Spark.Dsl.Extension`) - A list of DSL extensions to add to the `Spark.Dsl`\n\n* `:otp_app` (`t:atom/0`) - The otp_app to use for any application configurable options\n\n* `:fragments` (list of `t:module/0`) - Fragments to include in the `Spark.Dsl`. See the fragments guide for more.","ref":"Reactor.html#module-options"},{"type":"type","title":"Reactor.async_option/0","doc":"When set to `false` forces the Reactor to run every step synchronously,\nregardless of the step configuration.\n\nDefaults to `true`.","ref":"Reactor.html#t:async_option/0"},{"type":"type","title":"Reactor.concurrency_key_option/0","doc":"Use a `Reactor.Executor.ConcurrencyTracker.pool_key` to allow this Reactor to\nshare it's concurrency pool with other Reactor instances.\n\nIf you do not specify one then the Reactor will initialise a new pool and\nplace it in it's context for any child Reactors to re-use.\n\nOnly used if `async?` is set to `true`.","ref":"Reactor.html#t:concurrency_key_option/0"},{"type":"type","title":"Reactor.context/0","doc":"","ref":"Reactor.html#t:context/0"},{"type":"type","title":"Reactor.context_arg/0","doc":"","ref":"Reactor.html#t:context_arg/0"},{"type":"type","title":"Reactor.halt_timeout_option/0","doc":"How long to wait for asynchronous steps to complete when halting.\n\nDefaults to 5000ms.","ref":"Reactor.html#t:halt_timeout_option/0"},{"type":"type","title":"Reactor.inputs/0","doc":"","ref":"Reactor.html#t:inputs/0"},{"type":"macro","title":"Reactor.is_reactor/1","doc":"A guard which returns true if the value is a Reactor struct","ref":"Reactor.html#is_reactor/1"},{"type":"type","title":"Reactor.max_concurrency_option/0","doc":"Specify the maximum number of asynchronous steps which can be run in parallel.\n\nDefaults to the result of `System.schedulers_online/0`.  Only used if\n`async?` is set to `true`.","ref":"Reactor.html#t:max_concurrency_option/0"},{"type":"type","title":"Reactor.max_iterations_option/0","doc":"The maximum number of iterations which after which the Reactor will halt.\n\nDefaults to `:infinity`.","ref":"Reactor.html#t:max_iterations_option/0"},{"type":"type","title":"Reactor.options/0","doc":"","ref":"Reactor.html#t:options/0"},{"type":"function","title":"Reactor.run/4","doc":"Attempt to run a Reactor.","ref":"Reactor.html#run/4"},{"type":"function","title":"Arguments - Reactor.run/4","doc":"* `reactor` - The Reactor to run, either a Reactor DSL module, or a Reactor\n  struct.\n* `inputs` - A map of values passed in to satisfy the Reactor's expected\n  inputs.\n* `context` - An arbitrary map that will be merged into the Reactor context\n  and passed into each step.","ref":"Reactor.html#run/4-arguments"},{"type":"function","title":"Options - Reactor.run/4","doc":"* `:max_concurrency` (`t:pos_integer/0`) - The maximum number of processes to use to run the Reactor\n\n* `:timeout` - How long to allow the Reactor to run for The default value is `:infinity`.\n\n* `:max_iterations` - The maximum number of times to allow the Reactor to loop The default value is `:infinity`.\n\n* `:async?` (`t:boolean/0`) - Whether to allow the Reactor to start processes The default value is `true`.\n\n* `:run_id` (`t:term/0`) - A unique identifier for the Reactor run","ref":"Reactor.html#run/4-options"},{"type":"function","title":"Reactor.run!/4","doc":"Raising version of `run/4`.","ref":"Reactor.html#run!/4"},{"type":"type","title":"Reactor.state/0","doc":"","ref":"Reactor.html#t:state/0"},{"type":"type","title":"Reactor.t/0","doc":"","ref":"Reactor.html#t:t/0"},{"type":"type","title":"Reactor.timeout_option/0","doc":"Specify the amount of execution time after which to halt processing.\n\nNote that this is not a hard limit. The Reactor will stop when the first step\ncompletes _after_ the timeout has expired.\n\nDefaults to `:infinity`.","ref":"Reactor.html#t:timeout_option/0"},{"type":"module","title":"Reactor.Dsl.Argument","doc":"The struct used to store `argument` DSL entities.\n\nSee `d:Reactor.step.argument`.","ref":"Reactor.Dsl.Argument.html"},{"type":"function","title":"Reactor.Dsl.Argument.element/2","doc":"The `element` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#element/2"},{"type":"function","title":"Example - Reactor.Dsl.Argument.element/2","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  input :numbers\n\n  map :double_numbers do\n    source input(:numbers)\n\n    step :double do\n      argument :number, element(:double_numbers)\n\n      run fn args, _ ->\n        {:ok, args.number * 2}\n      end\n    end\n\n    return :double\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#element/2-example"},{"type":"function","title":"Reactor.Dsl.Argument.input/2","doc":"The `input` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#input/2"},{"type":"function","title":"Example - Reactor.Dsl.Argument.input/2","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  input :name\n\n  step :greet do\n    # here: --------↓↓↓↓↓\n    argument :name, input(:name)\n    run fn\n      %{name: nil}, _, _ -> {:ok, \"Hello, World!\"}\n      %{name: name}, _, _ -> {:ok, \"Hello, #{name}!\"}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#input/2-example"},{"type":"function","title":"Extracting nested values - Reactor.Dsl.Argument.input/2","doc":"You can provide a list of keys to extract from a data structure, similar to\n`Kernel.get_in/2` with the condition that the input value is either a struct\nor implements the `Access` protocol.","ref":"Reactor.Dsl.Argument.html#input/2-extracting-nested-values"},{"type":"function","title":"Reactor.Dsl.Argument.result/2","doc":"The `result` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#result/2"},{"type":"function","title":"Example - Reactor.Dsl.Argument.result/2","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  step :whom do\n    run fn _, _ ->\n      {:ok, Enum.random([\"Marty\", \"Doc\", \"Jennifer\", \"Lorraine\", \"George\", nil])}\n    end\n  end\n\n  step :greet do\n    # here: --------↓↓↓↓↓↓\n    argument :name, result(:whom)\n    run fn\n      %{name: nil}, _, _ -> {:ok, \"Hello, World!\"}\n      %{name: name}, _, _ -> {:ok, \"Hello, #{name}!\"}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#result/2-example"},{"type":"function","title":"Extracting nested values - Reactor.Dsl.Argument.result/2","doc":"You can provide a list of keys to extract from a data structure, similar to\n`Kernel.get_in/2` with the condition that the result is either a struct or\nimplements the `Access` protocol.","ref":"Reactor.Dsl.Argument.html#result/2-extracting-nested-values"},{"type":"type","title":"Reactor.Dsl.Argument.t/0","doc":"","ref":"Reactor.Dsl.Argument.html#t:t/0"},{"type":"function","title":"Reactor.Dsl.Argument.value/1","doc":"The `value` template helper for the Reactor DSL.","ref":"Reactor.Dsl.Argument.html#value/1"},{"type":"function","title":"Example - Reactor.Dsl.Argument.value/1","doc":"```elixir\ndefmodule ExampleReactor do\n  use Reactor\n\n  input :number\n\n  step :times_three do\n    argument :lhs, input(:number)\n    # here: -------↓↓↓↓↓\n    argument :rhs, value(3)\n\n    run fn args, _ ->\n      {:ok, args.lhs * args.rhs}\n    end\n  end\nend\n```","ref":"Reactor.Dsl.Argument.html#value/1-example"},{"type":"module","title":"Reactor.Dsl.Around","doc":"The `around` DSL entity struct.\n\nSee `d:Reactor.around`.","ref":"Reactor.Dsl.Around.html"},{"type":"type","title":"Reactor.Dsl.Around.t/0","doc":"","ref":"Reactor.Dsl.Around.html#t:t/0"},{"type":"protocol","title":"Reactor.Dsl.Build","doc":"A protocol which DSL entities must implement.","ref":"Reactor.Dsl.Build.html"},{"type":"function","title":"Reactor.Dsl.Build.build/2","doc":"Build an entity into a Reactor.\n\nThis function is called during conversion of a DSL module into a Reactor\nstruct.  This allows extensions to specify the behaviour of how they want\nto alter the structure of the Reactor.","ref":"Reactor.Dsl.Build.html#build/2"},{"type":"type","title":"Reactor.Dsl.Build.t/0","doc":"All the types that implement this protocol.","ref":"Reactor.Dsl.Build.html#t:t/0"},{"type":"function","title":"Reactor.Dsl.Build.verify/2","doc":"Perform any after-compilation verification that is needed to make the entity\nwork.\n\nSee `Spark.Dsl.Verifier` for more information.","ref":"Reactor.Dsl.Build.html#verify/2"},{"type":"module","title":"Reactor.Dsl.Collect","doc":"The struct used to store collect DSL entities.\n\nSee `d:Reactor.collect`.","ref":"Reactor.Dsl.Collect.html"},{"type":"type","title":"Reactor.Dsl.Collect.t/0","doc":"","ref":"Reactor.Dsl.Collect.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Compose","doc":"The `compose` DSL entity struct.\n\nSee the `d:Reactor.compose`.","ref":"Reactor.Dsl.Compose.html"},{"type":"type","title":"Reactor.Dsl.Compose.t/0","doc":"","ref":"Reactor.Dsl.Compose.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Debug","doc":"The `debug` DSL entity struct.\n\nSee `d:Reactor.debug`.","ref":"Reactor.Dsl.Debug.html"},{"type":"type","title":"Reactor.Dsl.Debug.t/0","doc":"","ref":"Reactor.Dsl.Debug.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Flunk","doc":"The struct used to store flunk DSL entities.\n\nSee `d:Reactor.flunk`.","ref":"Reactor.Dsl.Flunk.html"},{"type":"type","title":"Reactor.Dsl.Flunk.t/0","doc":"","ref":"Reactor.Dsl.Flunk.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Group","doc":"The `group` DSL entity struct.\n\nSee `d:Reactor.group`.","ref":"Reactor.Dsl.Group.html"},{"type":"type","title":"Reactor.Dsl.Group.t/0","doc":"","ref":"Reactor.Dsl.Group.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Guard","doc":"A struct used to store the `guard` DSL entity.\n\nSee `d:Reactor.step.guard`","ref":"Reactor.Dsl.Guard.html"},{"type":"type","title":"Reactor.Dsl.Guard.t/0","doc":"","ref":"Reactor.Dsl.Guard.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Info","doc":"Introspection for the Reactor DSL.","ref":"Reactor.Dsl.Info.html"},{"type":"function","title":"Reactor.Dsl.Info.reactor/1","doc":"reactor DSL entities","ref":"Reactor.Dsl.Info.html#reactor/1"},{"type":"function","title":"Reactor.Dsl.Info.reactor_middlewares/1","doc":"reactor.middlewares DSL entities","ref":"Reactor.Dsl.Info.html#reactor_middlewares/1"},{"type":"function","title":"Reactor.Dsl.Info.reactor_options/1","doc":"reactor DSL options\n\nReturns a map containing the and any configured or default values.","ref":"Reactor.Dsl.Info.html#reactor_options/1"},{"type":"function","title":"Reactor.Dsl.Info.reactor_return/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Dsl.Info.html#reactor_return/1"},{"type":"function","title":"Reactor.Dsl.Info.reactor_return!/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Dsl.Info.html#reactor_return!/1"},{"type":"function","title":"Reactor.Dsl.Info.to_struct/1","doc":"Convert a reactor DSL module into a reactor struct.","ref":"Reactor.Dsl.Info.html#to_struct/1"},{"type":"function","title":"Reactor.Dsl.Info.to_struct!/1","doc":"Raising version of `to_struct/1`.","ref":"Reactor.Dsl.Info.html#to_struct!/1"},{"type":"module","title":"Reactor.Dsl.Input","doc":"The struct used to store input DSL entities.\n\nSee `d:Reactor.input`.","ref":"Reactor.Dsl.Input.html"},{"type":"type","title":"Reactor.Dsl.Input.t/0","doc":"","ref":"Reactor.Dsl.Input.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Map","doc":"The `map` DSL entity struct.\n\nSee `d:Reactor.map`.","ref":"Reactor.Dsl.Map.html"},{"type":"type","title":"Reactor.Dsl.Map.t/0","doc":"","ref":"Reactor.Dsl.Map.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Middleware","doc":"The `middleware` DSL entity struct.\n\nSee `d:Reactor.middleware.middleware`.","ref":"Reactor.Dsl.Middleware.html"},{"type":"type","title":"Reactor.Dsl.Middleware.t/0","doc":"","ref":"Reactor.Dsl.Middleware.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Recurse","doc":"The `recurse` DSL entity struct.\n\nSee the `d:Reactor.recurse`.","ref":"Reactor.Dsl.Recurse.html"},{"type":"type","title":"Reactor.Dsl.Recurse.t/0","doc":"","ref":"Reactor.Dsl.Recurse.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Step","doc":"The struct used to store step DSL entities.\n\nSee `d:Reactor.step`.","ref":"Reactor.Dsl.Step.html"},{"type":"type","title":"Reactor.Dsl.Step.t/0","doc":"","ref":"Reactor.Dsl.Step.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch","doc":"The `switch` DSL entity struct.\n\nSee `d:Reactor.switch`.","ref":"Reactor.Dsl.Switch.html"},{"type":"type","title":"Reactor.Dsl.Switch.t/0","doc":"","ref":"Reactor.Dsl.Switch.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch.Default","doc":"The `default` DSL entity struct.\n\nSee `d:Reactor.switch.default`.","ref":"Reactor.Dsl.Switch.Default.html"},{"type":"type","title":"Reactor.Dsl.Switch.Default.t/0","doc":"","ref":"Reactor.Dsl.Switch.Default.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Switch.Match","doc":"The `matches?` DSL entity struct.\n\nSee `d:Reactor.switch.matches?`.","ref":"Reactor.Dsl.Switch.Match.html"},{"type":"type","title":"Reactor.Dsl.Switch.Match.t/0","doc":"","ref":"Reactor.Dsl.Switch.Match.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Template","doc":"The `template` DSL entity struct.\n\nSee `d:Reactor.template`.","ref":"Reactor.Dsl.Template.html"},{"type":"type","title":"Reactor.Dsl.Template.t/0","doc":"","ref":"Reactor.Dsl.Template.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Verifier","doc":"Runs `Reactor.Dsl.Build.verify/2` for all the entities in the reactor.","ref":"Reactor.Dsl.Verifier.html"},{"type":"module","title":"Reactor.Dsl.WaitFor","doc":"The struct used to store `wait_for` DSL entities.\n\nSee `d:Reactor.step.wait_for`.","ref":"Reactor.Dsl.WaitFor.html"},{"type":"type","title":"Reactor.Dsl.WaitFor.t/0","doc":"","ref":"Reactor.Dsl.WaitFor.html#t:t/0"},{"type":"module","title":"Reactor.Dsl.Where","doc":"A struct used to store the `where` DSL entity.\n\nSee `d:Reactor.step.where`.","ref":"Reactor.Dsl.Where.html"},{"type":"type","title":"Reactor.Dsl.Where.t/0","doc":"","ref":"Reactor.Dsl.Where.html#t:t/0"},{"type":"behaviour","title":"Reactor.Step","doc":"The Step behaviour and struct.\n\nImplement this behaviour to make steps for your Reactor.","ref":"Reactor.Step.html"},{"type":"type","title":"Reactor.Step.arguments/0","doc":"","ref":"Reactor.Step.html#t:arguments/0"},{"type":"function","title":"Reactor.Step.async?/1","doc":"Is the step able to be run asynchronously?","ref":"Reactor.Step.html#async?/1"},{"type":"callback","title":"Reactor.Step.async?/1","doc":"Detect if the step can be run asynchronously at runtime.\n\n> This callback is automatically defined by `use Reactor.Step` however you're\n> free to override it if you need a specific behaviour.\n\nThis callback is called when Reactor is deciding whether to run a step\nasynchronously.\n\nThe default implementation of this callback checks returns the the value of\nthe steps's `async?` key if it is boolean, or calls it with the steps's\noptions if it is a function.","ref":"Reactor.Step.html#c:async?/1"},{"type":"function","title":"Reactor.Step.can?/2","doc":"Find out of a step has a capability.","ref":"Reactor.Step.html#can?/2"},{"type":"callback","title":"Reactor.Step.can?/2","doc":"Detect the capabilities of the step at runtime.\n\n> This callback is automatically defined by `use Reactor.Step` however you're\n> free to override it if you need specific behaviour.\n\nWhenever Reactor would like to either undo a change made by the step, or\ncompensate a step failure this callback is called to detect whether the step\nmodule is capable of the desired action.\n\nThe default implementation of this callback checks to see if the optional\ncallback is defined on the current module.","ref":"Reactor.Step.html#c:can?/2"},{"type":"type","title":"Reactor.Step.capability/0","doc":"Optional capabilities which may be implemented by the step module.\n\nThis allows us to optimise out calls steps which cannot be undone, etc.","ref":"Reactor.Step.html#t:capability/0"},{"type":"function","title":"Reactor.Step.compensate/4","doc":"Compensate a step","ref":"Reactor.Step.html#compensate/4"},{"type":"callback","title":"Reactor.Step.compensate/4","doc":"Compensate for the failure of the step.\n\n> Do not implement this callback if your step doesn't support compensation.\n\nIf `run/3` returned an error then this callback will be called the error\nreason and the original arguments.\n\nThis provides you the opportunity to handle the error in a number of ways and\ndirect the reactor as to what to do next.","ref":"Reactor.Step.html#c:compensate/4"},{"type":"callback","title":"Arguments - Reactor.Step.compensate/4","doc":"- `reason` - the error reason returned from `c:run/3`.\n  - `arguments` - the arguments passed to the step.\n  - `context` - the reactor context.\n  - `options` - a keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:compensate/4-arguments"},{"type":"callback","title":"Return values - Reactor.Step.compensate/4","doc":"- `{:continue, value}` if you're able to provide a valid result for the step\n    (perhaps by re-running the original computation) then return that within a\n    `:continue` tuple and execution will continue as planned.\n  - `:ok` the step was successfully compensated and the reactor should\n    continue undoing upstream changes.\n  - `:retry` or `{:retry, reason}` if you would like the reactor to attempt to\n    re-run the step. You can optionally supply an error reason which will be\n    used in the event that the step runs out of retries, otherwise a\n    `Reactor.Error.Invalid.RetriesExceededError` will be used.\n  - `{:error, reason}` if compensation was unsuccessful.","ref":"Reactor.Step.html#c:compensate/4-return-values"},{"type":"type","title":"Reactor.Step.compensate_result/0","doc":"Possible valid return values for the `c:compensate/4` callback.","ref":"Reactor.Step.html#t:compensate_result/0"},{"type":"type","title":"Reactor.Step.context/0","doc":"","ref":"Reactor.Step.html#t:context/0"},{"type":"function","title":"Reactor.Step.run/3","doc":"Execute a step.","ref":"Reactor.Step.html#run/3"},{"type":"callback","title":"Reactor.Step.run/3","doc":"Execute the step.\n\nThis is the function that implements the behaviour you wish to execute.  You\nwill receive arguments as per the `t:Step.t` definition along with their\ncorresponding values as a map and a copy of the current reactor context.","ref":"Reactor.Step.html#c:run/3"},{"type":"callback","title":"Arguments - Reactor.Step.run/3","doc":"- `arguments` - A map of arguments as per the `t:Step.t` definition we're\n    called from.\n  - `context` - The reactor context.\n  - `options` - A keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:run/3-arguments"},{"type":"callback","title":"Return values - Reactor.Step.run/3","doc":"- `{:ok, value}` the step completed successfully it returns the value in an\n    ok tuple.\n  - `{:ok, value, [step]}` the step completed successfully and wants to add\n    new steps to the reactor.\n  - `{:error, reason}` the if step failed, return an error tuple.\n  - `:retry` or `{:retry, reason}` the step failed, but is retryable.  You can\n    optionally supply an error reason which will be used in the event that the\n    step runs out of retries, otherwise a `Reactor.Error.RetriesExceededError`\n    will be used.\n  - `{:halt, reason}` terminate (or pause) reactor execution.  If there are\n    actively running steps the reactor will wait for them to finish and then\n    return the incomplete state for later resumption.","ref":"Reactor.Step.html#c:run/3-return-values"},{"type":"type","title":"Reactor.Step.run_result/0","doc":"Possible valid return values for the `c:run/3` callback.","ref":"Reactor.Step.html#t:run_result/0"},{"type":"type","title":"Reactor.Step.step/0","doc":"","ref":"Reactor.Step.html#t:step/0"},{"type":"type","title":"Reactor.Step.t/0","doc":"","ref":"Reactor.Step.html#t:t/0"},{"type":"function","title":"Reactor.Step.undo/4","doc":"Undo a step","ref":"Reactor.Step.html#undo/4"},{"type":"callback","title":"Reactor.Step.undo/4","doc":"Undo a previously successful execution of the step.\n\n> Do not implement this callback if your step doesn't support undoing.\n\nThis callback is called when the reactor encounters an unhandled error later\nin it's execution run and must undo the work previously done.","ref":"Reactor.Step.html#c:undo/4"},{"type":"callback","title":"Arguments - Reactor.Step.undo/4","doc":"- `value` - the return value of the previously successful call to `c:run/3`.\n  - `arguments` - the arguments passed to the step.\n  - `context` - the reactor context.\n  - `options` - a keyword list of options provided to the step (if any).","ref":"Reactor.Step.html#c:undo/4-arguments"},{"type":"callback","title":"Return values - Reactor.Step.undo/4","doc":"- `:ok` the step was successfully undo and the reactor should continue\n    rolling back.\n  - `{:error, reason}` there was an error while attempting to compensate.  The\n    reactor will collect the error and continue rolling back.\n  - `:retry` if you would like the reactor to attempt to undo the again later\n    - possibly in the case of a network failure for example.","ref":"Reactor.Step.html#c:undo/4-return-values"},{"type":"type","title":"Reactor.Step.undo_result/0","doc":"Possible valid return values for the `c:undo/4` callback.","ref":"Reactor.Step.html#t:undo_result/0"},{"type":"module","title":"Reactor.Step.AnonFn","doc":"The built-in step for executing in-line DSL anonymous functions.","ref":"Reactor.Step.AnonFn.html"},{"type":"module","title":"Options - Reactor.Step.AnonFn","doc":"* `run` - a one or two arity function or MFA which will be called as the run\n  function of the step.\n* `compensate` - a one to three arity function or MFA which will be called as\n  the compensate function of the step.  Optional.\n* `undo` - a one to three arity function or MFA which will be called as the\n  undo function of this step.  Optional.","ref":"Reactor.Step.AnonFn.html#module-options"},{"type":"module","title":"Reactor.Step.AnonFn.Mermaid","doc":"Mermaid rendering for `AnonFn` steps.","ref":"Reactor.Step.AnonFn.Mermaid.html"},{"type":"module","title":"Reactor.Step.Around","doc":"Wrap the execution of a number of steps in a function.\n\nThis allows you to provide custom context and filter the provided steps as\nneeded.","ref":"Reactor.Step.Around.html"},{"type":"module","title":"Options - Reactor.Step.Around","doc":"* `fun` - a four-arity function that will be called when executing this step.\n* `steps` - a list of steps which are will be provided to the above mentioned\n  function.\n* `allow_async?` - a boolean indicating whether the nested steps can be\n  executed asynchronously or must remain within the current process.","ref":"Reactor.Step.Around.html#module-options"},{"type":"module","title":"Wrapper function - Reactor.Step.Around","doc":"Your around function will be called by this step and will be passed the\nfollowing arguments:\n\n* `arguments` - the arguments passed to the step.\n* `context` - the context passed to the step.\n* `steps` - the list of steps passed in the options.\n* `callback` - a 3 arity function that you can call to execute steps.\n\nThis provides you the opportunity to modify the arguments, context and list of\nsteps to be executed.  You then can call the callback with the modified\narguments, context and steps and they will be executed in a Reactor of their\nown.  The callback will return `{:ok, results}` where results is a map of all\nof the step results by name, or an error tuple.\n\nYou can then modify the result in any way before returning it as the return of\nthe around step.","ref":"Reactor.Step.Around.html#module-wrapper-function"},{"type":"module","title":"Callback function - Reactor.Step.Around","doc":"The callback function will spawn a separate Reactor and run provided steps to\ncompletion using `arguments` as input.\n\nIt expects the following three arguments to be passed:\n\n1. `arguments` - a map of arguments to be used as input to the nested Reactor.\n2. `context` - the context passed to the nested Reactor.\n3. `steps` - the list of steps which will be executed in the nested Reactor.","ref":"Reactor.Step.Around.html#module-callback-function"},{"type":"module","title":"Example - Reactor.Step.Around","doc":"You could use a function like that below to cause some steps to be executed\ninside an Ecto database transaction.\n\n```elixir\ndef in_transaction(arguments, context, steps, callback) do\n  MyApp.Repo.transaction(fn ->\n    case callback.(arguments, context, steps) do\n      {:ok, results} -> result\n      {:error, reason} -> raise reason\n    end\n  end)\nend\n```","ref":"Reactor.Step.Around.html#module-example"},{"type":"type","title":"Reactor.Step.Around.allow_async_option/0","doc":"Should the inner Reactor be allowed to run tasks asynchronously?\n\nOptional. Defaults to `true`.","ref":"Reactor.Step.Around.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Around.around_fun/0","doc":"The type signature for the \"around\" function.","ref":"Reactor.Step.Around.html#t:around_fun/0"},{"type":"type","title":"Reactor.Step.Around.callback/0","doc":"The type signature for the provided callback function.","ref":"Reactor.Step.Around.html#t:callback/0"},{"type":"type","title":"Reactor.Step.Around.function_option/0","doc":"The MFA or 4-arity function which this step will call.","ref":"Reactor.Step.Around.html#t:function_option/0"},{"type":"type","title":"Reactor.Step.Around.options/0","doc":"","ref":"Reactor.Step.Around.html#t:options/0"},{"type":"type","title":"Reactor.Step.Around.steps_option/0","doc":"The initial steps to pass into the \"around\" function.\n\nOptional.","ref":"Reactor.Step.Around.html#t:steps_option/0"},{"type":"module","title":"Reactor.Step.Around.Mermaid","doc":"Mermaid rendering for `around` steps.","ref":"Reactor.Step.Around.Mermaid.html"},{"type":"module","title":"Reactor.Step.Compose","doc":"A built-in step which can embed one reactor inside another.\n\nThis step calls `Reactor.run/3` on the inner reactor and returns it's result.\nReactor will correctly share the concurrency availability over both the parent\nand child Reactors.","ref":"Reactor.Step.Compose.html"},{"type":"module","title":"Reactor.Step.Compose.Mermaid","doc":"Mermaid rendering for `compose` steps.","ref":"Reactor.Step.Compose.Mermaid.html"},{"type":"module","title":"Reactor.Step.Debug","doc":"A very simple step which sends debug information to `Logger`.","ref":"Reactor.Step.Debug.html"},{"type":"module","title":"Options - Reactor.Step.Debug","doc":"* `level` - the log level to use.","ref":"Reactor.Step.Debug.html#module-options"},{"type":"type","title":"Reactor.Step.Debug.level_option/0","doc":"","ref":"Reactor.Step.Debug.html#t:level_option/0"},{"type":"type","title":"Reactor.Step.Debug.options/0","doc":"","ref":"Reactor.Step.Debug.html#t:options/0"},{"type":"module","title":"Reactor.Step.Fail","doc":"A very simple step which immediately returns an error.","ref":"Reactor.Step.Fail.html"},{"type":"module","title":"Reactor.Step.Group","doc":"Wrap the execution of a number of steps with before/after functions.\n\nUnlike `Reactor.Step.Around`, this step doesn't need to run a nested Reactor\ninstance, but instead can emit the steps directly into the parent Reactor.","ref":"Reactor.Step.Group.html"},{"type":"module","title":"Options - Reactor.Step.Group","doc":"* `before` - a three-arity function that will be called before running any\n  child steps.\n* `after` - a three-arity function that will be called after running any\n  emitted steps.\n* `allow_async?` - a boolean indicating whether the emitted steps can be\n  executed asynchronously or must remain within the current process.","ref":"Reactor.Step.Group.html#module-options"},{"type":"module","title":"Before function - Reactor.Step.Group","doc":"The before function will be passed the following arguments:\n\n1. `arguments` - the values of any step arguments needed by the group.\n2. `context` - the Reactor context.\n3. `steps` - the list of steps passed in the options.\n\nThis provides you the opportunity to modify the arguments, context and list of\nsteps to be executed.\n\nThe successful return value should be `{:ok, arguments, context, steps}`.  The\nreturned arguments will be used to provide any `input` arguments to nested\nsteps.","ref":"Reactor.Step.Group.html#module-before-function"},{"type":"module","title":"Example - Reactor.Step.Group","doc":"```elixir\ndef no_time_travel(arguments, context, steps) do\n  steps = steps\n    |> Enum.filter(&(&1.name == :program_time_circuits))\n\n  arguments = arguments\n    |> Map.delete(:destination_time)\n\n  {:ok, arguments, context, steps}\nend\n```","ref":"Reactor.Step.Group.html#module-example"},{"type":"module","title":"After function - Reactor.Step.Group","doc":"The after function will be called with a single argument; a map of the nested\nstep results.\n\nThe successful return value should be `{:ok, any}` where `any` will be treated\nas the result of the group.\n\n```elixir\ndef find_current_year(results) do\n  case Map.fetch(results, :time_circuits) do\n    {:ok, %{present_time: present_time}} -> {:ok, present_time.year}\n    _ -> {:error, \"Unable to read the present time from the time circuits\"}\n  end\nend\n```","ref":"Reactor.Step.Group.html#module-after-function"},{"type":"type","title":"Reactor.Step.Group.after_fun/0","doc":"The after function.","ref":"Reactor.Step.Group.html#t:after_fun/0"},{"type":"type","title":"Reactor.Step.Group.after_option/0","doc":"The MFA or 1-arity function which this step will call after successfully\nrunning the steps.","ref":"Reactor.Step.Group.html#t:after_option/0"},{"type":"type","title":"Reactor.Step.Group.allow_async_option/0","doc":"Should the emitted steps be allowed to run asynchronously?\n\nOptional. Defaults to `true`.","ref":"Reactor.Step.Group.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Group.before_fun/0","doc":"The before function.","ref":"Reactor.Step.Group.html#t:before_fun/0"},{"type":"type","title":"Reactor.Step.Group.before_option/0","doc":"The MFA or 3-arity function which this step will call before running any\nsteps.","ref":"Reactor.Step.Group.html#t:before_option/0"},{"type":"type","title":"Reactor.Step.Group.options/0","doc":"","ref":"Reactor.Step.Group.html#t:options/0"},{"type":"type","title":"Reactor.Step.Group.steps_option/0","doc":"The initial steps to pass into the \"before\" function.\n\nOptional.","ref":"Reactor.Step.Group.html#t:steps_option/0"},{"type":"module","title":"Reactor.Step.Group.Mermaid","doc":"Mermaid rendering for `group` steps.","ref":"Reactor.Step.Group.Mermaid.html"},{"type":"module","title":"Reactor.Step.Map","doc":"Given an iterable input run the provided steps for each element and collect\nthe results into a new value.\n\n> #### A note on ordering {: .tip}\n>\n> If your application doesn't need the results back in the same order that\n> they were provided then setting `strict_ordering?` to `false` will increase\n> performance - especially on large input sets.","ref":"Reactor.Step.Map.html"},{"type":"module","title":"Options - Reactor.Step.Map","doc":"* `:state` - Required. The current execution state of the map.  This is required because it's recursive.\n   Valid values are :init, :iterating\n\n* `:batch_size` (`t:pos_integer/0`) - The number of elements to consume off the source when emitting steps. The default value is `100`.\n\n* `:steps` (list of struct of type `Reactor.Step`) - Required. The steps to use when mapping source elements.\n\n* `:return` (`t:atom/0`) - Required. The name of the step whose result will be used as the new value for each element.\n\n* `:strict_ordering?` (`t:boolean/0`) - Whether the mapped values must be returned in the same order that they were provided. The default value is `true`.\n\n* `:allow_async?` (`t:boolean/0`) - Whether the emitted steps should be allowed to run asynchronously. The default value is `true`.\n\n* `:descendant_step_names` (struct of type `MapSet`) - The cached names of all descendant steps to aid rewriting. You don't need to provide this value - it is calculated by the init pass.\n\n* `:extra_arguments` (list of struct of type `Reactor.Argument`) - Extra arguments to be passed by to every nested step.","ref":"Reactor.Step.Map.html#module-options"},{"type":"module","title":"Reactor.Step.Map.Mermaid","doc":"Mermaid rendering for `map` steps.","ref":"Reactor.Step.Map.Mermaid.html"},{"type":"module","title":"Reactor.Step.Recurse","doc":"A built-in step which can recursively execute a reactor.\n\nThis step executes a reactor repeatedly, using the output of each execution\nas the input for the next, until either:\n\n1. The exit condition is met\n2. The maximum number of iterations is reached\n3. An error occurs\n\nReactor will correctly share the concurrency availability between iterations.","ref":"Reactor.Step.Recurse.html"},{"type":"module","title":"Reactor.Step.Recurse.Mermaid","doc":"Mermaid rendering for `compose` steps.","ref":"Reactor.Step.Recurse.Mermaid.html"},{"type":"module","title":"Reactor.Step.ReturnAllArguments","doc":"A very simple step which simply returns all it's arguments unchanged.","ref":"Reactor.Step.ReturnAllArguments.html"},{"type":"module","title":"Reactor.Step.ReturnArgument","doc":"A very simple step which simply returns the named argument, if provided.","ref":"Reactor.Step.ReturnArgument.html"},{"type":"module","title":"Options. - Reactor.Step.ReturnArgument","doc":"* `argument` - the name of the argument to return.","ref":"Reactor.Step.ReturnArgument.html#module-options"},{"type":"module","title":"Reactor.Step.Switch","doc":"Conditionally decide which steps should be run at runtime.","ref":"Reactor.Step.Switch.html"},{"type":"module","title":"Options - Reactor.Step.Switch","doc":"* `matches` - a list of match consisting of predicates and a list of steps to\n  execute if the predicate returns a truthy value.  See `t:matches` for more\n  information.  Required.\n* `default` - a list of steps to execute if none of the predicates match.\n  Optional.\n* `allow_async?` - a boolean indicating whether to allow the steps to be\n  executed asynchronously.  Optional.  Defaults to `true`.\n* `on` - the name of the argument to pass into the predicates.  If this\n  argument is not provided to this step, then an error will be returned.","ref":"Reactor.Step.Switch.html#module-options"},{"type":"module","title":"Branching behaviour - Reactor.Step.Switch","doc":"Each of the predicates in `matches` are tried in order, until either one\nreturns a truthy value, or all the matches are exhausted.\n\nIf there is a match, then the matching steps are emitted into the parent\nrunning Reactor.\n\nIn the case that no match is found, then the steps provided in the `default`\noption are emitted.  If no default is provided, then an error is returned.\n\n> #### Tip {: .tip}\n>\n> Execution of predicates stops once the first match is found.  This means\n> that if multiple predicates potentially match, the subsequent ones will\n> never be called.","ref":"Reactor.Step.Switch.html#module-branching-behaviour"},{"type":"module","title":"Returning - Reactor.Step.Switch","doc":"By default the step returns `nil` as it's result.\n\nYou can have the step return the result of a branch by adding a step to the\nbranch with the same name as the switch which returns the expected value.\nThis will be handled by normal Reactor step emission rules.","ref":"Reactor.Step.Switch.html#module-returning"},{"type":"type","title":"Reactor.Step.Switch.allow_async_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:allow_async_option/0"},{"type":"type","title":"Reactor.Step.Switch.default_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:default_option/0"},{"type":"type","title":"Reactor.Step.Switch.match_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:match_option/0"},{"type":"type","title":"Reactor.Step.Switch.matches/0","doc":"A list of predicates and steps to execute if the predicate returns a truthy\nvalue.","ref":"Reactor.Step.Switch.html#t:matches/0"},{"type":"type","title":"Reactor.Step.Switch.on_option/0","doc":"","ref":"Reactor.Step.Switch.html#t:on_option/0"},{"type":"type","title":"Reactor.Step.Switch.options/0","doc":"","ref":"Reactor.Step.Switch.html#t:options/0"},{"type":"type","title":"Reactor.Step.Switch.predicate/0","doc":"A predicate is a 1-arity function.  It can return anything.  Any result which\nis not `nil` or `false` is considered true.","ref":"Reactor.Step.Switch.html#t:predicate/0"},{"type":"module","title":"Reactor.Step.Switch.Mermaid","doc":"Mermaid rendering for `switch` steps.","ref":"Reactor.Step.Switch.Mermaid.html"},{"type":"module","title":"Reactor.Step.Template","doc":"A step which renders an `EEx` template using it's arguments as the assigns.","ref":"Reactor.Step.Template.html"},{"type":"module","title":"Options - Reactor.Step.Template","doc":"* `:template` (`t:String.t/0`) - Required. The EEx template\n\n* `:trim` (`t:boolean/0`) - Whether to trim whitespace before and after the template. See `EEx#options` The default value is `false`.","ref":"Reactor.Step.Template.html#module-options"},{"type":"module","title":"Reactor.Step.Transform","doc":"The built-in step for executing input and argument transformations.\n\nExpects a single argument named `value` which contains the value to be\ntransformed.","ref":"Reactor.Step.Transform.html"},{"type":"module","title":"Options - Reactor.Step.Transform","doc":"* `fun` - a one or two arity function or MFA to use to modify the `value`\n  argument.\n\n> #### Tip {: .tip}\n>\n> This step is emitted by the builder when an argument needs to be transformed\n> before being passed into a step.\n>\n> Most likely you will never need to use this step directly.","ref":"Reactor.Step.Transform.html#module-options"},{"type":"module","title":"Reactor.Step.Transform.Mermaid","doc":"Mermaid rendering for transform steps.","ref":"Reactor.Step.Transform.Mermaid.html"},{"type":"module","title":"Reactor.Step.TransformAll","doc":"A built-in step which applies a transformation function to all it's arguments.\n\nThe returned map is used as the arguments to the step, instead of the step's\ndefined arguments.\n\n\n> #### Tip {: .tip}\n>\n> This step will be emitted by the builder when a step wants to transform all\n> it's arguments.\n>\n> Most likely you will never need to use this step directly.","ref":"Reactor.Step.TransformAll.html"},{"type":"behaviour","title":"Reactor.Middleware","doc":"The Middleware behaviour.\n\nBy implementing this behaviour you can modify the internal state of the\nReactor during startup, execution and shutdown.\n\nMiddlewares can be added to the reactor either with the `middlewares` DSL\nsection or by the `add_middleware/2`, etc, functions in `Reactor.Builder`.\n\nAll callbacks in this behaviour are optional so you can define only the parts\nyou need for your feature.","ref":"Reactor.Middleware.html"},{"type":"callback","title":"Reactor.Middleware.complete/2","doc":"The complete callback will be called with the successful result of the\nReactor.\n\nThis gives you the opportunity to modify the return value or to perform clean\nup of any non-reactor-managed resources (eg notifications).\n\nNote that these callbacks are called in an arbitrary order meaning that the\nresult value passed may have already been altered by another callback.\n\nIf any callback returns an error then any remaining callbacks will not\nbe called.","ref":"Reactor.Middleware.html#c:complete/2"},{"type":"type","title":"Reactor.Middleware.context/0","doc":"","ref":"Reactor.Middleware.html#t:context/0"},{"type":"callback","title":"Reactor.Middleware.error/2","doc":"The error callback will be called the final error value(s) of the Reactor.\n\nThis gives you the opportunity to modify the return value or to perform clean\nup of any non-reactor-managed resources (eg notifications).\n\nNote that these callbacks are called in an arbitrary order meaning that the\nerror value passed may have already been altered by another callback.\n\nHere a return value of `:ok` will continue calls to other callbacks without\nmodifying the error value.","ref":"Reactor.Middleware.html#c:error/2"},{"type":"type","title":"Reactor.Middleware.error_or_errors/0","doc":"","ref":"Reactor.Middleware.html#t:error_or_errors/0"},{"type":"callback","title":"Reactor.Middleware.event/3","doc":"Receive events about the execution of the Reactor.\n\nThis callback will block the Reactor, so it's encouraged that you do anything\nexpensive in another process.","ref":"Reactor.Middleware.html#c:event/3"},{"type":"callback","title":"Reactor.Middleware.get_process_context/0","doc":"Called before starting an asynchronous step in order to retrieve any context\ninformation that needs to be passed into the new process.\n\nThis is most likely used by tracers to copy span information from the parent\nprocess to the child process.","ref":"Reactor.Middleware.html#c:get_process_context/0"},{"type":"callback","title":"Reactor.Middleware.halt/1","doc":"The halt callback will be called with the Reactor context when halting.\n\nThis allows you to clean up any non-reactor-managed resources or modify the\ncontext for later re-use by a future `init/1` callback.","ref":"Reactor.Middleware.html#c:halt/1"},{"type":"callback","title":"Reactor.Middleware.init/1","doc":"The init callback will be called with the Reactor context when starting up.\n\nThis gives you the opportunity to modify the context or to perform any\ninitialisation of any non-reactor-managed resources (eg notifications).","ref":"Reactor.Middleware.html#c:init/1"},{"type":"type","title":"Reactor.Middleware.result/0","doc":"","ref":"Reactor.Middleware.html#t:result/0"},{"type":"callback","title":"Reactor.Middleware.set_process_context/1","doc":"Called inside a new asynchronous step in order to set context information into\nthe new process.\n\nThis is most likely used by tracers to copy span information from the parent\nprocess to the child process.","ref":"Reactor.Middleware.html#c:set_process_context/1"},{"type":"type","title":"Reactor.Middleware.step_event/0","doc":"","ref":"Reactor.Middleware.html#t:step_event/0"},{"type":"type","title":"Reactor.Middleware.t/0","doc":"","ref":"Reactor.Middleware.html#t:t/0"},{"type":"module","title":"Reactor.Middleware.Telemetry","doc":"A Reactor middleware that emits telemetry events.\n\nThe following events are emitted:\n\n* `[:reactor, :run, :start]`\n* `[:reactor, :run, :stop]`\n* `[:reactor, :step, :run, :start]`\n* `[:reactor, :step, :run, :stop]`\n* `[:reactor, :step, :process, :start]`\n* `[:reactor, :step, :process, :stop]`\n* `[:reactor, :step, :compensate, :start]`\n* `[:reactor, :step, :compensate, :stop]`\n* `[:reactor, :step, :undo, :start]`\n* `[:reactor, :step, :undo, :stop]`\n\nYou can provide additional metadata by placing a map under the `telemetry_metadata` context key.","ref":"Reactor.Middleware.Telemetry.html"},{"type":"module","title":"Reactor.Error","doc":"Uses `splode` to manage various classes of error.","ref":"Reactor.Error.html"},{"type":"macro","title":"Reactor.Error.__using__/1","doc":"Convenience wrapper around `use Splode.Error`","ref":"Reactor.Error.html#__using__/1"},{"type":"type","title":"Reactor.Error.class/0","doc":"","ref":"Reactor.Error.html#t:class/0"},{"type":"type","title":"Reactor.Error.class_module/0","doc":"","ref":"Reactor.Error.html#t:class_module/0"},{"type":"type","title":"Reactor.Error.error_class/0","doc":"","ref":"Reactor.Error.html#t:error_class/0"},{"type":"function","title":"Reactor.Error.fetch_error/2","doc":"Recursively searches through nested reactor errors for the first instance of\nthe provided module.","ref":"Reactor.Error.html#fetch_error/2"},{"type":"function","title":"Reactor.Error.find_errors/2","doc":"Recursively searches through nested reactor errors and returns any errors\nwhich match the provided module name.","ref":"Reactor.Error.html#find_errors/2"},{"type":"function","title":"Reactor.Error.splode_error?/2","doc":"","ref":"Reactor.Error.html#splode_error?/2"},{"type":"type","title":"Reactor.Error.t/0","doc":"","ref":"Reactor.Error.html#t:t/0"},{"type":"function","title":"Reactor.Error.unwrap!/2","doc":"Raises an error if the result is an error, otherwise returns the result\n\nAlternatively, you can use the `defsplode` macro, which does this automatically.","ref":"Reactor.Error.html#unwrap!/2"},{"type":"function","title":"Options - Reactor.Error.unwrap!/2","doc":"- `:error_opts` - Options to pass to `to_error/2` when converting the returned error\n- `:unknown_error_opts` - Options to pass to the unknown error if the function returns only `:error`.\n  not necessary if your function always returns `{:error, error}`.","ref":"Reactor.Error.html#unwrap!/2-options"},{"type":"function","title":"Examples - Reactor.Error.unwrap!/2","doc":"def function(arg) do\n    case do_something(arg) do\n      :success -> :ok\n      {:success, result} -> {:ok, result}\n      {:error, error} -> {:error, error}\n    end\n  end\n\n  def function!(arg) do\n    YourErrors.unwrap!(function(arg))\n  end","ref":"Reactor.Error.html#unwrap!/2-examples"},{"type":"exception","title":"Reactor.Error.Internal","doc":"The [Splode error class](e:splode:get-started-with-splode.html#error-classes)\nfor Reactor-caused errors.","ref":"Reactor.Error.Internal.html"},{"type":"function","title":"Reactor.Error.Internal.exception/1","doc":"Create an `Elixir.Reactor.Error.Internal` without raising it.","ref":"Reactor.Error.Internal.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Internal.exception/1","doc":"- :errors","ref":"Reactor.Error.Internal.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Internal.ComposeError","doc":"This error is returned when two Reactors cannot be composed together.","ref":"Reactor.Error.Internal.ComposeError.html"},{"type":"function","title":"Reactor.Error.Internal.ComposeError.exception/1","doc":"Create an `Elixir.Reactor.Error.Internal.ComposeError` without raising it.","ref":"Reactor.Error.Internal.ComposeError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Internal.ComposeError.exception/1","doc":"- :arguments\n- :inner_reactor\n- :message\n- :outer_reactor","ref":"Reactor.Error.Internal.ComposeError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Internal.MissingReturnResultError","doc":"This error is returned when the Reactor's return name doesn't match any of the\nknown step results.","ref":"Reactor.Error.Internal.MissingReturnResultError.html"},{"type":"function","title":"Reactor.Error.Internal.MissingReturnResultError.exception/1","doc":"Create an `Elixir.Reactor.Error.Internal.MissingReturnResultError` without raising it.","ref":"Reactor.Error.Internal.MissingReturnResultError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Internal.MissingReturnResultError.exception/1","doc":"- :reactor","ref":"Reactor.Error.Internal.MissingReturnResultError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Internal.PlanError","doc":"This error is returned when the step graph cannot be built.","ref":"Reactor.Error.Internal.PlanError.html"},{"type":"function","title":"Reactor.Error.Internal.PlanError.exception/1","doc":"Create an `Elixir.Reactor.Error.Internal.PlanError` without raising it.","ref":"Reactor.Error.Internal.PlanError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Internal.PlanError.exception/1","doc":"- :graph\n- :message\n- :reactor\n- :step","ref":"Reactor.Error.Internal.PlanError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Internal.UnreachableError","doc":"An error that should never happen.","ref":"Reactor.Error.Internal.UnreachableError.html"},{"type":"function","title":"Reactor.Error.Internal.UnreachableError.exception/1","doc":"Create an `Elixir.Reactor.Error.Internal.UnreachableError` without raising it.","ref":"Reactor.Error.Internal.UnreachableError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Internal.UnreachableError.exception/1","doc":"- :bindings\n- :message\n- :file\n- :line","ref":"Reactor.Error.Internal.UnreachableError.html#exception/1-keys"},{"type":"macro","title":"Reactor.Error.Internal.UnreachableError.unreachable/1","doc":"Create an unreachable error.","ref":"Reactor.Error.Internal.UnreachableError.html#unreachable/1"},{"type":"macro","title":"Reactor.Error.Internal.UnreachableError.unreachable!/1","doc":"Bang version of `unreachable/1`.","ref":"Reactor.Error.Internal.UnreachableError.html#unreachable!/1"},{"type":"exception","title":"Reactor.Error.Invalid","doc":"The [Splode error class](e:splode:get-started-with-splode.html#error-classes)\nfor user-caused errors.","ref":"Reactor.Error.Invalid.html"},{"type":"function","title":"Reactor.Error.Invalid.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid` without raising it.","ref":"Reactor.Error.Invalid.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.exception/1","doc":"- :errors","ref":"Reactor.Error.Invalid.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.ArgumentSubpathError","doc":"This error is returned when an argument cannot have a subpath applied to it.","ref":"Reactor.Error.Invalid.ArgumentSubpathError.html"},{"type":"function","title":"Reactor.Error.Invalid.ArgumentSubpathError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.ArgumentSubpathError` without raising it.","ref":"Reactor.Error.Invalid.ArgumentSubpathError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.ArgumentSubpathError.exception/1","doc":"- :argument\n- :culprit\n- :culprit_key\n- :culprit_path\n- :message\n- :step\n- :value","ref":"Reactor.Error.Invalid.ArgumentSubpathError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.CompensateStepError","doc":"This error is returned when an error occurs during step compensation.\n\nIts `error` key will contain the error that was raised or returned by the\n`c:Step.compensate/4` callback.","ref":"Reactor.Error.Invalid.CompensateStepError.html"},{"type":"function","title":"Reactor.Error.Invalid.CompensateStepError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.CompensateStepError` without raising it.","ref":"Reactor.Error.Invalid.CompensateStepError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.CompensateStepError.exception/1","doc":"- :error\n- :reactor\n- :step","ref":"Reactor.Error.Invalid.CompensateStepError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.ForcedFailureError","doc":"This error is returned when the `flunk` DSL entity or the `Reactor.Step.Fail`\nstep are called.","ref":"Reactor.Error.Invalid.ForcedFailureError.html"},{"type":"function","title":"Reactor.Error.Invalid.ForcedFailureError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.ForcedFailureError` without raising it.","ref":"Reactor.Error.Invalid.ForcedFailureError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.ForcedFailureError.exception/1","doc":"- :arguments\n- :step_name\n- :message\n- :context\n- :options","ref":"Reactor.Error.Invalid.ForcedFailureError.html#exception/1-keys"},{"type":"type","title":"Reactor.Error.Invalid.ForcedFailureError.t/0","doc":"","ref":"Reactor.Error.Invalid.ForcedFailureError.html#t:t/0"},{"type":"exception","title":"Reactor.Error.Invalid.InvalidResultError","doc":"This error is returned when a step returns an invalid result.","ref":"Reactor.Error.Invalid.InvalidResultError.html"},{"type":"function","title":"Reactor.Error.Invalid.InvalidResultError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.InvalidResultError` without raising it.","ref":"Reactor.Error.Invalid.InvalidResultError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.InvalidResultError.exception/1","doc":"- :reactor\n- :step\n- :result\n- :arguments","ref":"Reactor.Error.Invalid.InvalidResultError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.MissingArgumentError","doc":"This error is returned when an expected argument is not passed to a step.","ref":"Reactor.Error.Invalid.MissingArgumentError.html"},{"type":"function","title":"Reactor.Error.Invalid.MissingArgumentError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.MissingArgumentError` without raising it.","ref":"Reactor.Error.Invalid.MissingArgumentError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.MissingArgumentError.exception/1","doc":"- :argument\n- :arguments\n- :step","ref":"Reactor.Error.Invalid.MissingArgumentError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.MissingInputError","doc":"Error raised when a required Reactor input is missing.","ref":"Reactor.Error.Invalid.MissingInputError.html"},{"type":"function","title":"Reactor.Error.Invalid.MissingInputError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.MissingInputError` without raising it.","ref":"Reactor.Error.Invalid.MissingInputError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.MissingInputError.exception/1","doc":"- :argument\n- :reactor\n- :step","ref":"Reactor.Error.Invalid.MissingInputError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.MissingResultError","doc":"This error is returned when a step attempts to consume an intermediate result\nwhich is not present in the Reactor state.","ref":"Reactor.Error.Invalid.MissingResultError.html"},{"type":"function","title":"Reactor.Error.Invalid.MissingResultError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.MissingResultError` without raising it.","ref":"Reactor.Error.Invalid.MissingResultError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.MissingResultError.exception/1","doc":"- :argument\n- :reactor\n- :step","ref":"Reactor.Error.Invalid.MissingResultError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.RetriesExceededError","doc":"This error is returned when a step attempts to retry more times that is\nallowed.","ref":"Reactor.Error.Invalid.RetriesExceededError.html"},{"type":"function","title":"Reactor.Error.Invalid.RetriesExceededError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.RetriesExceededError` without raising it.","ref":"Reactor.Error.Invalid.RetriesExceededError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.RetriesExceededError.exception/1","doc":"- :retry_count\n- :step","ref":"Reactor.Error.Invalid.RetriesExceededError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.RunStepError","doc":"This error is returned when an error occurs during step execution.\n\nIts `error` key will contain the error that was raised or returned by the\n`c:Step.run/3` callback.","ref":"Reactor.Error.Invalid.RunStepError.html"},{"type":"function","title":"Reactor.Error.Invalid.RunStepError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.RunStepError` without raising it.","ref":"Reactor.Error.Invalid.RunStepError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.RunStepError.exception/1","doc":"- :error\n- :step","ref":"Reactor.Error.Invalid.RunStepError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.TransformError","doc":"An error which occurs when building and running transforms.","ref":"Reactor.Error.Invalid.TransformError.html"},{"type":"function","title":"Reactor.Error.Invalid.TransformError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.TransformError` without raising it.","ref":"Reactor.Error.Invalid.TransformError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.TransformError.exception/1","doc":"- :input\n- :output\n- :error","ref":"Reactor.Error.Invalid.TransformError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.UndoRetriesExceededError","doc":"An error used when a step runs out of retry events and no other error is\nthrown.","ref":"Reactor.Error.Invalid.UndoRetriesExceededError.html"},{"type":"function","title":"Reactor.Error.Invalid.UndoRetriesExceededError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.UndoRetriesExceededError` without raising it.","ref":"Reactor.Error.Invalid.UndoRetriesExceededError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.UndoRetriesExceededError.exception/1","doc":"- :step\n- :retry_count","ref":"Reactor.Error.Invalid.UndoRetriesExceededError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Invalid.UndoStepError","doc":"This error is returned when an error occurs when attempting to undo step execution.\n\nIts `error` key will contain the error that was raised or returned by the\n`c:Step.undo/4` callback.","ref":"Reactor.Error.Invalid.UndoStepError.html"},{"type":"function","title":"Reactor.Error.Invalid.UndoStepError.exception/1","doc":"Create an `Elixir.Reactor.Error.Invalid.UndoStepError` without raising it.","ref":"Reactor.Error.Invalid.UndoStepError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Invalid.UndoStepError.exception/1","doc":"- :step\n- :error","ref":"Reactor.Error.Invalid.UndoStepError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Unknown","doc":"The [Splode error class](e:splode:get-started-with-splode.html#error-classes)\nfor unknown errors.","ref":"Reactor.Error.Unknown.html"},{"type":"function","title":"Reactor.Error.Unknown.exception/1","doc":"Create an `Elixir.Reactor.Error.Unknown` without raising it.","ref":"Reactor.Error.Unknown.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Unknown.exception/1","doc":"- :errors","ref":"Reactor.Error.Unknown.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Unknown.UnknownError","doc":"An error used to wrap unknown errors.","ref":"Reactor.Error.Unknown.UnknownError.html"},{"type":"function","title":"Reactor.Error.Unknown.UnknownError.exception/1","doc":"Create an `Elixir.Reactor.Error.Unknown.UnknownError` without raising it.","ref":"Reactor.Error.Unknown.UnknownError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Unknown.UnknownError.exception/1","doc":"- :error","ref":"Reactor.Error.Unknown.UnknownError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Validation","doc":"The [Splode error class](e:splode:get-started-with-splode.html#error-classes)\nfor validation errors.","ref":"Reactor.Error.Validation.html"},{"type":"function","title":"Reactor.Error.Validation.exception/1","doc":"Create an `Elixir.Reactor.Error.Validation` without raising it.","ref":"Reactor.Error.Validation.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Validation.exception/1","doc":"- :errors","ref":"Reactor.Error.Validation.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Validation.MissingReturnError","doc":"An error returned when a Reactor cannot be validated because of a missing\nreturn value.","ref":"Reactor.Error.Validation.MissingReturnError.html"},{"type":"function","title":"Reactor.Error.Validation.MissingReturnError.exception/1","doc":"Create an `Elixir.Reactor.Error.Validation.MissingReturnError` without raising it.","ref":"Reactor.Error.Validation.MissingReturnError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Validation.MissingReturnError.exception/1","doc":"- :reactor","ref":"Reactor.Error.Validation.MissingReturnError.html#exception/1-keys"},{"type":"exception","title":"Reactor.Error.Validation.StateError","doc":"An error returned when a Reactor is in an unexpected state.","ref":"Reactor.Error.Validation.StateError.html"},{"type":"function","title":"Reactor.Error.Validation.StateError.exception/1","doc":"Create an `Elixir.Reactor.Error.Validation.StateError` without raising it.","ref":"Reactor.Error.Validation.StateError.html#exception/1"},{"type":"function","title":"Keys - Reactor.Error.Validation.StateError.exception/1","doc":"- :reactor\n- :state\n- :expected","ref":"Reactor.Error.Validation.StateError.html#exception/1-keys"},{"type":"module","title":"Reactor.Builder","doc":"Build a new Reactor programmatically.\n\nYou don't _have_ to use the Reactor DSL to create a Reactor.  The functions in\nthis module allow you to define a Reactor programmatically.  This is\nespecially useful if you need to create a reactor dynamically (maybe based on\na UI such as [React Flow](https://reactflow.dev/)).","ref":"Reactor.Builder.html"},{"type":"module","title":"Example - Reactor.Builder","doc":"```elixir\nreactor = Builder.new()\n{:ok, reactor} = Builder.add_input(reactor, :name)\nargument = Argument.from_input(:name)\n{:ok, reactor} = Builder.add_step(reactor, :greet, [argument])\n{:ok, reactor} = Builder.return(reactor, :greet)\n```","ref":"Reactor.Builder.html#module-example"},{"type":"function","title":"Reactor.Builder.add_input/3","doc":"Add a named input to the Reactor.\n\nThis both places the input in the Reactor for later input validation and adds\nsteps to the Reactor which will emit and (possibly) transform the input.","ref":"Reactor.Builder.html#add_input/3"},{"type":"function","title":"Reactor.Builder.add_input!/3","doc":"Raising version of `add_input/2..3`.","ref":"Reactor.Builder.html#add_input!/3"},{"type":"function","title":"Reactor.Builder.add_middleware/2","doc":"Add a middleware to the Reactor.\n\nReturns an error if the middleware is already present on the Reactor.","ref":"Reactor.Builder.html#add_middleware/2"},{"type":"function","title":"Reactor.Builder.add_middleware!/2","doc":"Raising version of `add_middleware/2`.","ref":"Reactor.Builder.html#add_middleware!/2"},{"type":"function","title":"Reactor.Builder.add_step/5","doc":"Add a step to the Reactor.\n\nAdd a new step to the Reactor.  Rewrites input arguments to use the result of\nthe input steps and injects transformation steps as required.","ref":"Reactor.Builder.html#add_step/5"},{"type":"function","title":"Reactor.Builder.add_step!/5","doc":"Raising version of `add_step/3..5`.","ref":"Reactor.Builder.html#add_step!/5"},{"type":"type","title":"Reactor.Builder.arguments_transform/0","doc":"Optionally transform all the arguments into new arguments","ref":"Reactor.Builder.html#t:arguments_transform/0"},{"type":"type","title":"Reactor.Builder.async?/0","doc":"Should the step be run asynchronously?","ref":"Reactor.Builder.html#t:async?/0"},{"type":"function","title":"Reactor.Builder.compose/5","doc":"Compose another Reactor inside this one.\n\nWhenever possible this function will extract the steps from inner Reactor and\nplace them inside the parent Reactor.  In order to achieve this the composer\nwill rename the steps to ensure that there are no conflicts.\n\nIf you're attempting to create a recursive Reactor (ie compose a Reactor\nwithin itself) then this will be detected and runtime composition will be used\ninstead.  See `Reactor.Step.Compose` for more details.","ref":"Reactor.Builder.html#compose/5"},{"type":"function","title":"Reactor.Builder.compose!/5","doc":"Raising version of `compose/4`.","ref":"Reactor.Builder.html#compose!/5"},{"type":"type","title":"Reactor.Builder.context/0","doc":"Optional context which will be merged with the reactor context when calling this step.","ref":"Reactor.Builder.html#t:context/0"},{"type":"type","title":"Reactor.Builder.description/0","doc":"An optional step description","ref":"Reactor.Builder.html#t:description/0"},{"type":"function","title":"Reactor.Builder.ensure_middleware/2","doc":"Ensure that a middleware is present on the Reactor.","ref":"Reactor.Builder.html#ensure_middleware/2"},{"type":"function","title":"Reactor.Builder.ensure_middleware!/2","doc":"Raising version of `ensure_middleware/2`.","ref":"Reactor.Builder.html#ensure_middleware!/2"},{"type":"type","title":"Reactor.Builder.guards/0","doc":"","ref":"Reactor.Builder.html#t:guards/0"},{"type":"type","title":"Reactor.Builder.impl/0","doc":"","ref":"Reactor.Builder.html#t:impl/0"},{"type":"type","title":"Reactor.Builder.max_retries/0","doc":"How many times is the step allowed to retry?","ref":"Reactor.Builder.html#t:max_retries/0"},{"type":"function","title":"Reactor.Builder.new/1","doc":"Build a new, empty Reactor.\n\nOptionally an identifier for the Reactor. This is primarily used for recursive\ncomposition tracking.","ref":"Reactor.Builder.html#new/1"},{"type":"function","title":"Reactor.Builder.new_step/4","doc":"Build a step which can be added to a reactor at runtime.\n\nNote that the built step doesn't support transformations - you should add an\nadditional step to do the transformation needed (this is what `add_step/5`\ndoes anyway).","ref":"Reactor.Builder.html#new_step/4"},{"type":"function","title":"Reactor.Builder.new_step!/4","doc":"Raising version of `new_step/2..4`.","ref":"Reactor.Builder.html#new_step!/4"},{"type":"function","title":"Reactor.Builder.recurse/5","doc":"Recurse a Reactor until an exit condition is met or maximum iterations are reached.\n\nRecursion takes the output of one execution of the reactor and feeds it as input\nto the next execution, continuing until either:\n1. The exit_condition function returns true when applied to the latest result\n2. The maximum number of iterations is reached\n3. An error occurs during execution\n\nThis provides a powerful way to implement iterative algorithms where the \nprocessing continues until a convergence condition is met.","ref":"Reactor.Builder.html#recurse/5"},{"type":"function","title":"Reactor.Builder.recurse!/5","doc":"Raising version of `recurse/4`.","ref":"Reactor.Builder.html#recurse!/5"},{"type":"type","title":"Reactor.Builder.ref/0","doc":"","ref":"Reactor.Builder.html#t:ref/0"},{"type":"function","title":"Reactor.Builder.return/2","doc":"Specify the return value of the Reactor.\n\nThe return value must be the name of a step.","ref":"Reactor.Builder.html#return/2"},{"type":"function","title":"Reactor.Builder.return!/2","doc":"Raising version of `return/2`.","ref":"Reactor.Builder.html#return!/2"},{"type":"type","title":"Reactor.Builder.step_argument/0","doc":"","ref":"Reactor.Builder.html#t:step_argument/0"},{"type":"type","title":"Reactor.Builder.step_options/0","doc":"","ref":"Reactor.Builder.html#t:step_options/0"},{"type":"module","title":"Reactor.Builder.Compose","doc":"Handle composition of Reactors for the builder.\n\nThe composition logic was getting complicated enough that it seemed sensible\nto extract it from the builder - if only to aid readability.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.compose/4`.","ref":"Reactor.Builder.Compose.html"},{"type":"function","title":"Reactor.Builder.Compose.compose/5","doc":"Compose another Reactor inside this one.","ref":"Reactor.Builder.Compose.html#compose/5"},{"type":"function","title":"Options - Reactor.Builder.Compose.compose/5","doc":"* `:guards` (list of value that implements the `Reactor.Guard.Build` protocol) - Any guards which need to be added to the generated step The default value is `[]`.\n\n* `:async?` (`t:boolean/0`) - Whether the nested Reactor is allowed to run async or not The default value is `true`.","ref":"Reactor.Builder.Compose.html#compose/5-options"},{"type":"function","title":"Reactor.Builder.Compose.verify_arguments/2","doc":"Verify that the arguments and reactor inputs match","ref":"Reactor.Builder.Compose.html#verify_arguments/2"},{"type":"module","title":"Reactor.Builder.Input","doc":"Handle adding inputs to Reactors for the builder.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.add_input/3`.","ref":"Reactor.Builder.Input.html"},{"type":"function","title":"Reactor.Builder.Input.add_input/3","doc":"Add a named input to the reactor.","ref":"Reactor.Builder.Input.html#add_input/3"},{"type":"type","title":"Reactor.Builder.Input.options/0","doc":"","ref":"Reactor.Builder.Input.html#t:options/0"},{"type":"type","title":"Reactor.Builder.Input.transform/0","doc":"","ref":"Reactor.Builder.Input.html#t:transform/0"},{"type":"module","title":"Reactor.Builder.Recurse","doc":"Handle recursive execution of Reactors for the builder.\n\nThis module provides functionality to recursively execute a reactor,\nwith the output of each iteration becoming the input of the next one.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.recurse/4`.","ref":"Reactor.Builder.Recurse.html"},{"type":"function","title":"Reactor.Builder.Recurse.recurse/5","doc":"Execute a Reactor recursively.","ref":"Reactor.Builder.Recurse.html#recurse/5"},{"type":"function","title":"Options - Reactor.Builder.Recurse.recurse/5","doc":"* `:guards` (list of value that implements the `Reactor.Guard.Build` protocol) - Any guards which need to be added to the generated step The default value is `[]`.\n\n* `:async?` (`t:boolean/0`) - Whether the nested Reactor is allowed to run async or not The default value is `true`.\n\n* `:max_iterations` (`t:pos_integer/0`) - Maximum number of iterations to execute\n\n* `:exit_condition` - Function to determine when to stop iteration","ref":"Reactor.Builder.Recurse.html#recurse/5-options"},{"type":"function","title":"Reactor.Builder.Recurse.verify_io_compatibility/1","doc":"Verify that the reactor has compatible inputs and outputs for recursion.\n\nFor recursion to work properly, the output type of the reactor must match\nits input type, so that the result of one iteration can be fed into the next.\n\nThis function checks:\n1. The reactor has a return value defined\n2. The output structure (as a map) contains all keys that are required as inputs","ref":"Reactor.Builder.Recurse.html#verify_io_compatibility/1"},{"type":"module","title":"Reactor.Builder.Step","doc":"Handle building and adding steps to Reactors for the builder.\n\nYou should not use this module directly, but instead use\n`Reactor.Builder.new_step/4` and `Reactor.Builder.add_step/5`.","ref":"Reactor.Builder.Step.html"},{"type":"function","title":"Reactor.Builder.Step.add_step/5","doc":"Build and add a new step to a Reactor.","ref":"Reactor.Builder.Step.html#add_step/5"},{"type":"function","title":"Arguments - Reactor.Builder.Step.add_step/5","doc":"* `reactor` - An existing Reactor struct to add the step to.\n* `name` - The proposed name of the new step.\n* `impl` - A module implementing the `Reactor.Step` behaviour (or a tuple\n  containing the module and options).\n* `arguments` - A list of `Reactor.Argument` structs or shorthand keyword\n  lists.","ref":"Reactor.Builder.Step.html#add_step/5-arguments"},{"type":"function","title":"Options - Reactor.Builder.Step.add_step/5","doc":"* `:async?` - Allow the step to be run asynchronously? The default value is `true`.\n\n* `:max_retries` - The maximum number of times the step can ask to be retried The default value is `100`.\n\n* `:transform` - A function which can modify all incoming arguments\n\n* `:context` (`t:map/0`) - Context which will be merged with the reactor context when calling this step The default value is `%{}`.\n\n* `:description` - An optional description for the step\n\n* `:guards` (list of value that implements the `Reactor.Guard.Build` protocol) - Optional guards to add to the step The default value is `[]`.\n\n* `:ref` - What sort of step reference to generate Valid values are :step_name, :make_ref The default value is `:make_ref`.","ref":"Reactor.Builder.Step.html#add_step/5-options"},{"type":"function","title":"Reactor.Builder.Step.new_step/4","doc":"Dynamically build a new step for later use.\n\nYou're most likely to use this when dynamically returning new steps from an\nexisting step.","ref":"Reactor.Builder.Step.html#new_step/4"},{"type":"function","title":"Arguments - Reactor.Builder.Step.new_step/4","doc":"* `name` - The name of the new step.\n* `impl` - A module implementing the `Reactor.Step` behaviour (or a tuple\n  containing the module and options).\n* `arguments` - A list of `Reactor.Argument` structs or shorthand keyword\n  lists.","ref":"Reactor.Builder.Step.html#new_step/4-arguments"},{"type":"function","title":"Options - Reactor.Builder.Step.new_step/4","doc":"* `:async?` - Allow the step to be run asynchronously? The default value is `true`.\n\n* `:max_retries` - The maximum number of times the step can ask to be retried The default value is `100`.\n\n* `:transform` - A function which can modify all incoming arguments\n\n* `:context` (`t:map/0`) - Context which will be merged with the reactor context when calling this step The default value is `%{}`.\n\n* `:description` - An optional description for the step\n\n* `:guards` (list of value that implements the `Reactor.Guard.Build` protocol) - Optional guards to add to the step The default value is `[]`.\n\n* `:ref` - What sort of step reference to generate Valid values are :step_name, :make_ref The default value is `:make_ref`.","ref":"Reactor.Builder.Step.html#new_step/4-options"},{"type":"module","title":"Reactor.Argument","doc":"A step argument.","ref":"Reactor.Argument.html"},{"type":"function","title":"Reactor.Argument.from_element/3","doc":"Build an argument which refers to to an element within a map step with an optional transformation applied.","ref":"Reactor.Argument.html#from_element/3"},{"type":"function","title":"Example - Reactor.Argument.from_element/3","doc":"iex> Argument.from_element(:argument_name, &Atom.to_string/1)","ref":"Reactor.Argument.html#from_element/3-example"},{"type":"function","title":"Reactor.Argument.from_input/3","doc":"Build an argument which refers to a reactor input with an optional\ntransformation applied.","ref":"Reactor.Argument.html#from_input/3"},{"type":"function","title":"Options - Reactor.Argument.from_input/3","doc":"* `:transform` - An optional transformation function which can be used to modify the argument before it is passed to the step. The default value is `nil`.\n\n* `:description` - An optional description for the argument. The default value is `nil`.","ref":"Reactor.Argument.html#from_input/3-options"},{"type":"function","title":"Example - Reactor.Argument.from_input/3","doc":"iex> Argument.from_input(:argument_name, :input_name, transform: &String.to_integer/1)","ref":"Reactor.Argument.html#from_input/3-example"},{"type":"function","title":"Reactor.Argument.from_result/3","doc":"Build an argument which refers to the result of another step with an optional\ntransformation applied.","ref":"Reactor.Argument.html#from_result/3"},{"type":"function","title":"Example - Reactor.Argument.from_result/3","doc":"iex> Argument.from_result(:argument_name, :step_name, &Atom.to_string/1)","ref":"Reactor.Argument.html#from_result/3-example"},{"type":"function","title":"Reactor.Argument.from_template/3","doc":"Build an argument directly from a template.","ref":"Reactor.Argument.html#from_template/3"},{"type":"function","title":"Example - Reactor.Argument.from_template/3","doc":"iex> Argument.from_template(:argument_name, Reactor.Dsl.Argument.input(:input_name))","ref":"Reactor.Argument.html#from_template/3-example"},{"type":"function","title":"Reactor.Argument.from_value/3","doc":"Build an argument which refers to a statically defined value.","ref":"Reactor.Argument.html#from_value/3"},{"type":"function","title":"Example - Reactor.Argument.from_value/3","doc":"iex> Argument.from_value(:argument_name, 10)","ref":"Reactor.Argument.html#from_value/3-example"},{"type":"macro","title":"Reactor.Argument.has_sub_path/1","doc":"Validate that the argument source has a sub_path","ref":"Reactor.Argument.html#has_sub_path/1"},{"type":"macro","title":"Reactor.Argument.has_transform/1","doc":"Validate that the argument has a transform.","ref":"Reactor.Argument.html#has_transform/1"},{"type":"macro","title":"Reactor.Argument.is_argument/1","doc":"Validate that the argument is an Argument struct.","ref":"Reactor.Argument.html#is_argument/1"},{"type":"macro","title":"Reactor.Argument.is_from_element/1","doc":"Validate that the argument contains an element.","ref":"Reactor.Argument.html#is_from_element/1"},{"type":"macro","title":"Reactor.Argument.is_from_input/1","doc":"Validate that the argument refers to a reactor input.","ref":"Reactor.Argument.html#is_from_input/1"},{"type":"macro","title":"Reactor.Argument.is_from_result/1","doc":"Validate that the argument refers to a step result.","ref":"Reactor.Argument.html#is_from_result/1"},{"type":"macro","title":"Reactor.Argument.is_from_value/1","doc":"Validate that the argument contains a static value.","ref":"Reactor.Argument.html#is_from_value/1"},{"type":"type","title":"Reactor.Argument.options/0","doc":"","ref":"Reactor.Argument.html#t:options/0"},{"type":"function","title":"Reactor.Argument.sub_path/2","doc":"Set a sub-path on the argument.","ref":"Reactor.Argument.html#sub_path/2"},{"type":"function","title":"Example - Reactor.Argument.sub_path/2","doc":"iex> Argument.from_value(:example, :value)\n    ...> |> Argument.sub_path([:nested, :values])","ref":"Reactor.Argument.html#sub_path/2-example"},{"type":"type","title":"Reactor.Argument.t/0","doc":"","ref":"Reactor.Argument.html#t:t/0"},{"type":"type","title":"Reactor.Argument.transform/0","doc":"","ref":"Reactor.Argument.html#t:transform/0"},{"type":"protocol","title":"Reactor.Argument.Build","doc":"A protocol which can be used to convert something into an argument.","ref":"Reactor.Argument.Build.html"},{"type":"function","title":"Reactor.Argument.Build.build/1","doc":"Convert the input into one or more arguments.","ref":"Reactor.Argument.Build.html#build/1"},{"type":"type","title":"Reactor.Argument.Build.t/0","doc":"All the types that implement this protocol.","ref":"Reactor.Argument.Build.html#t:t/0"},{"type":"module","title":"Reactor.Executor","doc":"The Reactor executor.\n\nThe executor handles the main loop of running a Reactor.\n\nThe algorithm is somewhat confusing, so here it is in pseudocode:\n\n1. Find any async tasks (from a previous loop) which are completed. Either\n   recurse or continue if none are found.\n2. Find any async steps in the plan which are ready to run (they have no\n   in-edges in the graph) and start as many as possible (given the constraints\n   of `max_concurrency` and the state of the concurrency pool).  Either start\n   over, or continue if none are found.\n3. Find a single synchronous step which is ready to run and execute it. If\n   there was one then recurse, otherwise continue.\n4. Check if there are no more steps left in the plan (there are zero\n   vertices).  If so, collect the return value and exit, otherwise recurse.\n\nWhenever a step is run, whether run synchronously or asynchronously, the\nfollowing happens:\n\n1. When the step is successful:\n  a. If the step is undoable (ie `Step.can?(module, :undo)?` returns `true`)\n     then the step and the result are stored in the Reactor's undo stack.\n  b. If the result is depended upon by another step (the graph has out-edges\n     for the step) _or_ the step is asking the reactor to halt then the\n     result is stored in the Reactor's intermediate results.\n  c. The step is removed from the graph (along with it's out-edges, freeing\n     up it's dependents to run).\n2. When the step is unsuccessful (returns an error tuple or raises):\n  a. If the step can be compensated then compensation is attempted up to five\n     times before giving up.\n  b. The reactor iterates it's undo stack calling undo on each step.\n3. When a step or compensation asks for a retry then the step is placed back\n   in the graph to be run again next iteration.","ref":"Reactor.Executor.html"},{"type":"function","title":"Reactor.Executor.run/4","doc":"Run a reactor.\n\nProvided a Reactor which has been planned and the correct inputs, then run\nthe Reactor until completion, halting or failure.\n\nYou probably shouldn't call this directly, but use `Reactor.run/4` instead.","ref":"Reactor.Executor.html#run/4"},{"type":"module","title":"Reactor.Executor.Async","doc":"Handle the asynchronous execution of a batch of steps, along with any\nmutations to the reactor or execution state.","ref":"Reactor.Executor.Async.html"},{"type":"function","title":"Reactor.Executor.Async.collect_remaining_tasks_for_shutdown/2","doc":"When the Reactor needs to shut down for any reason, we need to await all the\ncurrently running asynchronous steps and delete any task vertices.","ref":"Reactor.Executor.Async.html#collect_remaining_tasks_for_shutdown/2"},{"type":"function","title":"Reactor.Executor.Async.handle_completed_steps/2","doc":"Handle zero or one completed async steps and then decide what to do.","ref":"Reactor.Executor.Async.html#handle_completed_steps/2"},{"type":"function","title":"Reactor.Executor.Async.start_steps/4","doc":"Start as many of the provided steps as possible.\n\nTakes into account he maximum concurrency and available work slots.","ref":"Reactor.Executor.Async.html#start_steps/4"},{"type":"module","title":"Reactor.Executor.ConcurrencyTracker","doc":"Manage shared concurrency pools for multiple Reactors.\n\nWhen running a Reactor you can pass the `concurrency_key` option, which will\ncause the Reactor to use the specified pool to ensure that the combined\nReactors never exceed the pool's available concurrency limit.\n\nThis avoids nested Reactors spawning too many workers and thrashing the\nsystem.\n\nThe process calling `allocate_pool/1` is monitored, and when it terminates\nit's allocation is removed.  Any processes which are using that pool will\nnot be able to allocate any new resources.","ref":"Reactor.Executor.ConcurrencyTracker.html"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.acquire/2","doc":"Attempt to acquire a number of concurrency allocations from the pool.\n\nReturns `{:ok, n}` where `n` was the number of slots that were actually\nallocated.  It's important to note that whilst you may request `16` slots, if\nthere is only `3` available, then this function will return `{:ok, 3}` and you\nmust abide by it.\n\nIt is possible for this function to return `{:ok, 0}` if there is no slots\navailable.","ref":"Reactor.Executor.ConcurrencyTracker.html#acquire/2"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.allocate_pool/1","doc":"Allocate a new concurrency pool and set the maximum limit.","ref":"Reactor.Executor.ConcurrencyTracker.html#allocate_pool/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Reactor.Executor.ConcurrencyTracker.html#child_spec/1"},{"type":"type","title":"Reactor.Executor.ConcurrencyTracker.pool_key/0","doc":"","ref":"Reactor.Executor.ConcurrencyTracker.html#t:pool_key/0"},{"type":"type","title":"Reactor.Executor.ConcurrencyTracker.record/0","doc":"","ref":"Reactor.Executor.ConcurrencyTracker.html#t:record/0"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.release/2","doc":"Release concurrency allocation back to the pool.","ref":"Reactor.Executor.ConcurrencyTracker.html#release/2"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.release_pool/1","doc":"Release the concurrency pool.\n\nThis deletes the pool, however doesn't affect any processes currently using\nit.  No more resources can be acquired by users of the pool key.","ref":"Reactor.Executor.ConcurrencyTracker.html#release_pool/1"},{"type":"function","title":"Reactor.Executor.ConcurrencyTracker.status/1","doc":"Report the available and maximum concurrency for a pool.","ref":"Reactor.Executor.ConcurrencyTracker.html#status/1"},{"type":"module","title":"Reactor.Executor.Hooks","doc":"Handles the execution of reactor middleware hooks.","ref":"Reactor.Executor.Hooks.html"},{"type":"function","title":"Reactor.Executor.Hooks.complete/3","doc":"Run the completion hooks allowing the result to be replaced","ref":"Reactor.Executor.Hooks.html#complete/3"},{"type":"function","title":"Reactor.Executor.Hooks.error/3","doc":"Run the error hooks allowing the error to be replaced","ref":"Reactor.Executor.Hooks.html#error/3"},{"type":"function","title":"Reactor.Executor.Hooks.event/4","doc":"Emit a step event.","ref":"Reactor.Executor.Hooks.html#event/4"},{"type":"function","title":"Reactor.Executor.Hooks.get_process_contexts/1","doc":"Run any get_process_context hooks","ref":"Reactor.Executor.Hooks.html#get_process_contexts/1"},{"type":"function","title":"Reactor.Executor.Hooks.halt/2","doc":"Run the halt hooks collecting the new context as it goes","ref":"Reactor.Executor.Hooks.html#halt/2"},{"type":"function","title":"Reactor.Executor.Hooks.init/2","doc":"Run the init hooks collecting the new context as it goes","ref":"Reactor.Executor.Hooks.html#init/2"},{"type":"function","title":"Reactor.Executor.Hooks.set_process_contexts/1","doc":"Run set_process_context hooks given the result of `get_process_contexts/1`","ref":"Reactor.Executor.Hooks.html#set_process_contexts/1"},{"type":"module","title":"Reactor.Executor.Init","doc":"Handle argument checking and state setup for a Reactor run.","ref":"Reactor.Executor.Init.html"},{"type":"module","title":"Reactor.Executor.State","doc":"Contains the reactor execution state.\n\nThis is run-time only information.","ref":"Reactor.Executor.State.html"},{"type":"type","title":"Reactor.Executor.State.t/0","doc":"","ref":"Reactor.Executor.State.html#t:t/0"},{"type":"module","title":"Reactor.Executor.StepRunner","doc":"Run an individual step, including compensation if possible.","ref":"Reactor.Executor.StepRunner.html"},{"type":"function","title":"Reactor.Executor.StepRunner.run/4","doc":"Collect the arguments and and run a step, with compensation if required.","ref":"Reactor.Executor.StepRunner.html#run/4"},{"type":"function","title":"Reactor.Executor.StepRunner.run_async/5","doc":"Run a step inside a task.\n\nThis is a simple wrapper around `run/4` except that it emits more events.","ref":"Reactor.Executor.StepRunner.html#run_async/5"},{"type":"function","title":"Reactor.Executor.StepRunner.undo/5","doc":"Undo a step if possible.","ref":"Reactor.Executor.StepRunner.html#undo/5"},{"type":"module","title":"Reactor.Executor.Sync","doc":"Handle the synchronous execution of a single step, along with any mutations to\nthe reactor or execution state.","ref":"Reactor.Executor.Sync.html"},{"type":"function","title":"Reactor.Executor.Sync.run/3","doc":"Try and run a step synchronously.","ref":"Reactor.Executor.Sync.html#run/3"},{"type":"module","title":"Reactor.Guard","doc":"A Reactor guard.\n\nGuard types should implement the `Reactor.Guard.Build` protocol.\n\nThis struct contains a single two arity function, which when called with a\nstep's arguments and context returns one of the following:\n\n- `:cont` - this guard has passed - continue evaluating any additional guards\n  and if exhausted run the step.\n- `{:halt, result}` - the guard has failed - use `result` as the steps result.","ref":"Reactor.Guard.html"},{"type":"type","title":"Reactor.Guard.t/0","doc":"","ref":"Reactor.Guard.html#t:t/0"},{"type":"protocol","title":"Reactor.Guard.Build","doc":"A protocol which can be used to convert something into a guard.","ref":"Reactor.Guard.Build.html"},{"type":"function","title":"Reactor.Guard.Build.build/1","doc":"Convert the input into one or more guards.","ref":"Reactor.Guard.Build.html#build/1"},{"type":"type","title":"Reactor.Guard.Build.t/0","doc":"All the types that implement this protocol.","ref":"Reactor.Guard.Build.html#t:t/0"},{"type":"module","title":"Reactor.Info","doc":"Introspection for the Reactor DSL.","ref":"Reactor.Info.html"},{"type":"function","title":"Reactor.Info.reactor/1","doc":"reactor DSL entities","ref":"Reactor.Info.html#reactor/1"},{"type":"function","title":"Reactor.Info.reactor_middlewares/1","doc":"reactor.middlewares DSL entities","ref":"Reactor.Info.html#reactor_middlewares/1"},{"type":"function","title":"Reactor.Info.reactor_options/1","doc":"reactor DSL options\n\nReturns a map containing the and any configured or default values.","ref":"Reactor.Info.html#reactor_options/1"},{"type":"function","title":"Reactor.Info.reactor_return/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Info.html#reactor_return/1"},{"type":"function","title":"Reactor.Info.reactor_return!/1","doc":"Specify which step result to return upon completion.","ref":"Reactor.Info.html#reactor_return!/1"},{"type":"function","title":"Reactor.Info.to_struct/1","doc":"Convert a reactor DSL module into a reactor struct.","ref":"Reactor.Info.html#to_struct/1"},{"type":"function","title":"Reactor.Info.to_struct!/1","doc":"Raising version of `to_struct/1`.","ref":"Reactor.Info.html#to_struct!/1"},{"type":"behaviour","title":"Reactor.Mermaid","doc":"Converts Reactors and their related entities into a Mermaid diagram.\n\nSee [Mermaid](https://mermaid.js.org/) for more information.","ref":"Reactor.Mermaid.html"},{"type":"type","title":"Reactor.Mermaid.options/0","doc":"","ref":"Reactor.Mermaid.html#t:options/0"},{"type":"function","title":"Reactor.Mermaid.to_mermaid/2","doc":"Convert the Reactor into Mermaid.","ref":"Reactor.Mermaid.html#to_mermaid/2"},{"type":"function","title":"Options - Reactor.Mermaid.to_mermaid/2","doc":"* `:expand?` (`t:boolean/0`) - Whether or not to expand composed Reactors The default value is `false`.\n\n* `:describe?` (`t:boolean/0`) - Whether or not to include descriptions, if available The default value is `false`.\n\n* `:direction` - The direction to render the flowchart Valid values are :top_to_bottom, :bottom_to_top, :right_to_left, :left_to_right The default value is `:left_to_right`.\n\n* `:indent` (`t:non_neg_integer/0`) - How much to indent the resulting mermaid The default value is `0`.\n\n* `:output` - Specify the output format. `iodata` is more performant Valid values are :iodata, :binary The default value is `:iodata`.","ref":"Reactor.Mermaid.html#to_mermaid/2-options"},{"type":"callback","title":"Reactor.Mermaid.to_mermaid/2","doc":"","ref":"Reactor.Mermaid.html#c:to_mermaid/2"},{"type":"function","title":"Reactor.Mermaid.to_mermaid!/2","doc":"Convert the Reactor into Mermaid\n\nRaising version of `to_mermaid/2`","ref":"Reactor.Mermaid.html#to_mermaid!/2"},{"type":"module","title":"Reactor.Mermaid.Node","doc":"A node in the mermaid graph","ref":"Reactor.Mermaid.Node.html"},{"type":"type","title":"Reactor.Mermaid.Node.node_data/0","doc":"","ref":"Reactor.Mermaid.Node.html#t:node_data/0"},{"type":"type","title":"Reactor.Mermaid.Node.node_list/0","doc":"","ref":"Reactor.Mermaid.Node.html#t:node_list/0"},{"type":"function","title":"Reactor.Mermaid.Node.render/2","doc":"Render the node at the provided indent level.","ref":"Reactor.Mermaid.Node.html#render/2"},{"type":"type","title":"Reactor.Mermaid.Node.t/0","doc":"","ref":"Reactor.Mermaid.Node.html#t:t/0"},{"type":"module","title":"Reactor.Mermaid.Utils","doc":"Utilities for generating Mermaid.","ref":"Reactor.Mermaid.Utils.html"},{"type":"function","title":"Reactor.Mermaid.Utils.direction/1","doc":"Convert the direction atom into a mermaid direction","ref":"Reactor.Mermaid.Utils.html#direction/1"},{"type":"function","title":"Reactor.Mermaid.Utils.md_escape/1","doc":"Escape markdown as needed","ref":"Reactor.Mermaid.Utils.html#md_escape/1"},{"type":"function","title":"Reactor.Mermaid.Utils.mermaid_id/2","doc":"Generate a mermaid ID for a term","ref":"Reactor.Mermaid.Utils.html#mermaid_id/2"},{"type":"function","title":"Reactor.Mermaid.Utils.name/1","doc":"Generate a name which can be used within a Mermaid node","ref":"Reactor.Mermaid.Utils.html#name/1"},{"type":"module","title":"Reactor.Planner","doc":"Build an execution plan for a Reactor.\n\nConverts any unplanned steps into vertices in a graph with directed edges\nbetween them representing their dependencies (arguments).","ref":"Reactor.Planner.html"},{"type":"function","title":"Reactor.Planner.plan/1","doc":"Build an execution plan for a Reactor.\n\nBuilds a graph of the step dependencies, resolves them and then builds an execution plan.","ref":"Reactor.Planner.html#plan/1"},{"type":"function","title":"Reactor.Planner.plan!/1","doc":"Raising version of `plan/1`.","ref":"Reactor.Planner.html#plan!/1"},{"type":"module","title":"Reactor.Template","doc":"Templates used to refer to some sort of computed value.","ref":"Reactor.Template.html"},{"type":"macro","title":"Reactor.Template.is_element_template/1","doc":"A guard for element templates","ref":"Reactor.Template.html#is_element_template/1"},{"type":"macro","title":"Reactor.Template.is_input_template/1","doc":"A guard for input templates","ref":"Reactor.Template.html#is_input_template/1"},{"type":"macro","title":"Reactor.Template.is_result_template/1","doc":"A guard for result templates","ref":"Reactor.Template.html#is_result_template/1"},{"type":"macro","title":"Reactor.Template.is_template/1","doc":"A guard to detect all template types","ref":"Reactor.Template.html#is_template/1"},{"type":"macro","title":"Reactor.Template.is_value_template/1","doc":"A guard for value templates","ref":"Reactor.Template.html#is_value_template/1"},{"type":"type","title":"Reactor.Template.t/0","doc":"","ref":"Reactor.Template.html#t:t/0"},{"type":"function","title":"Reactor.Template.type/0","doc":"The type for use in option schemas","ref":"Reactor.Template.html#type/0"},{"type":"module","title":"Reactor.Template.Element","doc":"The `element` template.","ref":"Reactor.Template.Element.html"},{"type":"type","title":"Reactor.Template.Element.t/0","doc":"","ref":"Reactor.Template.Element.html#t:t/0"},{"type":"module","title":"Reactor.Template.Input","doc":"The `input` template.","ref":"Reactor.Template.Input.html"},{"type":"type","title":"Reactor.Template.Input.t/0","doc":"","ref":"Reactor.Template.Input.html#t:t/0"},{"type":"module","title":"Reactor.Template.Result","doc":"The `result` template.","ref":"Reactor.Template.Result.html"},{"type":"type","title":"Reactor.Template.Result.t/0","doc":"","ref":"Reactor.Template.Result.html#t:t/0"},{"type":"module","title":"Reactor.Template.Value","doc":"A statically `value` template.","ref":"Reactor.Template.Value.html"},{"type":"type","title":"Reactor.Template.Value.t/0","doc":"","ref":"Reactor.Template.Value.html#t:t/0"},{"type":"extras","title":"README","doc":"# Reactor\n\n![Elixir CI](https://github.com/ash-project/reactor/actions/workflows/elixir.yml/badge.svg)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Hex version badge](https://img.shields.io/hexpm/v/reactor.svg)](https://hex.pm/packages/reactor)\n\nReactor is a dynamic, concurrent, dependency resolving saga orchestrator.\n\nWoah. That's a lot. Let's break it down:\n\n- **Saga orchestrator** A [saga][saga pattern] is a way of providing\n  transaction-like semantics across multiple distinct resources.\n- **Dependency resolving** reactor allows you to describe the dependencies\n  between your saga steps using _arguments_ which are converted into a\n  [DAG][dag] and used to compute execution order.\n- **Concurrent** unless otherwise specified reactor will run as many steps as\n  possible concurrently whilst taking into account the results of the dependency\n  resolution.\n- **Dynamic** whilst you can define a reactor statically using our awesome DSL,\n  you can also build workflows dynamically - and even add steps while the\n  reactor is running.\n\n[saga pattern](https://learn.microsoft.com/en-us/azure/architecture/reference-architectures/saga/saga)","ref":"readme.html"},{"type":"extras","title":"Sponsors - README","doc":"Thanks to [Alembic Pty Ltd](https://alembic.com.au/) for sponsoring a portion of\nthis project's development.","ref":"readme.html#sponsors"},{"type":"extras","title":"Installation - README","doc":"The package can be installed by adding `reactor` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:reactor, \"~> 0.15.5\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Documentation - README","doc":"Our documentation is organized to help you find exactly what you need:","ref":"readme.html#documentation"},{"type":"extras","title":"🎓 Learning Reactor - Tutorials - README","doc":"Step-by-step guides that teach Reactor through hands-on practice:\n- **[Getting Started](documentation/tutorials/01-getting-started.md)** - Build your first Reactor\n- **[Recursive Execution](documentation/tutorials/05-recursive-execution.md)** - Advanced iterative patterns","ref":"readme.html#learning-reactor-tutorials"},{"type":"extras","title":"🔧 Solving Problems - How-to Guides - README","doc":"Practical solutions for real-world scenarios:\n- **[Payment Processing](documentation/how-to/payment-processing.md)** - E-commerce workflows\n- **[Data Pipelines](documentation/how-to/data-pipelines.md)** - ETL orchestration\n- **[Testing Strategies](documentation/how-to/testing-strategies.md)** - Testing approaches","ref":"readme.html#solving-problems-how-to-guides"},{"type":"extras","title":"📚 API Reference - README","doc":"Complete technical reference:\n- **[DSL Documentation](documentation/dsls/DSL-Reactor.md)** - Complete DSL syntax (auto-generated)\n- **[HexDocs](https://hexdocs.pm/reactor)** - Generated API documentation\n- **[Latest (main branch)](https://ash-project.github.io/reactor)** - Development docs","ref":"readme.html#api-reference"},{"type":"extras","title":"💡 Understanding Reactor - Explanations - README","doc":"Conceptual guides about how and why Reactor works:\n- **[Core Concepts](documentation/explanation/concepts.md)** - Sagas, DAGs, compensation\n- **[Architecture](documentation/explanation/architecture.md)** - Internal design\n- **[Ecosystem](documentation/explanation/ecosystem.md)** - Integration patterns\n\n**Quick Start:** New to Reactor? Start with the [Getting Started tutorial](documentation/tutorials/01-getting-started.md)!","ref":"readme.html#understanding-reactor-explanations"},{"type":"extras","title":"Contributing - README","doc":"- To contribute updates, fixes or new features please fork and open a\n  pull-request against `main`.\n- Please use [conventional\n  commits](https://www.conventionalcommits.org/en/v1.0.0/) - this allows us to\n  dynamically generate the changelog.\n- Feel free to ask any questions on the `#reactor` channel on the [Ash\n  Discord](https://discord.gg/D7FNG2q).","ref":"readme.html#contributing"},{"type":"extras","title":"Licence - README","doc":"`reactor` is licensed under the terms of the [MIT\nlicense](https://opensource.org/licenses/MIT). See the [`LICENSE` file in this\nrepository](https://github.com/ash-project/reactor/blob/main/LICENSE)\nfor details.\n\n[saga pattern]: https://learn.microsoft.com/en-us/azure/architecture/reference-architectures/saga/saga\n[dag]: https://en.wikipedia.org/wiki/Directed_acyclic_graph","ref":"readme.html#licence"},{"type":"extras","title":"Your First Reactor","doc":"# Your First Reactor\n\nWelcome to Reactor! In this tutorial, you'll build your first reactor step by step, learning the core concepts through hands-on practice.","ref":"01-getting-started.html"},{"type":"extras","title":"What you'll build - Your First Reactor","doc":"A user registration workflow that validates email, hashes passwords, and creates user records.","ref":"01-getting-started.html#what-you-ll-build"},{"type":"extras","title":"You'll learn - Your First Reactor","doc":"- Defining reactor inputs and steps\n- Creating dependencies between steps\n- Running reactors and handling results","ref":"01-getting-started.html#you-ll-learn"},{"type":"extras","title":"Workflow Overview - Your First Reactor","doc":"Here's what we'll build - a simple user registration workflow:\n\n```mermaid\ngraph TD\n    A[input: email] --> B[validate_email]\n    A2[input: password] --> C[hash_password]\n    B --> D[create_user]\n    C --> D\n    D --> E[Registration Complete]\n    \n    style A fill:#e1f5fe\n    style A2 fill:#e1f5fe\n    style E fill:#c8e6c9\n    style D fill:#fff3e0\n```","ref":"01-getting-started.html#workflow-overview"},{"type":"extras","title":"Prerequisites - Your First Reactor","doc":"- Basic Elixir knowledge\n- Igniter New installed: `mix archive.install hex igniter_new`","ref":"01-getting-started.html#prerequisites"},{"type":"extras","title":"Step 1: Create a new project with Reactor - Your First Reactor","doc":"Let's start by creating a new project with Reactor already installed and configured:\n\n```bash\nmix igniter.new reactor_tutorial --install reactor\ncd reactor_tutorial\n```\n\nThis creates a new Elixir project and automatically:\n- Adds Reactor to your dependencies\n- Installs the dependencies\n- Sets up any necessary configuration","ref":"01-getting-started.html#step-1-create-a-new-project-with-reactor"},{"type":"extras","title":"Step 2: Your first reactor - Your First Reactor","doc":"Create a new file `lib/user_registration.ex` and define your first reactor:\n\n```elixir\ndefmodule UserRegistration do\n  use Reactor\n\n  # Define what inputs this reactor expects\n  input :email\n  input :password\n\n  # Define a simple step that validates email\n  step :validate_email do\n    argument :email, input(:email)\n    \n    run fn %{email: email}, _context ->\n      if String.contains?(email, \"@\") do\n        {:ok, email}\n      else\n        {:error, \"Email must contain @\"}\n      end\n    end\n  end\n\n  # Define a step that hashes the password\n  step :hash_password do\n    argument :password, input(:password)\n    \n    run fn %{password: password}, _context ->\n      hashed = :crypto.hash(:sha256, password) |> Base.encode16()\n      {:ok, hashed}\n    end\n  end\n\n  # Define a step that creates the user\n  step :create_user do\n    argument :email, result(:validate_email)\n    argument :password_hash, result(:hash_password)\n    \n    run fn %{email: email, password_hash: password_hash}, _context ->\n      user = %{\n        id: :rand.uniform(10000),\n        email: email,\n        password_hash: password_hash,\n        created_at: DateTime.utc_now()\n      }\n      {:ok, user}\n    end\n  end\n\n  # Specify what to return when the reactor completes\n  return :create_user\nend\n```","ref":"01-getting-started.html#step-2-your-first-reactor"},{"type":"extras","title":"Step 3: Understanding your reactor - Your First Reactor","doc":"Before running it, let's understand what you've built:\n\n**Inputs** define the parameters your reactor accepts, like function arguments.\n\n**Steps** are units of work with three key parts:\n- **Arguments** specify what data the step needs and where it comes from\n- **Run function** performs the actual work\n- **Dependencies** are created automatically based on arguments\n\nNotice how `:create_user` depends on results from both `:validate_email` and `:hash_password`. This creates an execution order where the first two steps can run concurrently, then `:create_user` runs after both complete.\n\n**Return** specifies what value the reactor should return when everything succeeds.","ref":"01-getting-started.html#step-3-understanding-your-reactor"},{"type":"extras","title":"Step 4: Run your reactor - Your First Reactor","doc":"Let's test your reactor! Start an IEx session:\n\n```bash\niex -S mix\n```\n\nNow run your reactor:\n\n```elixir\n# Test with valid data\n{:ok, user} = Reactor.run(UserRegistration, %{\n  email: \"alice@example.com\",\n  password: \"secret123\"\n})\n\nIO.inspect(user)\n# Should output something like:\n# %{\n#   id: 1234,\n#   email: \"alice@example.com\", \n#   password_hash: \"2BB80D537B1DA3E38BD30361AA855686BDE0EACD7162FEF6A25FE97BF527A25B\",\n#   created_at: ~U[2024-01-15 10:30:00.123456Z]\n# }\n```\n\nTry with invalid data:\n\n```elixir\n# Test with invalid email\n{:error, reason} = Reactor.run(UserRegistration, %{\n  email: \"not-an-email\",\n  password: \"secret123\" \n})\n\nIO.inspect(reason)\n# Should output: \"Email must contain @\"\n```","ref":"01-getting-started.html#step-4-run-your-reactor"},{"type":"extras","title":"Step 5: Execution order - Your First Reactor","doc":"Here's something important: Reactor doesn't execute steps in the order you wrote them. Instead, it builds a [dependency graph](../reference/glossary.md#dependency-graph) and runs steps as soon as their dependencies are available.\n\nIn your reactor:\n1. `:validate_email` and `:hash_password` can run **simultaneously** because they only depend on inputs\n2. `:create_user` runs **after both** previous steps complete because it needs their results\n\nThis automatic parallelization is one of Reactor's key features - you get concurrent execution without having to think about it!","ref":"01-getting-started.html#step-5-execution-order"},{"type":"extras","title":"What you learned - Your First Reactor","doc":"You've built your first reactor and learned the fundamentals:\n\n- **Inputs** define reactor parameters\n- **Steps** contain arguments, run functions, and automatic dependencies\n- **Arguments** create execution order by specifying data sources\n- **Concurrent execution** happens automatically when steps don't depend on each other\n- **Dependencies** determine when steps run, not the order you write them","ref":"01-getting-started.html#what-you-learned"},{"type":"extras","title":"What's next - Your First Reactor","doc":"Now that you understand the basics, you're ready to learn more advanced concepts:\n\n- **[Error Handling](02-error-handling.md)** - Learn how to handle failures with compensation and undo\n- **[Async Workflows](03-async-workflows.md)** - Explore concurrent processing patterns\n- **[Composition](04-composition.md)** - Build complex workflows with sub-reactors\n\nOr explore specific use cases in our How-to Guides section.\n\nFor quick syntax reference, see the [Reactor Cheatsheet](reactor-cheatsheet.cheatmd).\n\nHappy building with Reactor! 🚀","ref":"01-getting-started.html#what-s-next"},{"type":"extras","title":"Error Handling with Compensation and Undo","doc":"# Error Handling with Compensation and Undo\n\nIn this tutorial, you'll learn how to make your reactors resilient by adding proper error handling, retry logic, and rollback capabilities.","ref":"02-error-handling.html"},{"type":"extras","title":"What you'll build - Error Handling with Compensation and Undo","doc":"You'll enhance the user registration workflow from the first tutorial to handle:\n1. Network failures with automatic retries\n2. Validation errors with graceful failure\n3. Rollback scenarios when later steps fail\n4. Email service failures with compensation","ref":"02-error-handling.html#what-you-ll-build"},{"type":"extras","title":"You'll learn - Error Handling with Compensation and Undo","doc":"- Implementing compensation for retryable errors\n- Adding undo logic for rollback scenarios\n- The difference between compensation and undo\n- Building resilient workflows that handle failures gracefully","ref":"02-error-handling.html#you-ll-learn"},{"type":"extras","title":"Error Handling Flow - Error Handling with Compensation and Undo","doc":"Here's how Reactor handles errors through compensation and undo:\n\n```mermaid\nsequenceDiagram\n    participant Reactor\n    participant StepA\n    participant StepB\n    participant StepC\n    \n    Reactor->>StepA: run()\n    StepA-->>Reactor: {:ok, result}\n    \n    Reactor->>StepB: run()\n    StepB-->>Reactor: {:ok, result}\n    \n    Reactor->>StepC: run()\n    StepC-->>Reactor: {:error, reason}\n    \n    Note over Reactor: Begin compensation\n    Reactor->>StepB: compensate()\n    StepB-->>Reactor: {:continue, context}\n    \n    Reactor->>StepA: compensate()\n    StepA-->>Reactor: {:continue, context}\n    \n    Reactor-->>Reactor: Return compensated error\n```","ref":"02-error-handling.html#error-handling-flow"},{"type":"extras","title":"Prerequisites - Error Handling with Compensation and Undo","doc":"- Complete the [Getting Started tutorial](01-getting-started.md)\n- Basic knowledge of Elixir error handling","ref":"02-error-handling.html#prerequisites"},{"type":"extras","title":"Step 1: Set up the project - Error Handling with Compensation and Undo","doc":"If you don't have the project from the previous tutorial, create it:\n\n```bash\nmix igniter.new reactor_tutorial --install reactor\ncd reactor_tutorial\n```","ref":"02-error-handling.html#step-1-set-up-the-project"},{"type":"extras","title":"Step 2: Understanding Reactor error handling - Error Handling with Compensation and Undo","doc":"Reactor provides two main mechanisms for error handling:","ref":"02-error-handling.html#step-2-understanding-reactor-error-handling"},{"type":"extras","title":"Compensation - Error Handling with Compensation and Undo","doc":"**When**: A step fails during execution  \n**Purpose**: Decide whether to retry, continue, or fail the reactor  \n**Return values**:\n- `:retry` - Try the step again\n- `{:continue, value}` - Continue execution with the provided value\n- `:ok` - Successfully compensated, but still triggers rollback\n- `{:error, reason}` - Fail the entire reactor","ref":"02-error-handling.html#compensation"},{"type":"extras","title":"Undo - Error Handling with Compensation and Undo","doc":"**When**: A step succeeded but a later step failed  \n**Purpose**: Roll back the successful step's changes  \n**Return values**:\n- `:ok` - Successfully undone\n- `{:error, reason}` - Failed to undo (this will fail the reactor)","ref":"02-error-handling.html#undo"},{"type":"extras","title":"Step 3: Create a step with error handling - Error Handling with Compensation and Undo","doc":"Let's create a step that can fail and shows how to handle those failures. Create `lib/email_service.ex`:\n\n```elixir\ndefmodule EmailService do\n  use Reactor.Step\n\n  @impl true\n  def run(arguments, _context, _options) do\n    if String.ends_with?(arguments.email, \"@example.com\") do\n      {:ok, %{message_id: \"msg_123\", sent_at: DateTime.utc_now()}}\n    else\n      {:error, %{type: :network_timeout, message: \"Email service unavailable\"}}\n    end\n  end\n\n  @impl true\n  def compensate(error, _arguments, _context, _options) do\n    case error do\n      %{type: :network_timeout} ->\n        # Network errors are usually temporary, so retry\n        :retry\n      \n      _other ->\n        # Other errors are permanent, don't retry\n        :ok\n    end\n  end\n\n  @impl true\n  def undo(result, _arguments, _context, _options) do\n    IO.puts(\"Canceling email message #{result.message_id}\")\n    :ok\n  end\nend\n```","ref":"02-error-handling.html#step-3-create-a-step-with-error-handling"},{"type":"extras","title":"Step 4: Create a database service that needs rollback - Error Handling with Compensation and Undo","doc":"Create `lib/database_service.ex`:\n\n```elixir\ndefmodule DatabaseService do\n  use Reactor.Step\n\n  @impl true\n  def run(arguments, _context, _options) do\n    user = %{\n      id: :rand.uniform(10000),\n      email: arguments.email,\n      password_hash: arguments.password_hash,\n      created_at: DateTime.utc_now()\n    }\n    \n    {:ok, user}\n  end\n\n  @impl true\n  def compensate(_error, _arguments, _context, _options) do\n    # Database errors are usually retryable\n    :retry\n  end\n\n  @impl true\n  def undo(user, _arguments, _context, _options) do\n    IO.puts(\"Rolling back user creation for #{user.email} (ID: #{user.id})\")\n    :ok\n  end\nend\n```","ref":"02-error-handling.html#step-4-create-a-database-service-that-needs-rollback"},{"type":"extras","title":"Step 5: Build a reactor with error handling - Error Handling with Compensation and Undo","doc":"Now create `lib/resilient_user_registration.ex`:\n\n```elixir\ndefmodule ResilientUserRegistration do\n  use Reactor\n\n  input :email\n  input :password\n\n  step :validate_email do\n    argument :email, input(:email)\n    \n    run fn %{email: email}, _context ->\n      if String.contains?(email, \"@\") and String.length(email) > 5 do\n        {:ok, email}\n      else\n        {:error, \"Email must contain @ and be longer than 5 characters\"}\n      end\n    end\n  end\n\n  step :hash_password do\n    argument :password, input(:password)\n    \n    run fn %{password: password}, _context ->\n      if String.length(password) >= 8 do\n        hashed = :crypto.hash(:sha256, password) |> Base.encode16()\n        {:ok, hashed}\n      else\n        {:error, \"Password must be at least 8 characters\"}\n      end\n    end\n  end\n\n  step :create_user, DatabaseService do\n    argument :email, result(:validate_email)\n    argument :password_hash, result(:hash_password)\n    max_retries 3\n  end\n\n  step :send_welcome_email, EmailService do\n    argument :email, result(:validate_email)\n    argument :user, result(:create_user)\n    max_retries 2\n  end\n\n  step :send_admin_notification, EmailService do\n    argument :email, value(\"admin@company.com\")\n    argument :user, result(:create_user)\n    max_retries 1\n  end\n\n  return :create_user\nend\n```","ref":"02-error-handling.html#step-5-build-a-reactor-with-error-handling"},{"type":"extras","title":"Step 6: Test the error handling - Error Handling with Compensation and Undo","doc":"Let's test our reactor in IEx:\n\n```bash\niex -S mix\n```\n\n```elixir\n# Test with a valid @example.com email (should succeed)\n{:ok, user} = Reactor.run(ResilientUserRegistration, %{\n  email: \"alice@example.com\",\n  password: \"secretpassword123\"\n})\n\n# Test with a non-@example.com email (will trigger retry logic)\n{:error, reason} = Reactor.run(ResilientUserRegistration, %{\n  email: \"bob@gmail.com\",\n  password: \"secretpassword123\"\n})\n\n# Test with invalid inputs\n{:error, reason} = Reactor.run(ResilientUserRegistration, %{\n  email: \"bad\",\n  password: \"short\"\n})\n```","ref":"02-error-handling.html#step-6-test-the-error-handling"},{"type":"extras","title":"Step 7: Understanding the behaviour - Error Handling with Compensation and Undo","doc":"When you run the tests, you'll see different behaviours:\n\n**Successful execution** (with @example.com email): All steps succeed, user is created and emails are sent.\n\n**Retry scenario** (with non-@example.com email): EmailService fails with network timeout, compensation returns `:retry`, step retries up to max_retries limit.\n\n**Validation failures**: Invalid input fails immediately without retries - compensation logic determines these are permanent errors.","ref":"02-error-handling.html#step-7-understanding-the-behaviour"},{"type":"extras","title":"What you learned - Error Handling with Compensation and Undo","doc":"You now understand Reactor's error handling mechanisms:\n\n- **[Compensation](../reference/glossary.md#compensation)** handles step failures with retry logic\n- **Undo operations** roll back successful steps when later steps fail\n- **Max retries** controls how many times compensation can retry a step\n- **Error types** should be handled differently (retry vs fail)\n- **Context contains retry state** for intelligent retry logic","ref":"02-error-handling.html#what-you-learned"},{"type":"extras","title":"What's next - Error Handling with Compensation and Undo","doc":"Now that you can handle errors, you're ready for more advanced concepts:\n\n- **[Async Workflows](03-async-workflows.md)** - Explore concurrent processing patterns\n- **[Composition](04-composition.md)** - Build complex workflows with sub-reactors  \n- **[Testing Strategies](documentation/how-to/testing-strategies.md)** - Learn how to test error scenarios","ref":"02-error-handling.html#what-s-next"},{"type":"extras","title":"Common issues - Error Handling with Compensation and Undo","doc":"**Steps retry infinitely**: Always set `max_retries` and ensure compensation doesn't always return `:retry`\n\n**Undo operations fail**: Make undo operations idempotent - they should succeed even if called multiple times\n\n**Reactor fails instead of retrying**: Check that your compensation function returns `:retry`, not `{:error, reason}`\n\nHappy building resilient workflows! 🛡️","ref":"02-error-handling.html#common-issues"},{"type":"extras","title":"Building Async Workflows","doc":"# Building Async Workflows\n\nIn this tutorial, you'll learn how to build efficient concurrent workflows that take advantage of Reactor's dependency resolution and async execution capabilities.","ref":"03-async-workflows.html"},{"type":"extras","title":"What you'll build - Building Async Workflows","doc":"A data processing pipeline that:\n1. Fetches user data from multiple sources concurrently\n2. Processes different data types in parallel\n3. Aggregates results efficiently\n4. Handles mixed sync/async requirements\n5. Optimizes performance with proper concurrency control","ref":"03-async-workflows.html#what-you-ll-build"},{"type":"extras","title":"You'll learn - Building Async Workflows","doc":"- How Reactor's concurrency model works\n- When to use async vs sync execution\n- How to optimize workflow performance\n- Managing dependencies for maximum parallelisation\n- Controlling concurrency limits and resource usage","ref":"03-async-workflows.html#you-ll-learn"},{"type":"extras","title":"Prerequisites - Building Async Workflows","doc":"- Complete the [Getting Started tutorial](01-getting-started.md)\n- Complete the [Error Handling tutorial](02-error-handling.md)\n- Basic understanding of Elixir processes","ref":"03-async-workflows.html#prerequisites"},{"type":"extras","title":"Step 1: Set up the project - Building Async Workflows","doc":"If you don't have a project from the previous tutorials:\n\n```bash\nmix igniter.new reactor_tutorial --install reactor\ncd reactor_tutorial\n```","ref":"03-async-workflows.html#step-1-set-up-the-project"},{"type":"extras","title":"Step 2: Understanding Reactor's concurrency model - Building Async Workflows","doc":"Reactor runs steps **asynchronously by default**:\n\n- **Independent steps run in parallel** - Steps with no dependencies execute simultaneously\n- **Dependencies control execution order** - Steps wait for their dependencies to complete\n- **Automatic task management** - Reactor manages Elixir tasks and supervision\n- **Configurable concurrency** - Control how many steps run at once","ref":"03-async-workflows.html#step-2-understanding-reactor-s-concurrency-model"},{"type":"extras","title":"Async vs Sync execution - Building Async Workflows","doc":"```elixir\n# Async (default) - runs in a separate task\nstep :fetch_data do\n  async? true  # This is the default\n  run &fetch_from_api/1\nend\n\n# Sync - runs in the main process\nstep :critical_operation do\n  async? false  # Forces synchronous execution\n  run &update_database/1\nend\n```","ref":"03-async-workflows.html#async-vs-sync-execution"},{"type":"extras","title":"Step 3: Create simple data operations - Building Async Workflows","doc":"Let's create some data operations that will run concurrently. Create `lib/data_sources.ex`:\n\n```elixir\ndefmodule DataSources do\n  def fetch_user_profile(user_id) do\n    Process.sleep(200)\n    \n    {:ok, %{\n      id: user_id,\n      name: \"User #{user_id}\",\n      email: \"user#{user_id}@example.com\"\n    }}\n  end\n\n  def fetch_user_preferences(user_id) do\n    Process.sleep(150)\n    \n    {:ok, %{\n      user_id: user_id,\n      theme: \"light\",\n      language: \"en\"\n    }}\n  end\n\n  def fetch_user_activity(user_id) do\n    Process.sleep(300)\n    \n    {:ok, %{\n      user_id: user_id,\n      last_login: DateTime.utc_now(),\n      login_count: 42\n    }}\n  end\nend\n```","ref":"03-async-workflows.html#step-3-create-simple-data-operations"},{"type":"extras","title":"Step 4: Build a concurrent data reactor - Building Async Workflows","doc":"Now let's build a reactor that fetches data concurrently. Create `lib/async_user_data_reactor.ex`:\n\n```elixir\ndefmodule AsyncUserDataReactor do\n  use Reactor\n\n  input :user_id\n\n  # These steps have no dependencies on each other, so they run in parallel\n  step :fetch_profile do\n    argument :user_id, input(:user_id)\n    run fn %{user_id: user_id}, _context ->\n      DataSources.fetch_user_profile(user_id)\n    end\n  end\n\n  step :fetch_preferences do\n    argument :user_id, input(:user_id)  \n    run fn %{user_id: user_id}, _context ->\n      DataSources.fetch_user_preferences(user_id)\n    end\n  end\n\n  step :fetch_activity do\n    argument :user_id, input(:user_id)\n    run fn %{user_id: user_id}, _context ->\n      DataSources.fetch_user_activity(user_id)\n    end\n  end\n\n  # This step waits for all the fetch steps to complete\n  step :aggregate_data do\n    argument :profile, result(:fetch_profile)\n    argument :preferences, result(:fetch_preferences)\n    argument :activity, result(:fetch_activity)\n    \n    run fn args, _context ->\n      user_data = %{\n        profile: args.profile,\n        preferences: args.preferences,\n        activity: args.activity,\n        summary: \"User #{args.profile.id} has #{args.activity.login_count} logins\"\n      }\n      {:ok, user_data}\n    end\n  end\n\n  return :aggregate_data\nend\n```","ref":"03-async-workflows.html#step-4-build-a-concurrent-data-reactor"},{"type":"extras","title":"Step 5: Test the concurrent execution - Building Async Workflows","doc":"Let's test our reactor to see the difference between concurrent and sequential execution:\n\n```bash\niex -S mix\n```\n\n```elixir\n# Test the concurrent execution\nstart_time = :erlang.monotonic_time(:millisecond)\n\n{:ok, result} = Reactor.run(AsyncUserDataReactor, %{user_id: 123})\n\nend_time = :erlang.monotonic_time(:millisecond)\nduration = end_time - start_time\n\nIO.puts(\"Completed in #{duration}ms\")\nIO.inspect(result.summary)\n```","ref":"03-async-workflows.html#step-5-test-the-concurrent-execution"},{"type":"extras","title":"Step 6: Understanding the execution flow - Building Async Workflows","doc":"The reactor completes in about **300ms** instead of 650ms (200+150+300) because:\n\n1. All three fetch steps run **concurrently** (they only depend on input)\n2. `:aggregate_data` waits for all three to complete\n3. Total time is limited by the slowest operation (300ms) not the sum\n\nCompare with synchronous execution:\n\n```elixir\n# Force all steps to run synchronously\n{:ok, result} = Reactor.run(AsyncUserDataReactor, %{user_id: 123}, %{}, async?: false)\n```\n\nThis takes the full 650ms because each step runs sequentially.","ref":"03-async-workflows.html#step-6-understanding-the-execution-flow"},{"type":"extras","title":"Step 7: Controlling async behavior - Building Async Workflows","doc":"You can control which steps run synchronously when needed:\n\n```elixir\ndefmodule SyncVsAsyncReactor do\n  use Reactor\n\n  input :user_id\n\n  # I/O operations - keep async (default)\n  step :fetch_profile do\n    argument :user_id, input(:user_id)\n    # async? true  # This is the default\n    run fn %{user_id: user_id}, _context ->\n      DataSources.fetch_user_profile(user_id)\n    end\n  end\n\n  # CPU-intensive work - tune reactor concurrency as needed\n  step :process_data do\n    argument :profile, result(:fetch_profile)\n    # async? true is the default - adjust reactor max_concurrency instead\n    \n    run fn %{profile: profile}, _context ->\n      Process.sleep(100)\n      {:ok, Map.put(profile, :processed, true)}\n    end\n  end\n\n  return :process_data\nend\n```","ref":"03-async-workflows.html#step-7-controlling-async-behavior"},{"type":"extras","title":"What you learned - Building Async Workflows","doc":"You now understand Reactor's concurrency model:\n\n- **Steps run async by default** - Enables automatic parallelisation\n- **Dependencies determine execution order** - Independent steps run concurrently\n- **Tune concurrency for your workload** - Adjust limits based on system capacity\n- **Performance optimisation** - Balance concurrency with system resources\n- **Concurrency control** - Manage resource usage with limits","ref":"03-async-workflows.html#what-you-learned"},{"type":"extras","title":"Performance guidelines: - Building Async Workflows","doc":"- **I/O operations** → Generally benefit from high concurrency\n- **CPU-intensive work** → Tune concurrency to match CPU cores and workload\n- **Resource limits** → Set concurrency limits based on system capacity","ref":"03-async-workflows.html#performance-guidelines"},{"type":"extras","title":"What's next - Building Async Workflows","doc":"Now that you understand concurrency, you're ready for advanced workflow patterns:\n\n- **[Composition](04-composition.md)** - Build complex workflows with sub-reactors\n- **[Recursive Execution](05-recursive-execution.md)** - Advanced iterative patterns\n- **[Testing Strategies](documentation/how-to/testing-strategies.md)** - Test concurrent workflows effectively","ref":"03-async-workflows.html#what-s-next"},{"type":"extras","title":"Common issues - Building Async Workflows","doc":"**Steps aren't running in parallel**: Check for hidden dependencies in arguments - each argument creates a dependency\n\nFor comprehensive performance and concurrency troubleshooting, see [Performance Optimization](documentation/how-to/performance-optimization.md#troubleshooting).\n\nHappy building concurrent workflows! ⚡","ref":"03-async-workflows.html#common-issues"},{"type":"extras","title":"Building Complex Workflows with Composition","doc":"# Building Complex Workflows with Composition\n\nIn this tutorial, you'll learn how to build large, maintainable workflows by composing smaller reactors together. This is essential for managing complexity in real-world applications.","ref":"04-composition.html"},{"type":"extras","title":"What you'll build - Building Complex Workflows with Composition","doc":"A multi-stage e-commerce order processing system that:\n1. **User Management** - Validates and enriches user data\n2. **Inventory Management** - Checks and reserves product inventory  \n3. **Payment Processing** - Handles payment authorization and capture\n4. **Order Fulfillment** - Coordinates shipping and notifications\n5. **Master Orchestrator** - Composes all sub-workflows together","ref":"04-composition.html#what-you-ll-build"},{"type":"extras","title":"You'll learn - Building Complex Workflows with Composition","doc":"- How to break complex workflows into composable reactors\n- When to use composition vs building one large reactor\n- How to pass data between composed reactors\n- Error handling and rollback across composed workflows\n- Testing strategies for composed systems","ref":"04-composition.html#you-ll-learn"},{"type":"extras","title":"Prerequisites - Building Complex Workflows with Composition","doc":"- Complete the [Getting Started tutorial](01-getting-started.md)\n- Complete the [Error Handling tutorial](02-error-handling.md)  \n- Complete the [Async Workflows tutorial](03-async-workflows.md)","ref":"04-composition.html#prerequisites"},{"type":"extras","title":"Step 1: Set up the project - Building Complex Workflows with Composition","doc":"If you don't have a project from the previous tutorials:\n\n```bash\nmix igniter.new reactor_tutorial --install reactor\ncd reactor_tutorial\n```","ref":"04-composition.html#step-1-set-up-the-project"},{"type":"extras","title":"Step 2: Understanding Reactor composition - Building Complex Workflows with Composition","doc":"Reactor composition allows you to:\n\n**Break down complexity** - Instead of one massive reactor, create focused sub-reactors:\n```elixir\n# Use composition:\ndefmodule OrderProcessor do\n  use Reactor\n  \n  compose :user_management, UserManagementReactor\n  compose :inventory_check, InventoryReactor  \n  compose :payment_processing, PaymentReactor\n  compose :fulfillment, FulfillmentReactor\nend\n```\n\n**Enable reusability** - Sub-reactors can be used in multiple contexts:\n```elixir\ncompose :validate_buyer, UserManagementReactor\ncompose :validate_seller, UserManagementReactor  \n```\n\n**Improve testability** - Test each sub-reactor independently.","ref":"04-composition.html#step-2-understanding-reactor-composition"},{"type":"extras","title":"Step 3: Create simple domain reactors - Building Complex Workflows with Composition","doc":"Let's start by building focused reactors for each domain. Create `lib/user_validation_reactor.ex`:\n\n```elixir\ndefmodule UserValidationReactor do\n  use Reactor\n\n  input :user_id\n\n  step :fetch_user do\n    argument :user_id, input(:user_id)\n    \n    run fn %{user_id: user_id}, _context ->\n      {:ok, %{\n        id: user_id,\n        name: \"User #{user_id}\",\n        email: \"user#{user_id}@example.com\",\n        active: true\n      }}\n    end\n  end\n\n  step :validate_user do\n    argument :user, result(:fetch_user)\n    \n    run fn %{user: user}, _context ->\n      if user.active do\n        {:ok, user}\n      else\n        {:error, \"User is not active\"}\n      end\n    end\n  end\n\n  return :validate_user\nend\n```\n\nCreate `lib/inventory_reactor.ex`:\n\n```elixir\ndefmodule InventoryReactor do\n  use Reactor\n\n  input :product_id\n  input :quantity\n\n  step :check_availability do\n    argument :product_id, input(:product_id)\n    argument :quantity, input(:quantity)\n    \n    run fn %{product_id: product_id, quantity: quantity}, _context ->\n      available = 50\n      \n      if quantity <= available do\n        {:ok, %{product_id: product_id, available: available}}\n      else\n        {:error, \"Not enough inventory\"}\n      end\n    end\n  end\n\n  step :reserve_items do\n    argument :availability, result(:check_availability)\n    argument :quantity, input(:quantity)\n    \n    run fn %{availability: avail, quantity: qty}, _context ->\n      reservation = %{\n        product_id: avail.product_id,\n        quantity: qty,\n        reserved_at: DateTime.utc_now()\n      }\n      {:ok, reservation}\n    end\n  end\n\n  return :reserve_items\nend\n```\n\nCreate `lib/payment_reactor.ex`:\n\n```elixir\ndefmodule PaymentReactor do\n  use Reactor\n\n  input :user_id\n  input :amount\n\n  step :validate_payment do\n    argument :user_id, input(:user_id)\n    argument :amount, input(:amount)\n    \n    run fn %{user_id: user_id, amount: amount}, _context ->\n      if amount > 0 and amount < 10000 do\n        {:ok, %{user_id: user_id, amount: amount, valid: true}}\n      else\n        {:error, \"Invalid payment amount\"}\n      end\n    end\n  end\n\n  step :process_payment do\n    argument :payment_info, result(:validate_payment)\n    \n    run fn %{payment_info: info}, _context ->\n      payment = %{\n        payment_id: \"pay_#{:rand.uniform(1000)}\",\n        user_id: info.user_id,\n        amount: info.amount,\n        status: :completed,\n        processed_at: DateTime.utc_now()\n      }\n      {:ok, payment}\n    end\n  end\n\n  return :process_payment\nend\n```","ref":"04-composition.html#step-3-create-simple-domain-reactors"},{"type":"extras","title":"Step 4: Create the master orchestrator - Building Complex Workflows with Composition","doc":"Now create the main reactor that composes all sub-reactors. Create `lib/order_processing_reactor.ex`:\n\n```elixir\ndefmodule OrderProcessingReactor do\n  use Reactor\n\n  input :user_id\n  input :product_id\n  input :quantity\n  input :amount\n\n  # Step 1: Validate user\n  compose :user_validation, UserValidationReactor do\n    argument :user_id, input(:user_id)\n  end\n\n  # Step 2: Check inventory (can run in parallel with payment)\n  compose :inventory_check, InventoryReactor do\n    argument :product_id, input(:product_id)\n    argument :quantity, input(:quantity)\n  end\n\n  # Step 3: Process payment\n  compose :payment_processing, PaymentReactor do\n    argument :user_id, input(:user_id)\n    argument :amount, input(:amount)\n  end\n\n  # Step 4: Create final order (depends on all previous steps)\n  step :create_order do\n    argument :user, result(:user_validation)\n    argument :reservation, result(:inventory_check)\n    argument :payment, result(:payment_processing)\n    \n    run fn %{user: user, reservation: res, payment: pay}, _context ->\n      order = %{\n        order_id: \"order_#{:rand.uniform(1000)}\",\n        user: user,\n        product_id: res.product_id,\n        quantity: res.quantity,\n        payment_id: pay.payment_id,\n        total: pay.amount,\n        created_at: DateTime.utc_now()\n      }\n      \n      {:ok, order}\n    end\n  end\n\n  return :create_order\nend\n```","ref":"04-composition.html#step-4-create-the-master-orchestrator"},{"type":"extras","title":"Step 5: Test the composition - Building Complex Workflows with Composition","doc":"Let's test our composed reactor:\n\n```bash\niex -S mix\n```\n\n```elixir\n# Test the composed order processing\n{:ok, order} = Reactor.run(OrderProcessingReactor, %{\n  user_id: 123,\n  product_id: 456,\n  quantity: 2,\n  amount: 99.99\n})\n\nIO.inspect(order.order_id)\n# Should output something like \"order_123\"\n\n# Test individual reactors too\n{:ok, user} = Reactor.run(UserValidationReactor, %{user_id: 123})\nIO.inspect(user.name)\n# Should output \"User 123\"\n```","ref":"04-composition.html#step-5-test-the-composition"},{"type":"extras","title":"What you learned - Building Complex Workflows with Composition","doc":"You now understand Reactor composition:\n\n- **Composition over monoliths** - Many small reactors are easier to manage than one large one\n- **Clear interfaces** - Define clear inputs and outputs for each sub-reactor\n- **Independent testing** - Test each reactor in isolation before testing compositions\n- **Error boundaries** - Failures in sub-reactors can be handled by the parent reactor\n- **Reusability** - Well-designed sub-reactors can be used in multiple contexts","ref":"04-composition.html#what-you-learned"},{"type":"extras","title":"Design guidelines: - Building Complex Workflows with Composition","doc":"- **Single responsibility** - Each reactor should have one clear purpose\n- **Loose coupling** - Minimise dependencies between reactors\n- **High cohesion** - Related steps belong in the same reactor","ref":"04-composition.html#design-guidelines"},{"type":"extras","title":"What's next - Building Complex Workflows with Composition","doc":"Now that you understand composition, you're ready for advanced patterns:\n\n- **[Recursive Execution](05-recursive-execution.md)** - Handle iterative and recursive workflows\n- **[Testing Strategies](documentation/how-to/testing-strategies.md)** - Comprehensive testing approaches\n- **[Performance Optimization](documentation/how-to/performance-optimization.md)** - Advanced performance techniques","ref":"04-composition.html#what-s-next"},{"type":"extras","title":"Common issues - Building Complex Workflows with Composition","doc":"**Sub-reactor outputs don't match expectations**: Use clear interface contracts and validate inputs/outputs in tests\n\n**Error handling doesn't work across boundaries**: Ensure compensation and undo are implemented in the appropriate reactor layer\n\n**Composed reactors are hard to debug**: Test each sub-reactor individually and use descriptive logging\n\nHappy building modular workflows! 🧩","ref":"04-composition.html#common-issues"},{"type":"extras","title":"Building Iterative Workflows with Recursive Execution","doc":"# Building Iterative Workflows with Recursive Execution\n\nIn this tutorial, you'll learn how to build iterative workflows that process data repeatedly until a condition is met. This is perfect for mathematical algorithms, data processing pipelines, and convergence calculations.","ref":"05-recursive-execution.html"},{"type":"extras","title":"What you'll build - Building Iterative Workflows with Recursive Execution","doc":"Simple iterative algorithms demonstrating recursive patterns:\n1. **Countdown Calculator** - Count down from a number to zero\n2. **List Processor** - Process items one at a time until done","ref":"05-recursive-execution.html#what-you-ll-build"},{"type":"extras","title":"You'll learn - Building Iterative Workflows with Recursive Execution","doc":"- How Reactor's recursive execution works\n- Setting up proper exit conditions and iteration limits\n- Managing state between iterations\n- When to use recursion vs other Reactor patterns\n- Building convergent algorithms and iterative processors","ref":"05-recursive-execution.html#you-ll-learn"},{"type":"extras","title":"Prerequisites - Building Iterative Workflows with Recursive Execution","doc":"- Complete the [Getting Started tutorial](01-getting-started.md)\n- Complete the [Error Handling tutorial](02-error-handling.md)\n- Complete the [Async Workflows tutorial](03-async-workflows.md)\n- Complete the [Composition tutorial](04-composition.md)","ref":"05-recursive-execution.html#prerequisites"},{"type":"extras","title":"Step 1: Set up the project - Building Iterative Workflows with Recursive Execution","doc":"If you don't have a project from the previous tutorials:\n\n```bash\nmix igniter.new reactor_tutorial --install reactor\ncd reactor_tutorial\n```","ref":"05-recursive-execution.html#step-1-set-up-the-project"},{"type":"extras","title":"Step 2: Understanding recursive execution - Building Iterative Workflows with Recursive Execution","doc":"Reactor's recursive execution allows you to:\n\n**Repeat until a condition is met:**\n```elixir\nrecurse :calculate, CalculatorReactor do\n  argument :value, input(:start_value)\n  exit_condition fn %{done: done} -> done == true end\n  max_iterations 100\nend\n```\n\n**Process data iteratively:**\n```elixir\nrecurse :process, ProcessorReactor do\n  argument :remaining, input(:data_list)\n  exit_condition fn %{remaining: list} -> Enum.empty?(list) end\n  max_iterations 1000\nend\n```","ref":"05-recursive-execution.html#step-2-understanding-recursive-execution"},{"type":"extras","title":"Step 3: Build a simple countdown reactor - Building Iterative Workflows with Recursive Execution","doc":"Let's start with a simple recursive algorithm. Create `lib/countdown_reactor.ex`:\n\n```elixir\ndefmodule CountdownReactor do\n  use Reactor\n\n  input :current_number\n\n  step :countdown_step do\n    argument :current_number, input(:current_number)\n\n    run fn %{current_number: num}, _context ->\n      new_number = num - 1\n      {:ok, %{current_number: new_number}}\n    end\n  end\n\n  return :countdown_step\nend\n\ndefmodule CountdownExample do\n  use Reactor\n\n  input :start_number\n\n  recurse :countdown, CountdownReactor do\n    argument :current_number, input(:start_number)\n    max_iterations 100\n    exit_condition fn %{current_number: num} -> num <= 0 end\n  end\n\n  step :show_result do\n    argument :final_state, result(:countdown)\n    argument :start_number, input(:start_number)\n\n    run fn %{final_state: state, start_number: start}, _context ->\n      result = %{\n        started_at: start,\n        finished_at: state.current_number,\n        message: \"Counted down from #{start} to #{state.current_number}\"\n      }\n      {:ok, result}\n    end\n  end\n\n  return :show_result\nend\n```","ref":"05-recursive-execution.html#step-3-build-a-simple-countdown-reactor"},{"type":"extras","title":"Step 4: Build an accumulator - Building Iterative Workflows with Recursive Execution","doc":"Create `lib/accumulator.ex`:\n\n```elixir\ndefmodule AccumulatorReactor do\n  use Reactor\n\n  input :current_total\n  input :target_total\n\n  step :add_random_amount do\n    argument :current_total, input(:current_total)\n    argument :target_total, input(:target_total)\n\n    run fn %{current_total: current, target_total: target}, _context ->\n      # Add a random amount between 1 and 10\n      addition = :rand.uniform(10)\n      new_total = current + addition\n      \n      reached_target = new_total >= target\n\n      {:ok, %{\n        current_total: new_total,\n        target_total: target,\n        last_addition: addition,\n        reached_target: reached_target\n      }}\n    end\n  end\n\n  return :add_random_amount\nend\n\ndefmodule AccumulatorExample do\n  use Reactor\n\n  input :target_score\n\n  recurse :accumulate, AccumulatorReactor do\n    argument :current_total, value(0)\n    argument :target_total, input(:target_score)\n    max_iterations 50\n    exit_condition fn %{reached_target: reached} -> reached == true end\n  end\n\n  step :show_results do\n    argument :final_state, result(:accumulate)\n    argument :target, input(:target_score)\n\n    run fn %{final_state: state, target: target}, _context ->\n      result = %{\n        target_score: target,\n        final_total: state.current_total,\n        last_addition: state.last_addition,\n        message: \"Reached #{state.current_total} (target was #{target})\"\n      }\n      {:ok, result}\n    end\n  end\n\n  return :show_results\nend\n```","ref":"05-recursive-execution.html#step-4-build-an-accumulator"},{"type":"extras","title":"Step 5: Advanced example - Convergence algorithm - Building Iterative Workflows with Recursive Execution","doc":"Now let's look at a more sophisticated pattern - iterative convergence. Create `lib/square_root_calculator.ex`:\n\n```elixir\ndefmodule NewtonMethodReactor do\n  use Reactor\n\n  input :current_guess\n  input :target_number\n\n  step :newton_iteration do\n    argument :current_guess, input(:current_guess)\n    argument :target_number, input(:target_number)\n\n    run fn %{current_guess: guess, target_number: target}, _context ->\n      # Newton's method: x_new = (x + target/x) / 2\n      new_guess = (guess + target / guess) / 2\n      difference = abs(new_guess - guess)\n      \n      # Consider converged when difference is very small\n      converged = difference < 0.0001\n\n      {:ok, %{\n        current_guess: new_guess,\n        target_number: target,\n        converged: converged\n      }}\n    end\n  end\n\n  return :newton_iteration\nend\n\ndefmodule SquareRootCalculator do\n  use Reactor\n\n  input :number\n\n  recurse :converge, NewtonMethodReactor do\n    argument :current_guess, input(:number)  # Start with the number itself\n    argument :target_number, input(:number)\n    max_iterations 20\n    exit_condition fn %{converged: converged} -> converged == true end\n  end\n\n  step :show_result do\n    argument :final_state, result(:converge)\n    argument :original_number, input(:number)\n\n    run fn %{final_state: state, original_number: num}, _context ->\n      result = %{\n        number: num,\n        square_root: state.current_guess,\n        message: \"√#{num} ≈ #{Float.round(state.current_guess, 6)}\"\n      }\n      {:ok, result}\n    end\n  end\n\n  return :show_result\nend\n```","ref":"05-recursive-execution.html#step-5-advanced-example-convergence-algorithm"},{"type":"extras","title":"Step 6: Test the examples - Building Iterative Workflows with Recursive Execution","doc":"Let's test our recursive reactors:\n\n```bash\niex -S mix\n```\n\n```elixir\n# Test the countdown example\n{:ok, result} = Reactor.run(CountdownExample, %{start_number: 5})\nIO.inspect(result.message)\n# Should output: \"Counted down from 5 to 0\"\n\n# Test the accumulator\n{:ok, result} = Reactor.run(AccumulatorExample, %{target_score: 50})\nIO.inspect(result.message)\n# Should output something like: \"Reached 53 (target was 50)\"\n\n# Test the square root calculator\n{:ok, result} = Reactor.run(SquareRootCalculator, %{number: 25})\nIO.inspect(result.message)\n# Should output: \"√25 ≈ 5.0\"\n```","ref":"05-recursive-execution.html#step-6-test-the-examples"},{"type":"extras","title":"What you learned - Building Iterative Workflows with Recursive Execution","doc":"You now understand Reactor's recursive execution:\n\n- **Always provide termination conditions** - Use both `exit_condition` and `max_iterations`\n- **Keep state simple** - Use flat maps with simple values\n- **Test with small examples** - Start simple before building complexity\n- **Convergence detection** - Check when iterative improvements become negligible","ref":"05-recursive-execution.html#what-you-learned"},{"type":"extras","title":"When to use recursion: - Building Iterative Workflows with Recursive Execution","doc":"**Use recursion for:**\n- Counting and decrementing patterns\n- Accumulation until a threshold is reached\n- Iterative calculations with clear stopping conditions\n- Mathematical convergence algorithms\n\n**Avoid recursion for:**\n- Simple sequential processing (use regular steps)\n- One-time transformations (use map steps)\n- Parallel processing of collections (use map steps with batching)","ref":"05-recursive-execution.html#when-to-use-recursion"},{"type":"extras","title":"What's next - Building Iterative Workflows with Recursive Execution","doc":"You've mastered all core Reactor patterns! Ready for specialized guides:\n\n- **[Testing Strategies](documentation/how-to/testing-strategies.md)** - Comprehensive testing approaches\n- **[Performance Optimization](documentation/how-to/performance-optimization.md)** - Advanced performance techniques  \n- **[Debugging Workflows](documentation/how-to/debugging-workflows.md)** - Troubleshooting complex reactors","ref":"05-recursive-execution.html#what-s-next"},{"type":"extras","title":"Common issues - Building Iterative Workflows with Recursive Execution","doc":"**Recursion never terminates**: Ensure exit conditions can actually be met; always include `max_iterations` as backup\n\n**\"Maximum iterations exceeded\" error**: Increase `max_iterations` or fix logic preventing proper convergence\n\n**State not passing correctly**: Ensure output structure exactly matches input requirements; keep state flat and simple\n\n**Exit condition never triggers**: Add debug logging to verify exit condition logic; check that condition field exists in state\n\nHappy building iterative workflows! 🔄","ref":"05-recursive-execution.html#common-issues"},{"type":"extras","title":"Reactor Cheatsheet","doc":"# Reactor Cheatsheet\n\nReactor is a dynamic, concurrent, dependency resolving saga orchestrator for Elixir.","ref":"reactor-cheatsheet.html"},{"type":"extras","title":"Basic Reactor Definition - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#basic-reactor-definition"},{"type":"extras","title":"Simple Reactor - Reactor Cheatsheet","doc":"```elixir\ndefmodule MyReactor do\n  use Reactor\n\n  input :email\n  input :password\n\n  step :hash_password do\n    argument :password, input(:password)\n    run &Bcrypt.hash_pwd_salt/1\n  end\n\n  step :create_user, MyApp.CreateUser do\n    argument :email, input(:email)\n    argument :password_hash, result(:hash_password)\n  end\n\n  return :create_user\nend\n```","ref":"reactor-cheatsheet.html#simple-reactor"},{"type":"extras","title":"Running Reactors - Reactor Cheatsheet","doc":"```elixir\n# Basic execution\n{:ok, result} = Reactor.run(MyReactor, \n  email: \"user@example.com\", \n  password: \"secret\"\n)\n\n# With context and options\n{:ok, result} = Reactor.run(MyReactor, \n  inputs, \n  %{current_user: user},\n  async?: false,\n  max_concurrency: 10\n)\n\n# Halting and resuming\n{:halted, state} = Reactor.run(MyReactor, inputs)\n{:ok, result} = Reactor.run(state, %{}, %{})\n```","ref":"reactor-cheatsheet.html#running-reactors"},{"type":"extras","title":"Step Types - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#step-types"},{"type":"extras","title":"Basic Steps - Reactor Cheatsheet","doc":"```elixir\n# Anonymous function\nstep :transform do\n  argument :data, input(:raw_data)\n  run fn %{data: data}, _context ->\n    {:ok, String.upcase(data)}\n  end\nend\n\n# Module implementation\nstep :create_user, MyApp.Steps.CreateUser do\n  argument :email, result(:validate_email)\n  argument :data, input(:user_data)\nend\n\n# Sync/async control\nstep :critical_operation do\n  async? false\n  run &important_work/1\nend\n```","ref":"reactor-cheatsheet.html#basic-steps"},{"type":"extras","title":"Debug Steps - Reactor Cheatsheet","doc":"```elixir\ndebug :log_user do\n  argument :user, result(:create_user)\n  argument :message, value(\"User created\")\nend\n```","ref":"reactor-cheatsheet.html#debug-steps"},{"type":"extras","title":"Map Steps - Reactor Cheatsheet","doc":"```elixir\nmap :process_users do\n  source input(:user_list)\n  batch_size 10\n  allow_async? true\n\n  step :validate_user do\n    argument :user, element(:process_users)\n    run &validate_user/1\n  end\nend\n```","ref":"reactor-cheatsheet.html#map-steps"},{"type":"extras","title":"Compose Steps - Reactor Cheatsheet","doc":"```elixir\ncompose :sub_workflow, MyApp.SubReactor do\n  argument :input_data, result(:prepare_data)\nend\n```","ref":"reactor-cheatsheet.html#compose-steps"},{"type":"extras","title":"Advanced Step Types - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#advanced-step-types"},{"type":"extras","title":"Switch Steps - Reactor Cheatsheet","doc":"```elixir\nswitch :handle_user_type do\n  on result(:user)\n\n  matches? &(&1.type == :premium) do\n    step :setup_premium do\n      argument :user, result(:user)\n      run &setup_premium_features/1\n    end\n  end\n\n  default do\n    step :setup_basic do\n      argument :user, result(:user)\n      run &setup_basic_features/1\n    end\n  end\nend\n```","ref":"reactor-cheatsheet.html#switch-steps"},{"type":"extras","title":"Group Steps - Reactor Cheatsheet","doc":"```elixir\ngroup :user_setup do\n  before_all &MyApp.setup_database/3\n  after_all &MyApp.cleanup_database/1\n\n  step :create_profile do\n    # Profile creation logic\n  end\n\n  step :send_welcome_email do\n    # Email logic\n  end\nend\n```","ref":"reactor-cheatsheet.html#group-steps"},{"type":"extras","title":"Around Steps - Reactor Cheatsheet","doc":"```elixir\naround :transaction, &MyApp.with_transaction/4 do\n  step :create_user do\n    # User creation in transaction\n  end\n\n  step :create_profile do\n    # Profile creation in transaction\n  end\nend\n```","ref":"reactor-cheatsheet.html#around-steps"},{"type":"extras","title":"Collect Steps - Reactor Cheatsheet","doc":"```elixir\ncollect :user_summary do\n  argument :user, result(:create_user)\n  argument :profile, result(:create_profile)\n  argument :settings, result(:create_settings)\n\n  transform fn inputs ->\n    %{\n      user: inputs.user,\n      profile: inputs.profile,\n      settings: inputs.settings\n    }\n  end\nend\n```","ref":"reactor-cheatsheet.html#collect-steps"},{"type":"extras","title":"Arguments and Dependencies - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#arguments-and-dependencies"},{"type":"extras","title":"Argument Sources - Reactor Cheatsheet","doc":"```elixir\n# From input\nargument :email, input(:email)\n\n# From step result\nargument :user, result(:create_user)\n\n# Static value\nargument :timeout, value(5000)\n\n# Nested value extraction\nargument :user_id, result(:create_user, [:id])\n\n# Input with path\nargument :year, input(:date, [:year])\n```","ref":"reactor-cheatsheet.html#argument-sources"},{"type":"extras","title":"Argument Transformations - Reactor Cheatsheet","doc":"```elixir\n# Inline transformation\nargument :user_id, result(:create_user) do\n  transform &(&1.id)\nend\n\n# Block form with source\nargument :age do\n  source input(:birth_year)\n  transform fn year -> \n    Date.utc_today().year - year \n  end\nend\n```","ref":"reactor-cheatsheet.html#argument-transformations"},{"type":"extras","title":"Dependencies - Reactor Cheatsheet","doc":"```elixir\n# Wait for step without using data\nwait_for :verify_email\n\n# Wait for multiple steps\nwait_for [:setup_user, :setup_profile]\n```","ref":"reactor-cheatsheet.html#dependencies"},{"type":"extras","title":"Conditional Execution - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#conditional-execution"},{"type":"extras","title":"Guards - Reactor Cheatsheet","doc":"```elixir\nstep :read_file_via_cache do\n  argument :path, input(:path)\n  run &File.read(&1.path)\n  \n  guard fn %{path: path}, %{cache: cache} ->\n    case Cache.get(cache, path) do\n      {:ok, content} -> {:halt, {:ok, content}}\n      _ -> :cont\n    end\n  end\nend\n```","ref":"reactor-cheatsheet.html#guards"},{"type":"extras","title":"Where Clauses - Reactor Cheatsheet","doc":"```elixir\nstep :conditional_step do\n  argument :user, result(:create_user)\n  \n  where fn %{user: user} ->\n    user.active? and user.plan == :premium\n  end\n  \n  run &process_premium_user/1\nend\n\n# Simple predicate\nstep :read_file do\n  argument :path, input(:path)\n  run &File.read(&1.path)\n  where &File.exists?(&1.path)\nend\n```","ref":"reactor-cheatsheet.html#where-clauses"},{"type":"extras","title":"Step Implementation - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#step-implementation"},{"type":"extras","title":"Step Module - Reactor Cheatsheet","doc":"```elixir\ndefmodule MyApp.Steps.CreateUser do\n  use Reactor.Step\n\n  @impl true\n  def run(arguments, context, options) do\n    case create_user(arguments) do\n      {:ok, user} -> {:ok, user}\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl true\n  def compensate(reason, arguments, context, options) do\n    case reason do\n      %DBConnection.ConnectionError{} -> :retry\n      _other -> :ok\n    end\n  end\n\n  @impl true\n  def undo(user, arguments, context, options) do\n    case delete_user(user) do\n      :ok -> :ok\n      {:error, reason} -> {:error, reason}\n    end\n  end\nend\n```","ref":"reactor-cheatsheet.html#step-module"},{"type":"extras","title":"Return Values - Reactor Cheatsheet","doc":"```elixir\n# Success\n{:ok, value}\n\n# Success with additional steps\n{:ok, value, [additional_step]}\n\n# Failure (triggers compensation)\n{:error, reason}\n\n# Retry step\n:retry\n{:retry, reason}\n\n# Pause reactor\n{:halt, reason}\n```","ref":"reactor-cheatsheet.html#return-values"},{"type":"extras","title":"Error Handling - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#error-handling"},{"type":"extras","title":"Compensation - Reactor Cheatsheet","doc":"```elixir\ndef compensate(reason, arguments, context, options) do\n  case reason do\n    # Retry on network errors\n    %HTTPoison.Error{reason: :timeout} -> :retry\n    %HTTPoison.Error{reason: :econnrefused} -> :retry\n    \n    # Continue with error for other failures\n    _other -> :ok\n  end\nend\n```","ref":"reactor-cheatsheet.html#compensation"},{"type":"extras","title":"Undo Operations - Reactor Cheatsheet","doc":"```elixir\ndef undo(resource, arguments, context, options) do\n  case delete_resource(resource) do\n    :ok -> :ok\n    {:error, :not_found} -> :ok  # Already deleted\n    {:error, reason} -> {:error, reason}\n  end\nend\n```","ref":"reactor-cheatsheet.html#undo-operations"},{"type":"extras","title":"Inputs and Transformations - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#inputs-and-transformations"},{"type":"extras","title":"Input Definition - Reactor Cheatsheet","doc":"```elixir\n# Basic input\ninput :name\n\n# Input with transformation\ninput :age do\n  transform &String.to_integer/1\nend\n\n# Input with description\ninput :email, description: \"User's email address\"\n```","ref":"reactor-cheatsheet.html#input-definition"},{"type":"extras","title":"Template Steps - Reactor Cheatsheet","doc":"```elixir\ntemplate :welcome_message do\n  argument :user, result(:create_user)\n  template \"\"\"\n  Welcome <%= @user.name %>! 🎉\n  Your account is now active.\n  \"\"\"\nend\n```","ref":"reactor-cheatsheet.html#template-steps"},{"type":"extras","title":"Middleware - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#middleware"},{"type":"extras","title":"Adding Middleware - Reactor Cheatsheet","doc":"```elixir\ndefmodule MyReactor do\n  use Reactor\n\n  middlewares do\n    middleware MyApp.LoggingMiddleware\n    middleware Reactor.Middleware.Telemetry\n  end\n\n  # Steps...\nend\n```","ref":"reactor-cheatsheet.html#adding-middleware"},{"type":"extras","title":"Custom Middleware - Reactor Cheatsheet","doc":"```elixir\ndefmodule MyApp.LoggingMiddleware do\n  use Reactor.Middleware\n\n  def init(context) do\n    Logger.info(\"Reactor starting\")\n    {:ok, context}\n  end\n\n  def complete(result, context) do\n    Logger.info(\"Reactor completed\")\n    {:ok, result}\n  end\n\n  def error(errors, context) do\n    Logger.error(\"Reactor failed: #{inspect(errors)}\")\n    :ok\n  end\n\n  def event({:run_start, args}, step, context) do\n    Logger.debug(\"Step #{step.name} starting\")\n  end\nend\n```","ref":"reactor-cheatsheet.html#custom-middleware"},{"type":"extras","title":"Common Patterns - Reactor Cheatsheet","doc":"{: .col-2}","ref":"reactor-cheatsheet.html#common-patterns"},{"type":"extras","title":"Data Pipeline - Reactor Cheatsheet","doc":"```elixir\ninput :raw_data\n\nstep :validate do\n  argument :data, input(:raw_data)\n  run &validate_data/1\nend\n\nstep :transform do\n  argument :data, result(:validate)\n  run &transform_data/1\nend\n\nstep :store do\n  argument :data, result(:transform)\n  run &store_data/1\nend\n\nreturn :store\n```","ref":"reactor-cheatsheet.html#data-pipeline"},{"type":"extras","title":"Parallel Processing - Reactor Cheatsheet","doc":"```elixir\nstep :fetch_user do\n  argument :id, input(:user_id)\n  run &fetch_user/1\nend\n\nstep :fetch_settings do\n  argument :id, input(:user_id)\n  run &fetch_settings/1\nend\n\ncollect :user_with_settings do\n  argument :user, result(:fetch_user)\n  argument :settings, result(:fetch_settings)\nend\n```","ref":"reactor-cheatsheet.html#parallel-processing"},{"type":"extras","title":"Error Recovery - Reactor Cheatsheet","doc":"```elixir\nstep :risky_operation do\n  run &might_fail/1\n  max_retries 3\n  \n  compensate fn reason, _, _, _ ->\n    case reason do\n      %NetworkError{} -> :retry\n      %TimeoutError{} -> :retry\n      _other -> :ok\n    end\n  end\nend\n```","ref":"reactor-cheatsheet.html#error-recovery"},{"type":"extras","title":"Quick Reference - Reactor Cheatsheet","doc":"{: .col-3}","ref":"reactor-cheatsheet.html#quick-reference"},{"type":"extras","title":"Functions - Reactor Cheatsheet","doc":"- `input/1` - Define reactor input\n- `step/2` - Define step\n- `argument/2` - Define step dependency\n- `result/1` - Reference step result\n- `value/1` - Static value\n- `wait_for/1` - Dependency without data\n- `return/1` - Set reactor return","ref":"reactor-cheatsheet.html#functions"},{"type":"extras","title":"Built-in Steps - Reactor Cheatsheet","doc":"- `step` - Basic step\n- `debug` - Log information\n- `map` - Process collections\n- `compose` - Embed reactors\n- `switch` - Conditional logic\n- `group` - Shared setup/teardown\n- `around` - Wrap execution\n- `collect` - Gather values\n- `template` - EEx templates\n- `flunk` - Force failure","ref":"reactor-cheatsheet.html#built-in-steps"},{"type":"extras","title":"Control Flow - Reactor Cheatsheet","doc":"- `guard` - Conditional execution\n- `where` - Simple conditional\n- `async?` - Control sync/async\n- `max_retries` - Retry limits\n- `batch_size` - Map processing","ref":"reactor-cheatsheet.html#control-flow"},{"type":"extras","title":"How to Build a Payment Processing Workflow","doc":"# How to Build a Payment Processing Workflow","ref":"payment-processing.html"},{"type":"extras","title":"Problem - How to Build a Payment Processing Workflow","doc":"You need to coordinate payment authorization, inventory reservation, and order fulfillment with proper rollback on failures. When any step fails, you need to clean up previous successful operations to maintain data consistency.","ref":"payment-processing.html#problem"},{"type":"extras","title":"Solution Overview - How to Build a Payment Processing Workflow","doc":"This guide shows you how to build a robust payment processing workflow using Reactor's compensation and undo patterns. You'll learn to handle failures gracefully with automatic rollback.","ref":"payment-processing.html#solution-overview"},{"type":"extras","title":"Workflow Architecture - How to Build a Payment Processing Workflow","doc":"Here's the complete payment processing flow with error handling:\n\n```mermaid\nflowchart TD\n    Start([Order Received]) --> ValidatePayment[Validate Payment Details]\n    ValidatePayment --> CheckInventory[Check Inventory]\n    ValidatePayment --> ValidateAddress[Validate Shipping Address]\n    \n    CheckInventory --> ReserveItems[Reserve Items]\n    ValidateAddress --> CalculateShipping[Calculate Shipping Cost]\n    ValidatePayment --> AuthorizePayment[Authorize Payment]\n    \n    ReserveItems --> ProcessOrder{All Valid?}\n    CalculateShipping --> ProcessOrder\n    AuthorizePayment --> ProcessOrder\n    \n    ProcessOrder -->|Success| CapturePayment[Capture Payment]\n    ProcessOrder -->|Failure| ReleaseReservation[Release Inventory]\n    ProcessOrder -->|Failure| VoidAuthorization[Void Payment Authorization]\n    \n    CapturePayment --> FulfillOrder[Fulfill Order]\n    FulfillOrder --> ShipItems[Ship Items]\n    FulfillOrder --> SendConfirmation[Send Confirmation Email]\n    \n    ShipItems --> UpdateInventory[Update Inventory]\n    ShipItems --> SendTracking[Send Tracking Information]\n    \n    SendConfirmation --> Complete([Order Complete])\n    UpdateInventory --> Complete\n    SendTracking --> Complete\n    \n    ReleaseReservation --> Failed([Order Failed])\n    VoidAuthorization --> Failed\n    \n    %% Compensation flows\n    CapturePayment -.->|Error| RefundPayment[Refund Payment]\n    FulfillOrder -.->|Error| CancelOrder[Cancel Order]\n    ShipItems -.->|Error| RecallShipment[Recall Shipment]\n    \n    style Start fill:#e1f5fe\n    style Complete fill:#c8e6c9\n    style Failed fill:#ffcdd2\n    style ProcessOrder fill:#fff3e0\n```","ref":"payment-processing.html#workflow-architecture"},{"type":"extras","title":"Prerequisites - How to Build a Payment Processing Workflow","doc":"- Understanding of Reactor basics (inputs, steps, arguments)\n- Familiarity with error handling patterns\n- Basic knowledge of payment processing concepts","ref":"payment-processing.html#prerequisites"},{"type":"extras","title":"Complete Working Example - How to Build a Payment Processing Workflow","doc":"Here's a complete e-commerce payment workflow with proper error handling:\n\n```elixir\ndefmodule ECommerce.PaymentWorkflow do\n  use Reactor\n\n  input :order_id\n  input :payment_details\n  input :customer_id\n\n  # Step 1: Validate order and check inventory\n  step :validate_order, ECommerce.Steps.ValidateOrder do\n    argument :order_id, input(:order_id)\n  end\n\n  # Step 2: Reserve inventory (can be undone)\n  step :reserve_inventory, ECommerce.Steps.ReserveInventory do\n    argument :order, result(:validate_order)\n  end\n\n  # Step 3: Authorize payment (can be undone)  \n  step :authorize_payment, ECommerce.Steps.AuthorizePayment do\n    argument :payment_details, input(:payment_details)\n    argument :order, result(:validate_order)\n    wait_for :reserve_inventory  # Ensure inventory is reserved first\n  end\n\n  # Step 4: Capture payment (compensatable)\n  step :capture_payment, ECommerce.Steps.CapturePayment do\n    argument :authorization, result(:authorize_payment)\n    argument :order, result(:validate_order)\n  end\n\n  # Step 5: Fulfill order (compensatable)\n  step :fulfill_order, ECommerce.Steps.FulfillOrder do\n    argument :order, result(:validate_order)\n    argument :inventory_reservation, result(:reserve_inventory)\n  end\n\n  # Step 6: Send confirmation (basic step, no rollback needed)\n  step :send_confirmation, ECommerce.Steps.SendConfirmation do\n    argument :order, result(:validate_order)\n    argument :customer_id, input(:customer_id)\n    argument :payment, result(:capture_payment)\n  end\n\n  return :send_confirmation\nend\n```","ref":"payment-processing.html#complete-working-example"},{"type":"extras","title":"Step Implementations - How to Build a Payment Processing Workflow","doc":"","ref":"payment-processing.html#step-implementations"},{"type":"extras","title":"1. Inventory Reservation (Undoable) - How to Build a Payment Processing Workflow","doc":"```elixir\ndefmodule ECommerce.Steps.ReserveInventory do\n  use Reactor.Step\n\n  @impl true\n  def run(%{order: order}, _context, _options) do\n    case InventoryService.reserve_items(order.items) do\n      {:ok, reservation} -> {:ok, reservation}\n      {:error, :insufficient_stock} -> {:error, \"Not enough inventory for order\"}\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl true\n  def undo(reservation, _arguments, _context, _options) do\n    # Release the reserved inventory if later steps fail\n    case InventoryService.release_reservation(reservation.id) do\n      :ok -> :ok\n      {:error, :already_released} -> :ok  # Already cleaned up\n      {:error, reason} -> {:error, reason}\n    end\n  end\nend\n```","ref":"payment-processing.html#1-inventory-reservation-undoable"},{"type":"extras","title":"2. Payment Authorization (Undoable) - How to Build a Payment Processing Workflow","doc":"```elixir\ndefmodule ECommerce.Steps.AuthorizePayment do\n  use Reactor.Step\n\n  @impl true\n  def run(%{payment_details: details, order: order}, _context, _options) do\n    case PaymentGateway.authorize(details, order.total) do\n      {:ok, authorization} -> {:ok, authorization}\n      {:error, :card_declined} -> {:error, \"Payment was declined\"}\n      {:error, :insufficient_funds} -> {:error, \"Insufficient funds\"}\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl true \n  def undo(authorization, _arguments, _context, _options) do\n    # Void the authorization if fulfillment fails\n    case PaymentGateway.void_authorization(authorization.id) do\n      :ok -> :ok\n      {:error, :already_voided} -> :ok\n      {:error, reason} -> {:error, reason}\n    end\n  end\nend\n```","ref":"payment-processing.html#2-payment-authorization-undoable"},{"type":"extras","title":"3. Payment Capture (Compensatable with Retry) - How to Build a Payment Processing Workflow","doc":"```elixir\ndefmodule ECommerce.Steps.CapturePayment do\n  use Reactor.Step\n\n  @impl true\n  def run(%{authorization: auth, order: order}, _context, _options) do\n    case PaymentGateway.capture(auth.id, order.total) do\n      {:ok, capture} -> {:ok, capture}\n      {:error, :network_timeout} -> :retry  # Retry network issues\n      {:error, :authorization_expired} -> {:error, \"Authorization expired\"}\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl true\n  def compensate(reason, %{authorization: auth}, _context, _options) do\n    case reason do\n      # Retry on transient network issues\n      %PaymentGateway.NetworkError{} -> :retry\n      %PaymentGateway.TimeoutError{} -> :retry\n      \n      # Don't retry on business logic errors\n      _other -> :ok\n    end\n  end\nend\n```","ref":"payment-processing.html#3-payment-capture-compensatable-with-retry"},{"type":"extras","title":"4. Order Fulfillment (Compensatable) - How to Build a Payment Processing Workflow","doc":"```elixir\ndefmodule ECommerce.Steps.FulfillOrder do\n  use Reactor.Step\n\n  @impl true\n  def run(%{order: order, inventory_reservation: reservation}, _context, _options) do\n    case FulfillmentService.create_shipment(order, reservation) do\n      {:ok, shipment} -> {:ok, shipment}\n      {:error, :warehouse_unavailable} -> :retry\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl true\n  def compensate(reason, _arguments, _context, _options) do\n    case reason do\n      # Retry warehouse issues\n      %FulfillmentService.WarehouseError{} -> :retry\n      \n      # Don't retry validation errors\n      _other -> :ok\n    end\n  end\nend\n```","ref":"payment-processing.html#4-order-fulfillment-compensatable"},{"type":"extras","title":"Running the Workflow - How to Build a Payment Processing Workflow","doc":"```elixir\n# Successful execution\n{:ok, confirmation} = Reactor.run(ECommerce.PaymentWorkflow,\n  order_id: \"order_123\",\n  payment_details: %{\n    card_number: \"4111111111111111\",\n    exp_month: 12,\n    exp_year: 2025,\n    cvv: \"123\"\n  },\n  customer_id: \"customer_456\"\n)\n\n# Handle failures gracefully\ncase Reactor.run(ECommerce.PaymentWorkflow, invalid_inputs) do\n  {:ok, result} -> \n    # Process successful order\n    handle_successful_order(result)\n    \n  {:error, errors} ->\n    # Log errors and notify customer\n    Logger.error(\"Payment workflow failed: #{inspect(errors)}\")\n    notify_customer_of_failure(customer_id, errors)\nend\n```","ref":"payment-processing.html#running-the-workflow"},{"type":"extras","title":"Key Patterns Explained - How to Build a Payment Processing Workflow","doc":"","ref":"payment-processing.html#key-patterns-explained"},{"type":"extras","title":"Compensation vs Undo - How to Build a Payment Processing Workflow","doc":"**Undo** is called when a step succeeded but a later step failed:\n- Use for cleaning up successful operations (release inventory, void payments)\n- Called in reverse order of step execution\n\n**Compensate** is called when the step itself failed:\n- Use for retry logic on transient failures\n- Return `:retry` to try the step again, `:ok` to continue with failure","ref":"payment-processing.html#compensation-vs-undo"},{"type":"extras","title":"Error Recovery Strategy - How to Build a Payment Processing Workflow","doc":"```elixir\ndef compensate(reason, arguments, context, options) do\n  case reason do\n    # Retry transient failures\n    %HTTPoison.Error{reason: :timeout} -> :retry\n    %HTTPoison.Error{reason: :econnrefused} -> :retry\n    %PaymentGateway.NetworkError{} -> :retry\n    \n    # Don't retry business logic failures  \n    %PaymentGateway.CardDeclinedError{} -> :ok\n    %InventoryService.OutOfStockError{} -> :ok\n    \n    # Default: don't retry unknown errors\n    _other -> :ok\n  end\nend\n```","ref":"payment-processing.html#error-recovery-strategy"},{"type":"extras","title":"Dependency Management - How to Build a Payment Processing Workflow","doc":"Use `wait_for` to control execution order without passing data:\n\n```elixir\nstep :authorize_payment do\n  argument :payment_details, input(:payment_details)\n  wait_for :reserve_inventory  # Ensure inventory reserved first\n  run &authorize_payment/1\nend\n```","ref":"payment-processing.html#dependency-management"},{"type":"extras","title":"Variations - How to Build a Payment Processing Workflow","doc":"","ref":"payment-processing.html#variations"},{"type":"extras","title":"Multi-Warehouse Fulfillment - How to Build a Payment Processing Workflow","doc":"```elixir\nmap :check_warehouses do\n  source input(:preferred_warehouses)\n  \n  step :check_availability do\n    argument :warehouse, element(:check_warehouses)\n    argument :order, result(:validate_order)\n    run &check_warehouse_inventory/1\n  end\nend\n\nstep :select_warehouse do\n  argument :availability_results, result(:check_warehouses)\n  run &select_best_warehouse/1\nend\n```","ref":"payment-processing.html#multi-warehouse-fulfillment"},{"type":"extras","title":"Multiple Payment Methods - How to Build a Payment Processing Workflow","doc":"```elixir\nswitch :payment_method do\n  on input(:payment_details, [:type])\n  \n  matches? &(&1 == :credit_card) do\n    step :process_card_payment, CreditCardProcessor do\n      argument :details, input(:payment_details)\n    end\n  end\n  \n  matches? &(&1 == :paypal) do\n    step :process_paypal_payment, PayPalProcessor do\n      argument :details, input(:payment_details)\n    end\n  end\n  \n  default do\n    flunk \"Unsupported payment method\"\n  end\nend\n```","ref":"payment-processing.html#multiple-payment-methods"},{"type":"extras","title":"Partial Fulfillment - How to Build a Payment Processing Workflow","doc":"```elixir\nstep :check_partial_fulfillment do\n  argument :order, result(:validate_order)\n  argument :inventory, result(:check_inventory)\n  \n  where fn %{inventory: inventory, order: order} ->\n    InventoryService.can_partially_fulfill?(order, inventory)\n  end\n  \n  run &create_partial_fulfillment/1\nend\n```","ref":"payment-processing.html#partial-fulfillment"},{"type":"extras","title":"Troubleshooting - How to Build a Payment Processing Workflow","doc":"","ref":"payment-processing.html#troubleshooting"},{"type":"extras","title":"Common issues - How to Build a Payment Processing Workflow","doc":"**Inventory not released after payment failure:**\n- Ensure `ReserveInventory` implements `undo/4` correctly\n- Check that `undo` handles already-released reservations gracefully\n\n**Payment captured but order not fulfilled:**\n- Implement proper compensation in fulfillment steps\n- Consider implementing refund logic for captured payments\n\nFor general error handling and retry issues, see [Error Handling Tutorial](../tutorials/02-error-handling.md#common-issues).","ref":"payment-processing.html#common-issues"},{"type":"extras","title":"Debugging Tips - How to Build a Payment Processing Workflow","doc":"Add debug steps to track workflow progress:\n\n```elixir\ndebug :log_payment_status do\n  argument :payment, result(:capture_payment)\n  argument :message, value(\"Payment captured successfully\")\nend\n```\n\nUse telemetry middleware for observability:\n\n```elixir\ndefmodule ECommerce.PaymentWorkflow do\n  use Reactor\n  \n  middlewares do\n    middleware Reactor.Middleware.Telemetry\n  end\n  \n  # ... steps\nend\n```","ref":"payment-processing.html#debugging-tips"},{"type":"extras","title":"Related Guides - How to Build a Payment Processing Workflow","doc":"- [Error Handling Tutorial](../tutorials/02-error-handling.md) - Learn compensation basics\n- [Testing Strategies](testing-strategies.md) - How to test payment workflows\n- [Performance Optimization](performance-optimization.md) - Scaling payment processing","ref":"payment-processing.html#related-guides"},{"type":"extras","title":"How to Build Data Processing Pipelines","doc":"# How to Build Data Processing Pipelines","ref":"data-pipelines.html"},{"type":"extras","title":"Problem - How to Build Data Processing Pipelines","doc":"You need to process large datasets through multiple transformation steps with error handling, progress tracking, and efficient batch processing. Your data pipeline should handle millions of records while being resilient to failures.","ref":"data-pipelines.html#problem"},{"type":"extras","title":"Solution Overview - How to Build Data Processing Pipelines","doc":"This guide shows you how to build robust ETL (Extract, Transform, Load) and batch processing workflows using Reactor's `map` steps, async processing, and collect patterns. You'll learn to process data efficiently at scale.","ref":"data-pipelines.html#solution-overview"},{"type":"extras","title":"Prerequisites - How to Build Data Processing Pipelines","doc":"- Understanding of Reactor basics (inputs, steps, arguments)\n- Familiarity with Elixir's `Enum` and `Stream` modules\n- Basic knowledge of data processing concepts","ref":"data-pipelines.html#prerequisites"},{"type":"extras","title":"Processing Large Datasets - How to Build Data Processing Pipelines","doc":"Reactor uses [Iterex](https://hexdocs.pm/iterex/) internally for efficient, resumable iteration over large datasets. This provides several advantages over standard Elixir streams:\n\n**Key Benefits:**\n- **Resumable**: Can pause and resume processing from any point\n- **Memory efficient**: Processes data in configurable chunks without loading everything into memory\n- **Lazy evaluation**: Only processes data as needed, not all at once\n- **Resource management**: Proper cleanup of external resources (files, database connections)\n\n**When you need these patterns:**\n- Processing files larger than available memory (> 1GB)\n- Long-running ETL jobs that may need to pause/resume\n- Data pipelines requiring checkpoint/recovery capabilities\n- Streaming data from external APIs with rate limits","ref":"data-pipelines.html#processing-large-datasets"},{"type":"extras","title":"Complete ETL Pipeline Example - How to Build Data Processing Pipelines","doc":"Here's a complete data processing pipeline that extracts user data, transforms it, and loads it into multiple destinations:\n\n```elixir\ndefmodule DataPipeline.UserETL do\n  use Reactor\n\n  input :source_file\n  input :output_destinations\n\n  # Step 1: Extract - Read and parse data\n  step :extract_data, DataPipeline.Steps.ExtractCSV do\n    argument :file_path, input(:source_file)\n  end\n\n  # Step 2: Validate data quality\n  step :validate_data_quality, DataPipeline.Steps.DataQualityCheck do\n    argument :raw_data, result(:extract_data)\n  end\n\n  # Step 3: Transform users in batches\n  map :transform_users do\n    source result(:extract_data, [:users])\n    allow_async? true\n    return :validate_user\n\n    step :clean_user do\n      argument :user, element(:transform_users)\n      run fn %{user: user} ->\n        clean_and_normalize_user(%{user: user, rules: %{}})\n      end\n    end\n\n    step :enrich_user do\n      argument :clean_user, result(:clean_user)\n      run &enrich_with_external_data/1\n    end\n\n    step :validate_user do\n      argument :user, result(:enrich_user)\n      run &validate_business_rules/1\n    end\n  end\n\n  # Step 4: Collect transformation results\n  collect :process_results do\n    argument :transformed_users, result(:transform_users)\n    argument :source_stats, result(:extract_data, [:stats])\n    \n    transform fn %{transformed_users: users, source_stats: stats} ->\n      successful_users = Enum.filter(users, &match?({:ok, _}, &1))\n      failed_users = Enum.filter(users, &match?({:error, _}, &1))\n      \n      %{\n        successful: Enum.map(successful_users, &elem(&1, 1)),\n        failed: failed_users,\n        source_count: stats.total_count,\n        success_rate: length(successful_users) / length(users)\n      }\n    end\n  end\n\n  # Step 5: Load to multiple destinations in parallel\n  step :load_to_database, DataPipeline.Steps.LoadToDatabase do\n    argument :users, result(:process_results, [:successful])\n    async? true\n  end\n\n  step :load_to_search_index, DataPipeline.Steps.LoadToElasticsearch do\n    argument :users, result(:process_results, [:successful])\n    async? true\n  end\n\n  step :generate_report, DataPipeline.Steps.GenerateProcessingReport do\n    argument :results, result(:process_results)\n    wait_for [:load_to_database, :load_to_search_index]\n  end\n\n  return :generate_report\nend\n```","ref":"data-pipelines.html#complete-etl-pipeline-example"},{"type":"extras","title":"Step Implementations - How to Build Data Processing Pipelines","doc":"","ref":"data-pipelines.html#step-implementations"},{"type":"extras","title":"1. Data Extraction - How to Build Data Processing Pipelines","doc":"```elixir\ndefmodule DataPipeline.Steps.ExtractCSV do\n  use Reactor.Step\n\n  @impl true\n  def run(%{file_path: path}, _context, _options) do\n    case File.exists?(path) do\n      true ->\n        users = \n          path\n          |> File.stream!()\n          |> CSV.decode!(headers: true)\n          |> Enum.to_list()\n        \n        stats = %{\n          total_count: length(users),\n          file_size: File.stat!(path).size,\n          extracted_at: DateTime.utc_now()\n        }\n        \n        {:ok, %{users: users, stats: stats}}\n        \n      false ->\n        {:error, \"Source file not found: #{path}\"}\n    end\n  end\nend\n```","ref":"data-pipelines.html#1-data-extraction"},{"type":"extras","title":"2. Data Quality Validation - How to Build Data Processing Pipelines","doc":"```elixir\ndefmodule DataPipeline.Steps.DataQualityCheck do\n  use Reactor.Step\n\n  @impl true\n  def run(%{raw_data: %{users: users}}, _context, _options) do\n    # Analyze data quality\n    quality_issues = analyze_quality(users)\n    \n    rules = %{\n      email_required: true,\n      phone_format: ~r/^\\+?[\\d\\s\\-\\(\\)]+$/,\n      name_min_length: 2,\n      max_age: 120\n    }\n    \n    case quality_issues do\n      [] -> \n        {:ok, %{rules: rules, issues: [], status: :passed}}\n      issues when length(issues) < 100 ->\n        {:ok, %{rules: rules, issues: issues, status: :warnings}}\n      issues ->\n        {:error, \"Too many quality issues: #{length(issues)} problems found\"}\n    end\n  end\n\n  defp analyze_quality(users) do\n    users\n    |> Enum.with_index()\n    |> Enum.flat_map(fn {user, index} ->\n      check_user_quality(user, index)\n    end)\n  end\n\n  defp check_user_quality(user, index) do\n    issues = []\n    \n    issues = if is_nil(user[\"email\"]) or user[\"email\"] == \"\" do\n      [\"Row #{index + 1}: Missing email\" | issues]\n    else\n      issues\n    end\n    \n    issues = if is_nil(user[\"name\"]) or String.length(user[\"name\"]) < 2 do\n      [\"Row #{index + 1}: Invalid name\" | issues]\n    else\n      issues\n    end\n    \n    issues\n  end\nend\n```","ref":"data-pipelines.html#2-data-quality-validation"},{"type":"extras","title":"3. User Transformation Functions - How to Build Data Processing Pipelines","doc":"```elixir\ndef clean_and_normalize_user(%{user: user, rules: _rules}) do\n  cleaned = %{\n    id: user[\"id\"],\n    email: String.downcase(String.trim(user[\"email\"] || \"\")),\n    name: String.trim(user[\"name\"] || \"\"),\n    phone: normalize_phone(user[\"phone\"]),\n    age: parse_age(user[\"age\"]),\n    created_at: DateTime.utc_now()\n  }\n  \n  {:ok, cleaned}\nrescue\n  e -> {:error, \"Failed to clean user #{user[\"id\"]}: #{inspect(e)}\"}\nend\n\ndef enrich_with_external_data(%{clean_user: user}) do\n  case ExternalAPI.get_user_profile(user.email) do\n    {:ok, profile} ->\n      enriched = Map.merge(user, %{\n        company: profile.company,\n        location: profile.location,\n        verified: profile.verified\n      })\n      {:ok, enriched}\n      \n    {:error, :not_found} ->\n      # Continue without enrichment\n      {:ok, Map.put(user, :verified, false)}\n      \n    {:error, reason} ->\n      {:error, \"Enrichment failed for #{user.email}: #{reason}\"}\n  end\nend\n\ndef validate_business_rules(%{user: user}) do\n  errors = []\n  \n  errors = if String.length(user.name) < 2 do\n    [\"Name too short\" | errors]\n  else\n    errors\n  end\n  \n  errors = if not String.contains?(user.email, \"@\") do\n    [\"Invalid email format\" | errors]\n  else\n    errors\n  end\n  \n  errors = if user.age && user.age > 120 do\n    [\"Unrealistic age\" | errors]\n  else\n    errors\n  end\n  \n  case errors do\n    [] -> {:ok, user}\n    errors -> {:error, \"Validation failed: #{Enum.join(errors, \", \")}\"}\n  end\nend\n```","ref":"data-pipelines.html#3-user-transformation-functions"},{"type":"extras","title":"4. Data Loading Steps - How to Build Data Processing Pipelines","doc":"```elixir\ndefmodule DataPipeline.Steps.LoadToDatabase do\n  use Reactor.Step\n\n  @impl true\n  def run(%{users: users}, _context, _options) do\n    batches = Enum.chunk_every(users, 1000)\n    \n    results = Enum.map(batches, fn batch ->\n      case MyApp.Repo.insert_all(\"users\", batch, \n           on_conflict: :replace_all,\n           conflict_target: [:email]) do\n        {count, _} -> {:ok, count}\n        error -> {:error, error}\n      end\n    end)\n    \n    total_inserted = results\n    |> Enum.filter(&match?({:ok, _}, &1))\n    |> Enum.map(&elem(&1, 1))\n    |> Enum.sum()\n    \n    {:ok, %{inserted: total_inserted, total_batches: length(batches)}}\n  end\n\n  @impl true\n  def compensate(_reason, %{users: users}, _context, _options) do\n    # Cleanup on failure - remove any partially inserted data\n    user_emails = Enum.map(users, & &1.email)\n    MyApp.Repo.delete_all(from u in \"users\", where: u.email in ^user_emails)\n    :ok\n  end\nend\n\ndefmodule DataPipeline.Steps.LoadToElasticsearch do\n  use Reactor.Step\n\n  @impl true\n  def run(%{users: users}, _context, _options) do\n    bulk_requests = Enum.map(users, fn user ->\n      %{\n        index: %{\n          _index: \"users\",\n          _id: user.id,\n          _source: user\n        }\n      }\n    end)\n    \n    case Elasticsearch.bulk_request(bulk_requests) do\n      {:ok, response} ->\n        indexed = response[\"items\"] |> length()\n        {:ok, %{indexed: indexed}}\n        \n      {:error, reason} ->\n        {:error, \"Elasticsearch indexing failed: #{reason}\"}\n    end\n  end\nend\n```","ref":"data-pipelines.html#4-data-loading-steps"},{"type":"extras","title":"Running the Pipeline - How to Build Data Processing Pipelines","doc":"```elixir\n# Process a large CSV file\n{:ok, report} = Reactor.run(DataPipeline.UserETL, %{\n  source_file: \"/data/users_export.csv\",\n  output_destinations: [:database, :elasticsearch]\n})\n\nIO.puts(\"Processing completed!\")\nIO.puts(\"Success rate: #{report.success_rate * 100}%\")\nIO.puts(\"Total processed: #{report.source_count}\")\n```","ref":"data-pipelines.html#running-the-pipeline"},{"type":"extras","title":"Advanced Patterns - How to Build Data Processing Pipelines","doc":"","ref":"data-pipelines.html#advanced-patterns"},{"type":"extras","title":"Streaming Large Files - How to Build Data Processing Pipelines","doc":"For very large files, process data in chunks to avoid memory issues and enable efficient error handling:\n\n```elixir\nstep :extract_streaming_data do\n  run fn %{file_path: path} ->\n    iter = \n      path\n      |> File.stream!()\n      |> CSV.decode!(headers: true)\n      |> Iter.from()\n    \n    {:ok, iter}\n  end\nend\n\nmap :process_streaming_data do\n  source result(:extract_streaming_data)\n  allow_async? true\n  return :process_chunk\n\n  compose :process_chunk, ChunkProcessor do\n    argument :chunk_data, element(:process_streaming_data)\n  end\nend\n\ndefmodule ChunkProcessor do\n  use Reactor\n\n  input :chunk_data\n\n  step :validate_chunk do\n    argument :data, input(:chunk_data)\n    run &validate_chunk_data/1\n  end\n\n  step :transform_records do\n    argument :data, result(:validate_chunk)\n    run fn %{data: records} ->\n      processed = Enum.map(records, &transform_record/1)\n      {:ok, processed}\n    end\n  end\n\n  step :save_checkpoint do\n    argument :original, input(:chunk_data)\n    argument :processed, result(:transform_records)\n    \n    run fn %{original: orig, processed: proc} ->\n      save_chunk_checkpoint(orig, proc)\n      {:ok, proc}\n    end\n    \n    compensate fn _reason, %{original: chunk_data}, _context, _options ->\n      log_failed_chunk(chunk_data)\n      :retry\n    end\n  end\n\n  return :save_checkpoint\nend\n```","ref":"data-pipelines.html#streaming-large-files"},{"type":"extras","title":"Resumable Data Processing - How to Build Data Processing Pipelines","doc":"Pause and resume processing across reactor executions:\n\n```elixir\nstep :prepare_processing_iter do\n  argument :source_iter, result(:extract_streaming_data)\n  \n  run fn %{source_iter: iter} ->\n    if File.exists?(\"/tmp/processing_checkpoint.json\") do\n      # Resume from checkpoint\n      checkpoint = load_checkpoint()\n      remaining_iter = Iter.drop(iter, checkpoint.processed_count)\n      {:ok, remaining_iter}\n    else\n      # Start fresh processing\n      {:ok, iter}\n    end\n  end\nend\n```","ref":"data-pipelines.html#resumable-data-processing"},{"type":"extras","title":"Parallel Processing with Error Handling - How to Build Data Processing Pipelines","doc":"```elixir\nmap :process_files do\n  source input(:file_list)\n  allow_async? true\n\n  step :process_file do\n    argument :file_path, element(:process_files)\n    max_retries 3\n    \n    run fn %{file_path: path} ->\n      case process_single_file(path) do\n        {:ok, result} -> {:ok, result}\n        {:error, :temp_failure} -> :retry\n        {:error, reason} -> {:error, reason}\n      end\n    end\n    \n    compensate fn reason, %{file_path: path}, _, _ ->\n      case reason do\n        %File.Error{} -> :retry\n        _other -> :ok\n      end\n    end\n  end\nend\n```","ref":"data-pipelines.html#parallel-processing-with-error-handling"},{"type":"extras","title":"Data Aggregation - How to Build Data Processing Pipelines","doc":"```elixir\nmap :calculate_totals do\n  source input(:sales_data)\n  \n  step :sum_by_region do\n    argument :region_data, element(:calculate_totals)\n    \n    run fn %{region_data: data} ->\n      total = Enum.sum(Enum.map(data, & &1.amount))\n      {:ok, %{region: data.region, total: total}}\n    end\n  end\nend\n\ncollect :overall_totals do\n  argument :region_totals, result(:calculate_totals)\n  \n  transform fn %{region_totals: totals} ->\n    grand_total = Enum.sum(Enum.map(totals, & &1.total))\n    %{grand_total: grand_total, by_region: totals}\n  end\nend\n```","ref":"data-pipelines.html#data-aggregation"},{"type":"extras","title":"Monitoring and Observability - How to Build Data Processing Pipelines","doc":"","ref":"data-pipelines.html#monitoring-and-observability"},{"type":"extras","title":"Progress Tracking - How to Build Data Processing Pipelines","doc":"```elixir\ndebug :log_progress do\n  argument :batch_results, result(:transform_users)\n  argument :message, value(\"Batch processing completed\")\nend\n\n# Or use a regular step for custom progress tracking\nstep :track_progress do\n  argument :batch_results, result(:transform_users)\n  \n  run fn %{batch_results: results} ->\n    completed = length(results)\n    successful = Enum.count(results, &match?({:ok, _}, &1))\n    \n    IO.puts(\"Processed #{completed} records, #{successful} successful\")\n    {:ok, %{completed: completed, successful: successful}}\n  end\nend\n```","ref":"data-pipelines.html#progress-tracking"},{"type":"extras","title":"Telemetry Integration - How to Build Data Processing Pipelines","doc":"Reactor provides observability using the conventional [telemetry](https://hex.pm/packages/telemetry) package. Add the telemetry middleware to emit events for monitoring your data pipelines:\n\n```elixir\ndefmodule DataPipeline.UserETL do\n  use Reactor\n  \n  middlewares do\n    middleware Reactor.Middleware.Telemetry\n  end\n  \n  # Steps...\nend\n\n# In your application\n:telemetry.attach(\"data-pipeline-handler\", \n  [:reactor, :step, :stop], \n  &DataPipeline.Telemetry.handle_event/4, \n  %{}\n)\n```","ref":"data-pipelines.html#telemetry-integration"},{"type":"extras","title":"Troubleshooting - How to Build Data Processing Pipelines","doc":"","ref":"data-pipelines.html#troubleshooting"},{"type":"extras","title":"Common Issues - How to Build Data Processing Pipelines","doc":"**Data inconsistency:**\n- Implement proper compensation for database operations\n- Use database transactions where appropriate\n- Add data validation at multiple stages\n\nFor memory and performance issues, see [Performance Optimization](performance-optimization.md#troubleshooting).","ref":"data-pipelines.html#common-issues"},{"type":"extras","title":"Debugging Tips - How to Build Data Processing Pipelines","doc":"Add debug steps to monitor data flow:\n\n```elixir\ndebug :inspect_batch do\n  argument :batch, element(:transform_users)\n  argument :label, value(\"Processing batch\")\nend\n```","ref":"data-pipelines.html#debugging-tips"},{"type":"extras","title":"Related Guides - How to Build Data Processing Pipelines","doc":"- [Performance Optimization](performance-optimization.md) - Scaling techniques\n- [Testing Strategies](testing-strategies.md) - Testing data pipelines  \n- [Async Workflows Tutorial](../tutorials/03-async-workflows.md) - Concurrency basics","ref":"data-pipelines.html#related-guides"},{"type":"extras","title":"How to Orchestrate HTTP APIs with Reactor","doc":"# How to Orchestrate HTTP APIs with Reactor","ref":"api-orchestration.html"},{"type":"extras","title":"Problem - How to Orchestrate HTTP APIs with Reactor","doc":"You need to integrate with multiple HTTP APIs in production workflows, handling authentication, rate limits, circuit breakers, API versioning, and service discovery patterns.","ref":"api-orchestration.html#problem"},{"type":"extras","title":"Solution Overview - How to Orchestrate HTTP APIs with Reactor","doc":"This guide shows you how to build production-ready API orchestration using `reactor_req`, covering real-world concerns like authentication management, circuit breakers, rate limiting, and service resilience patterns.","ref":"api-orchestration.html#solution-overview"},{"type":"extras","title":"Prerequisites - How to Orchestrate HTTP APIs with Reactor","doc":"- Understanding of Reactor basics (inputs, steps, arguments)\n- Familiarity with error handling and compensation patterns\n- Basic knowledge of HTTP APIs and the `Req` HTTP client","ref":"api-orchestration.html#prerequisites"},{"type":"extras","title":"Setup - How to Orchestrate HTTP APIs with Reactor","doc":"Add `reactor_req` to your dependencies:\n\n```elixir\n# mix.exs\ndef deps do\n  [\n    {:reactor, \"~> 0.15\"},\n    {:reactor_req, \"~> 0.1\"},\n    {:req, \"~> 0.5\"}\n  ]\nend\n```","ref":"api-orchestration.html#setup"},{"type":"extras","title":"HTTP Client Integration with Reactor - How to Orchestrate HTTP APIs with Reactor","doc":"The `reactor_req` package provides direct integration between Reactor and the Req HTTP client. Create `lib/api_client.ex`:\n\n```elixir\ndefmodule ApiClient do\n  use Reactor\n\n  input :base_url\n  input :user_id\n\n  req_new :setup_client do\n    base_url input(:base_url)\n    headers value([{\"user-agent\", \"MyApp/1.0\"}])\n    retry value(:transient)\n    retry_delay value(fn attempt -> 200 * attempt end)\n  end\n\n  template :build_profile_url do\n    argument :user_id, input(:user_id)\n    template \"/users/<%= @user_id %>\"\n  end\n\n  template :build_preferences_url do\n    argument :user_id, input(:user_id)\n    template \"/users/<%= @user_id %>/preferences\"\n  end\n\n  req_get :fetch_profile do\n    request result(:setup_client)\n    url result(:build_profile_url)\n    headers value(%{\"accept\" => \"application/json\"})\n  end\n\n  req_get :fetch_preferences do\n    request result(:setup_client)\n    url result(:build_preferences_url)\n  end\n\n  step :combine_data do\n    argument :profile, result(:fetch_profile, [:body])\n    argument :preferences, result(:fetch_preferences, [:body])\n\n    run fn %{profile: profile, preferences: prefs}, _context ->\n      {:ok, %{profile: profile, preferences: prefs}}\n    end\n  end\n\n  return :combine_data\nend\n```","ref":"api-orchestration.html#http-client-integration-with-reactor"},{"type":"extras","title":"Authentication Management - How to Orchestrate HTTP APIs with Reactor","doc":"Handle API authentication and token refresh patterns:\n\n```elixir\ndefmodule AuthenticatedApiClient do\n  use Reactor\n\n  input :client_id\n  input :client_secret\n  input :api_endpoint\n\n  step :build_oauth_payload do\n    argument :client_id, input(:client_id)\n    argument :client_secret, input(:client_secret)\n    \n    run fn %{client_id: client_id, client_secret: client_secret}, _context ->\n      payload = %{\n        grant_type: \"client_credentials\",\n        client_id: client_id,\n        client_secret: client_secret\n      }\n      {:ok, payload}\n    end\n  end\n\n  req_post :get_auth_token do\n    url value(\"https://auth.example.com/oauth/token\")\n    json result(:build_oauth_payload)\n  end\n\n  step :extract_access_token do\n    argument :token_response, result(:get_auth_token, [:body])\n    \n    run fn %{token_response: resp}, _context ->\n      {:ok, resp[\"access_token\"]}\n    end\n  end\n\n  template :build_auth_header do\n    argument :token, result(:extract_access_token)\n    template \"Bearer <%= @token %>\"\n  end\n\n  req_new :prepare_authenticated_client do\n    base_url input(:api_endpoint)\n    headers [{\"authorization\", result(:build_auth_header)}]\n  end\n\n  req_get :fetch_protected_data do\n    request result(:prepare_authenticated_client)\n    url value(\"/protected/data\")\n  end\n\n  return :fetch_protected_data\nend\n```","ref":"api-orchestration.html#authentication-management"},{"type":"extras","title":"API Versioning - How to Orchestrate HTTP APIs with Reactor","doc":"Handle different API versions by composing version-specific reactors:\n\n```elixir\ndefmodule UserApiV1 do\n  use Reactor\n\n  input :user_id\n\n  req_new :client do\n    base_url value(\"https://api.example.com/v1\")\n  end\n\n  template :build_user_path do\n    argument :user_id, input(:user_id)\n    template \"/users/<%= @user_id %>\"\n  end\n\n  req_get :fetch_user do\n    request result(:client)\n    url result(:build_user_path)\n  end\n\n  step :normalize_response do\n    argument :response, result(:fetch_user, [:body])\n    \n    run fn %{response: resp}, _context ->\n      normalized = %{\n        id: resp[\"user_id\"],\n        name: resp[\"full_name\"],\n        email: resp[\"email_address\"]\n      }\n      {:ok, normalized}\n    end\n  end\n\n  return :normalize_response\nend\n\ndefmodule UserApiV2 do\n  use Reactor\n\n  input :user_id\n\n  req_new :client do\n    base_url value(\"https://api.example.com/v2\")\n  end\n\n  template :build_user_path do\n    argument :user_id, input(:user_id)\n    template \"/users/<%= @user_id %>\"\n  end\n\n  req_get :fetch_user do\n    request result(:client)\n    url result(:build_user_path)\n  end\n\n  return :fetch_user\nend\n\ndefmodule VersionedUserApi do\n  use Reactor\n\n  input :api_version, default: \"v1\"\n  input :user_id\n\n  switch :fetch_user_by_version do\n    on input(:api_version)\n\n    match \"v1\" do\n      compose :get_user, UserApiV1 do\n        argument :user_id, input(:user_id)\n      end\n      return :get_user\n    end\n\n    match \"v2\" do\n      compose :get_user, UserApiV2 do\n        argument :user_id, input(:user_id)\n      end\n      return :get_user\n    end\n\n    default do\n      flunk :unsupported_version do\n        argument :version, input(:api_version)\n        message \"Unsupported API version: <%= @version %>\"\n      end\n    end\n  end\n\n  return :fetch_user_by_version\nend\n```","ref":"api-orchestration.html#api-versioning"},{"type":"extras","title":"Testing API Integration - How to Orchestrate HTTP APIs with Reactor","doc":"Test your API orchestration patterns:\n\n```bash\niex -S mix\n```\n\n```elixir\n# Test basic HTTP client integration\n{:ok, result} = Reactor.run(ApiClient, %{\n  base_url: \"https://jsonplaceholder.typicode.com\", \n  user_id: \"1\"\n})\n\n# Test rate limiting\nrequests = [%{id: 1}, %{id: 2}, %{id: 3}]\n{:ok, results} = Reactor.run(RateLimitedApi, %{requests: requests})\n\n# Test versioning\n{:ok, normalized} = Reactor.run(VersionedApi, %{\n  api_version: \"v2\",\n  user_id: \"123\"\n})\n```","ref":"api-orchestration.html#testing-api-integration"},{"type":"extras","title":"Related Guides - How to Orchestrate HTTP APIs with Reactor","doc":"- [Error Handling Tutorial](../tutorials/02-error-handling.md) - Compensation and undo patterns\n- [Async Workflows Tutorial](../tutorials/03-async-workflows.md) - Concurrent processing\n- [Performance Optimization](performance-optimization.md) - Scaling and monitoring","ref":"api-orchestration.html#related-guides"},{"type":"extras","title":"How to Test Reactors and Steps","doc":"# How to Test Reactors and Steps","ref":"testing-strategies.html"},{"type":"extras","title":"Problem - How to Test Reactors and Steps","doc":"You need comprehensive testing strategies for your reactors, including unit tests for individual steps, integration tests for complete workflows, and proper testing of error handling and compensation logic.","ref":"testing-strategies.html#problem"},{"type":"extras","title":"Solution Overview - How to Test Reactors and Steps","doc":"This guide shows you different approaches to testing reactors, from testing individual step modules to full workflow integration tests. We'll cover unit testing, integration testing, error scenario testing, and concurrent execution testing.","ref":"testing-strategies.html#solution-overview"},{"type":"extras","title":"Prerequisites - How to Test Reactors and Steps","doc":"- Understanding of Reactor basics\n- Familiarity with ExUnit testing\n- Basic knowledge of mocking with Mimic (for advanced scenarios)","ref":"testing-strategies.html#prerequisites"},{"type":"extras","title":"Testing Strategies - How to Test Reactors and Steps","doc":"","ref":"testing-strategies.html#testing-strategies"},{"type":"extras","title":"1. Unit Testing Individual Steps - How to Test Reactors and Steps","doc":"The most granular level is testing individual step modules directly.\n\n#### Basic Step Testing\n\n```elixir\ndefmodule MyApp.Steps.ValidateEmailTest do\n  use ExUnit.Case, async: true\n  \n  alias MyApp.Steps.ValidateEmail\n  \n  test \"validates correct email format\" do\n    arguments = %{email: \"user@example.com\"}\n    context = %{}\n    options = []\n    \n    assert {:ok, \"user@example.com\"} = ValidateEmail.run(arguments, context, options)\n  end\n  \n  test \"returns error for invalid email\" do\n    arguments = %{email: \"invalid-email\"}\n    context = %{}\n    options = []\n    \n    assert {:error, %ArgumentError{}} = ValidateEmail.run(arguments, context, options)\n  end\nend\n```\n\n#### Testing Anonymous Function Steps\n\nInstead of testing anonymous functions inline, extract them to public functions that can be unit tested:\n\n```elixir\ndefmodule MyApp.UserReactor do\n  use Reactor\n  \n  input :name\n  \n  step :greet do\n    argument :name, input(:name)\n    run &greet_user/2\n  end\n  \n  def greet_user(%{name: name}, _context) do\n    {:ok, \"Hello, #{name}!\"}\n  end\nend\n\ndefmodule MyApp.UserReactorTest do\n  use ExUnit.Case, async: true\n  \n  test \"greet_user formats greeting correctly\" do\n    assert {:ok, \"Hello, Marty!\"} = \n      MyApp.UserReactor.greet_user(%{name: \"Marty\"}, %{})\n  end\n  \n  test \"greet_user handles edge cases\" do\n    assert {:ok, \"Hello, !\"} = \n      MyApp.UserReactor.greet_user(%{name: \"\"}, %{})\n  end\nend\n```","ref":"testing-strategies.html#1-unit-testing-individual-steps"},{"type":"extras","title":"2. Integration Testing Complete Reactors - How to Test Reactors and Steps","doc":"Test entire workflows by running complete reactors.\n\n#### Module-Based Reactor Testing\n\nTest your actual reactor modules directly. When testing reactors that interact with databases or other shared resources, you'll typically want to disable async execution to ensure proper isolation with test sandboxes:\n\n```elixir\ndefmodule MyApp.UserRegistrationReactor do\n  use Reactor\n  \n  input :email\n  input :password\n  \n  step :validate_email, MyApp.Steps.ValidateEmail do\n    argument :email, input(:email)\n  end\n  \n  step :hash_password, MyApp.Steps.HashPassword do\n    argument :password, input(:password)\n  end\n  \n  step :create_user, MyApp.Steps.CreateUser do\n    argument :email, result(:validate_email)\n    argument :password_hash, result(:hash_password)\n  end\nend\n\ndefmodule MyApp.UserRegistrationReactorTest do\n  use ExUnit.Case, async: false\n  \n  alias MyApp.UserRegistrationReactor\n  \n  test \"successful user registration flow\" do\n    inputs = %{\n      email: \"user@example.com\",\n      password: \"secure_password\"\n    }\n    \n    assert {:ok, %{id: user_id}} = Reactor.run(UserRegistrationReactor, inputs, async?: false)\n    assert is_binary(user_id)\n  end\n  \n  test \"handles invalid email gracefully\" do\n    inputs = %{\n      email: \"invalid-email\",\n      password: \"secure_password\"\n    }\n    \n    assert {:error, _reason} = Reactor.run(UserRegistrationReactor, inputs, async?: false)\n  end\nend\n```","ref":"testing-strategies.html#2-integration-testing-complete-reactors"},{"type":"extras","title":"3. Testing Error Handling and Compensation - How to Test Reactors and Steps","doc":"Test how your reactors handle failures and compensation using Mimic to control step behavior.\n\n#### Setting Up Mimic for Step Mocking\n\nWe recommend using [Mimic](https://hex.pm/packages/mimic) as your mocking library for testing Reactor steps. Mimic allows you to stub function calls without modifying your production code.\n\nFirst, set up your test helper to copy the step modules you want to mock:\n\n```elixir\n# test/test_helper.exs\nMimic.copy(MyApp.Steps.ProcessPayment)\nMimic.copy(MyApp.Steps.ReserveInventory)\nMimic.copy(MyApp.Steps.SendConfirmation)\nExUnit.start()\n```\n\n#### Testing Error Scenarios\n\n```elixir\ndefmodule MyApp.PaymentReactor do\n  use Reactor\n  \n  input :payment_data\n  input :items\n  \n  step :process_payment, MyApp.Steps.ProcessPayment do\n    argument :payment_data, input(:payment_data)\n  end\n  \n  step :reserve_inventory, MyApp.Steps.ReserveInventory do\n    argument :items, input(:items)\n    argument :payment_id, result(:process_payment, [:id])\n  end\n  \n  step :send_confirmation, MyApp.Steps.SendConfirmation do\n    argument :payment_id, result(:process_payment, [:id])\n  end\nend\n\ndefmodule MyApp.PaymentReactorTest do\n  use ExUnit.Case, async: false\n  use Mimic\n  \n  alias MyApp.PaymentReactor\n  \n  test \"successful payment flow\" do\n    MyApp.Steps.ProcessPayment\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, %{id: \"payment_123\", status: :completed}}\n    end)\n    \n    MyApp.Steps.ReserveInventory\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, %{reservation_id: \"res_456\"}}\n    end)\n    \n    MyApp.Steps.SendConfirmation\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, :sent}\n    end)\n    \n    inputs = %{payment_data: %{amount: 100}, items: [%{id: 1}]}\n    assert {:ok, :sent} = Reactor.run(PaymentReactor, inputs, async?: false)\n  end\n  \n  test \"payment failure with no compensation needed\" do\n    MyApp.Steps.ProcessPayment\n    |> stub(:run, fn _args, _context, _opts -> \n      {:error, :insufficient_funds}\n    end)\n    \n    inputs = %{payment_data: %{amount: 100}, items: [%{id: 1}]}\n    assert {:error, _reason} = Reactor.run(PaymentReactor, inputs, async?: false)\n  end\n  \n  test \"inventory failure triggers payment compensation\" do\n    MyApp.Steps.ProcessPayment\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, %{id: \"payment_123\", status: :completed}}\n    end)\n    |> expect(:compensate, fn _reason, _args, _context, _opts ->\n      :ok\n    end)\n    \n    MyApp.Steps.ReserveInventory\n    |> stub(:run, fn _args, _context, _opts -> \n      {:error, :out_of_stock}\n    end)\n    \n    inputs = %{payment_data: %{amount: 100}, items: [%{id: 1}]}\n    assert {:error, _reason} = Reactor.run(PaymentReactor, inputs, async?: false)\n  end\n  \n  test \"compensation returns continue value\" do\n    MyApp.Steps.ProcessPayment\n    |> stub(:run, fn _args, _context, _opts -> \n      {:error, :temporary_failure}\n    end)\n    |> stub(:compensate, fn _reason, _args, _context, _opts ->\n      {:continue, %{id: \"fallback_payment\", status: :manual_review}}\n    end)\n    \n    MyApp.Steps.ReserveInventory\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, %{reservation_id: \"res_456\"}}\n    end)\n    \n    MyApp.Steps.SendConfirmation\n    |> stub(:run, fn _args, _context, _opts -> \n      {:ok, :sent}\n    end)\n    \n    inputs = %{payment_data: %{amount: 100}, items: [%{id: 1}]}\n    assert {:ok, :sent} = Reactor.run(PaymentReactor, inputs, async?: false)\n  end\nend\n```\n\n#### Testing Undo Logic\n\n```elixir\ntest \"later step failure triggers undo of earlier steps\" do\n  MyApp.Steps.ProcessPayment\n  |> stub(:run, fn _args, _context, _opts -> \n    {:ok, %{id: \"payment_123\"}}\n  end)\n  |> expect(:undo, fn _result, _args, _context, _opts ->\n    :ok\n  end)\n  \n  MyApp.Steps.ReserveInventory\n  |> stub(:run, fn _args, _context, _opts -> \n    {:ok, %{reservation_id: \"res_456\"}}\n  end)\n  \n  MyApp.Steps.SendConfirmation\n  |> stub(:run, fn _args, _context, _opts -> \n    {:error, :email_service_down}\n  end)\n  \n  inputs = %{payment_data: %{amount: 100}, items: [%{id: 1}]}\n  assert {:error, _reason} = Reactor.run(PaymentReactor, inputs, async?: false)\nend\n```","ref":"testing-strategies.html#3-testing-error-handling-and-compensation"},{"type":"extras","title":"Best Practices - How to Test Reactors and Steps","doc":"","ref":"testing-strategies.html#best-practices"},{"type":"extras","title":"Test Organization - How to Test Reactors and Steps","doc":"1. **Unit Tests**: Create one test file per step module\n2. **Integration Tests**: Group related workflow tests together\n3. **Use async: true**: For pure unit tests that don't depend on external state\n4. **Use async: false**: For integration tests that need deterministic execution","ref":"testing-strategies.html#test-organization"},{"type":"extras","title":"Related Guides - How to Test Reactors and Steps","doc":"- [Debugging Workflows](debugging-workflows.md) - Troubleshooting techniques\n- [Error Handling Tutorial](../tutorials/02-error-handling.md) - Learn compensation patterns\n- [Async Workflows Tutorial](../tutorials/03-async-workflows.md) - Understanding concurrency\n\nThis comprehensive testing approach ensures your reactors are reliable, maintainable, and perform well under various conditions.","ref":"testing-strategies.html#related-guides"},{"type":"extras","title":"How to Debug Reactor Workflows","doc":"# How to Debug Reactor Workflows","ref":"debugging-workflows.html"},{"type":"extras","title":"Problem - How to Debug Reactor Workflows","doc":"Your reactor isn't behaving as expected and you need effective techniques to troubleshoot and debug workflow execution, identify bottlenecks, and understand what's happening during execution.","ref":"debugging-workflows.html#problem"},{"type":"extras","title":"Solution Overview - How to Debug Reactor Workflows","doc":"This guide shows you debugging techniques, tools, and patterns for identifying and fixing issues in your reactors using Reactor's built-in debugging features, telemetry, visualization tools, and error analysis.","ref":"debugging-workflows.html#solution-overview"},{"type":"extras","title":"Prerequisites - How to Debug Reactor Workflows","doc":"- Understanding of Reactor basics\n- Experience building reactors\n- Basic knowledge of Elixir debugging techniques","ref":"debugging-workflows.html#prerequisites"},{"type":"extras","title":"Debugging Strategies - How to Debug Reactor Workflows","doc":"","ref":"debugging-workflows.html#debugging-strategies"},{"type":"extras","title":"1. Using Debug Steps - How to Debug Reactor Workflows","doc":"The simplest way to understand what's happening in your reactor is to add debug steps that log intermediate values.\n\n#### Basic Debug Step\n\n```elixir\ndefmodule UserProcessingReactor do\n  use Reactor\n\n  input :user_data\n\n  step :validate_user, MyApp.Steps.ValidateUser do\n    argument :user_data, input(:user_data)\n  end\n\n  debug :check_validation do\n    argument :user, result(:validate_user)\n    level :info\n  end\n\n  step :create_user, MyApp.Steps.CreateUser do\n    argument :user, result(:validate_user)\n  end\n\n  debug :check_creation do\n    argument :result, result(:create_user)\n  end\nend\n```\n\n#### Debug Output\n\nDebug steps log comprehensive information about arguments, context, and options:\n\n```\n[info] # Debug information for step `:check_validation`.","ref":"debugging-workflows.html#1-using-debug-steps"},{"type":"extras","title":"Arguments - How to Debug Reactor Workflows","doc":"%{\n  user: %{\n    email: \"user@example.com\",\n    name: \"John Doe\",\n    validated: true\n  }\n}","ref":"debugging-workflows.html#arguments"},{"type":"extras","title":"Context - How to Debug Reactor Workflows","doc":"%{\n  concurrency_key: #Reference<0.123.456.789>,\n  current_step: :check_validation\n}","ref":"debugging-workflows.html#context"},{"type":"extras","title":"Options - How to Debug Reactor Workflows","doc":"[]\n```","ref":"debugging-workflows.html#options"},{"type":"extras","title":"2. Synchronous Execution for Debugging - How to Debug Reactor Workflows","doc":"When debugging, disable async execution to get predictable, deterministic behavior:\n\n```elixir\ndefmodule DebuggingTest do\n  test \"debug complex workflow\" do\n    inputs = %{user_data: sample_data()}\n    \n    result = Reactor.run(UserProcessingReactor, inputs, async?: false)\n    \n    case result do\n      {:ok, user} -> \n        IO.puts(\"Success: #{inspect(user)}\")\n      {:error, errors} -> \n        IO.puts(\"Failed: #{inspect(errors)}\")\n    end\n  end\nend\n```","ref":"debugging-workflows.html#2-synchronous-execution-for-debugging"},{"type":"extras","title":"3. Visual Workflow Debugging - How to Debug Reactor Workflows","doc":"Generate visual diagrams to understand reactor structure and flow:\n\n#### Creating Mermaid Diagrams\n\nUse the built-in Mix task to generate visual diagrams:\n\n```bash\n# Generate basic diagram\nmix reactor.mermaid UserProcessingReactor\n\n# Include descriptions and expand sub-reactors\nmix reactor.mermaid UserProcessingReactor --describe --expand\n\n# Save to specific file\nmix reactor.mermaid UserProcessingReactor --output debug_flow.mmd\n\n# Display diagram for copy-pasting into Mermaid Live\nmix reactor.mermaid UserProcessingReactor --format copy\n\n# Generate direct Mermaid Live Editor URL\nmix reactor.mermaid UserProcessingReactor --format url\n```\n\nOr generate diagrams programmatically:\n\n```elixir\ndefmodule DebugHelpers do\n  def visualize_reactor(reactor_module, filename \\\\ \"reactor_debug.mmd\") do\n    {:ok, diagram} = Reactor.Mermaid.to_mermaid(reactor_module, \n      describe?: true,\n      expand?: true\n    )\n    \n    File.write!(filename, diagram)\n    IO.puts(\"Diagram saved to #{filename}\")\n    IO.puts(\"View at: https://mermaid.live/edit\")\n  end\nend\n\nDebugHelpers.visualize_reactor(UserProcessingReactor)\n```\n\n#### Understanding the Diagram\n\nThe generated diagram shows:\n- Input dependencies\n- Step execution order\n- Data flow between steps\n- Conditional branches and loops","ref":"debugging-workflows.html#3-visual-workflow-debugging"},{"type":"extras","title":"4. Telemetry-Based Debugging - How to Debug Reactor Workflows","doc":"Set up telemetry handlers to monitor reactor execution in real-time:\n\n#### Setting Up Telemetry\n\n```elixir\ndefmodule MyApp.ReactorTelemetry do\n  def setup_debugging() do\n    events = [\n      [:reactor, :run, :start],\n      [:reactor, :run, :stop],\n      [:reactor, :step, :run, :start],\n      [:reactor, :step, :run, :stop]\n    ]\n    \n    :telemetry.attach_many(\n      \"reactor-debug\",\n      events,\n      &handle_event/4,\n      nil\n    )\n  end\n  \n  def handle_event([:reactor, :run, :start], _measurements, metadata, _config) do\n    IO.puts(\"🚀 Starting reactor: #{metadata.reactor.id}\")\n  end\n  \n  def handle_event([:reactor, :run, :stop], measurements, metadata, _config) do\n    duration_ms = measurements.duration / 1_000_000\n    \n    case metadata.outcome do\n      :ok -> \n        IO.puts(\"✅ Reactor completed in #{duration_ms}ms\")\n      :error -> \n        IO.puts(\"❌ Reactor failed in #{duration_ms}ms\")\n      :halt ->\n        IO.puts(\"⏸️ Reactor halted in #{duration_ms}ms\")\n    end\n  end\n  \n  def handle_event([:reactor, :step, :run, :start], _measurements, metadata, _config) do\n    IO.puts(\"   🔄 Starting step: #{metadata.step.name}\")\n  end\n  \n  def handle_event([:reactor, :step, :run, :stop], measurements, metadata, _config) do\n    duration_ms = measurements.duration / 1_000_000\n    \n    case metadata.outcome do\n      :ok -> \n        IO.puts(\"   ✅ Step #{metadata.step.name} completed in #{duration_ms}ms\")\n      :error -> \n        IO.puts(\"   ❌ Step #{metadata.step.name} failed in #{duration_ms}ms\")\n      :retry ->\n        IO.puts(\"   🔄 Step #{metadata.step.name} retrying after #{duration_ms}ms\")\n    end\n  end\nend\n```\n\n#### Using Telemetry in Your Reactor\n\n```elixir\ndefmodule MyReactor do\n  use Reactor\n\n  middlewares do\n    middleware Reactor.Middleware.Telemetry\n  end\n\n  # ... steps\nend\n\n# Set up debugging and run\nMyApp.ReactorTelemetry.setup_debugging()\nReactor.run(MyReactor, inputs)\n```","ref":"debugging-workflows.html#4-telemetry-based-debugging"},{"type":"extras","title":"5. Error Analysis and Debugging - How to Debug Reactor Workflows","doc":"Understand and debug errors using Reactor's structured error system:\n\n#### Analyzing Reactor Errors\n\n```elixir\ndefmodule ErrorAnalyzer do\n  def debug_error(error) do\n    IO.puts(\"🔍 Analyzing reactor error...\")\n    \n    case error do\n      %Reactor.Error.Invalid.RunStepError{} = step_error ->\n        IO.puts(\"❌ Step execution failed:\")\n        IO.puts(\"   Step: #{step_error.step.name}\")\n        IO.puts(\"   Error: #{inspect(step_error.error)}\")\n        debug_step_context(step_error.step)\n        \n      %Reactor.Error.Invalid.MissingArgumentError{} = arg_error ->\n        IO.puts(\"❌ Missing argument:\")\n        IO.puts(\"   Argument: #{arg_error.argument.name}\")\n        IO.puts(\"   Step: #{arg_error.step.name}\")\n        debug_dependency_chain(arg_error.step, arg_error.argument)\n        \n      %Reactor.Error.Validation.MissingReturnError{} ->\n        IO.puts(\"❌ No return value specified for reactor\")\n        IO.puts(\"💡 Add a return statement to your reactor\")\n        \n      errors when is_list(errors) ->\n        IO.puts(\"❌ Multiple errors occurred:\")\n        Enum.with_index(errors, 1)\n        |> Enum.each(fn {err, idx} ->\n          IO.puts(\"   #{idx}. #{inspect(err)}\")\n        end)\n        \n      _ ->\n        IO.puts(\"❌ Unknown error: #{inspect(error)}\")\n    end\n  end\n  \n  defp debug_step_context(step) do\n    IO.puts(\"   Arguments:\")\n    Enum.each(step.arguments, fn arg ->\n      IO.puts(\"     - #{arg.name}: #{inspect(arg.source)}\")\n    end)\n  end\n  \n  defp debug_dependency_chain(step, argument) do\n    IO.puts(\"💡 Check if the source step completed successfully:\")\n    \n    case argument.source do\n      %Reactor.Template.Result{name: source_step} ->\n        IO.puts(\"   Source step: #{source_step}\")\n        IO.puts(\"   Add a debug step after #{source_step} to verify its output\")\n        \n      %Reactor.Template.Input{name: input_name} ->\n        IO.puts(\"   Missing input: #{input_name}\")\n        IO.puts(\"   Ensure this input is provided when running the reactor\")\n        \n      _ ->\n        IO.puts(\"   Source: #{inspect(argument.source)}\")\n    end\n  end\nend\n\n# Usage\ncase Reactor.run(MyReactor, inputs) do\n  {:ok, result} -> \n    IO.puts(\"Success: #{inspect(result)}\")\n  {:error, errors} ->\n    Enum.each(List.wrap(errors), &ErrorAnalyzer.debug_error/1)\nend\n```","ref":"debugging-workflows.html#5-error-analysis-and-debugging"},{"type":"extras","title":"6. Performance Debugging - How to Debug Reactor Workflows","doc":"Identify performance bottlenecks in your workflows:\n\n#### Performance Monitoring\n\n```elixir\ndefmodule PerformanceMonitor do\n  def setup_performance_monitoring(slow_threshold_ms \\\\ 100) do\n    :telemetry.attach(\n      \"reactor-performance\",\n      [:reactor, :step, :run, :stop],\n      &monitor_step_performance/4,\n      %{threshold: slow_threshold_ms * 1_000_000}  # Convert to nanoseconds\n    )\n  end\n  \n  def monitor_step_performance(_event, measurements, metadata, config) do\n    if measurements.duration > config.threshold do\n      duration_ms = measurements.duration / 1_000_000\n      \n      IO.puts(\"🐌 Slow step detected:\")\n      IO.puts(\"   Step: #{metadata.step.name}\")\n      IO.puts(\"   Duration: #{duration_ms}ms\")\n      IO.puts(\"   Module: #{metadata.step.impl}\")\n    end\n  end\nend\n\nPerformanceMonitor.setup_performance_monitoring(50)  # 50ms threshold\n```\n\n#### Concurrency Debugging\n\n```elixir\ndefmodule ConcurrencyDebugger do\n  def debug_concurrency(reactor_module, inputs) do\n    # Monitor async step spawning\n    :telemetry.attach(\n      \"concurrency-debug\",\n      [:reactor, :step, :process, :start],\n      fn _event, _measurements, metadata, _config ->\n        IO.puts(\"🚀 Spawned async step: #{metadata.step.name} (PID: #{inspect(metadata.pid)})\")\n      end,\n      nil\n    )\n    \n    # Run with limited concurrency for debugging\n    Reactor.run(reactor_module, inputs, max_concurrency: 2)\n  end\nend\n```","ref":"debugging-workflows.html#6-performance-debugging"},{"type":"extras","title":"7. Interactive Debugging - How to Debug Reactor Workflows","doc":"Use IEx for interactive debugging sessions:\n\n#### Adding Breakpoints\n\n```elixir\ndefmodule MyApp.Steps.DebugStep do\n  use Reactor.Step\n  \n  def run(arguments, context, _options) do\n    IO.puts(\"🔍 Interactive debugging point\")\n    IO.puts(\"Arguments: #{inspect(arguments)}\")\n    IO.puts(\"Context: #{inspect(context)}\")\n    \n    # Start interactive session\n    require IEx; IEx.pry()\n    \n    {:ok, arguments}\n  end\nend\n\n# Use in reactor\nstep :debug_point, MyApp.Steps.DebugStep do\n  argument :data, result(:previous_step)\nend\n```\n\n#### Runtime Inspection\n\n```elixir\ndefmodule ReactorInspector do\n  def inspect_reactor_state(reactor_module) do\n    {:ok, reactor_struct} = Reactor.Info.to_struct(reactor_module)\n    \n    IO.puts(\"📊 Reactor Analysis:\")\n    IO.puts(\"   Inputs: #{length(reactor_struct.inputs)}\")\n    IO.puts(\"   Steps: #{length(reactor_struct.steps)}\")\n    \n    IO.puts(\"\\n📝 Step Details:\")\n    Enum.each(reactor_struct.steps, fn step ->\n      IO.puts(\"   #{step.name}:\")\n      IO.puts(\"     Module: #{step.impl}\")\n      IO.puts(\"     Async: #{step.async?}\")\n      IO.puts(\"     Arguments: #{length(step.arguments)}\")\n    end)\n    \n    if reactor_struct.plan do\n      IO.puts(\"\\n🗺️ Execution Plan:\")\n      IO.inspect(reactor_struct.plan)\n    end\n  end\nend\n```","ref":"debugging-workflows.html#7-interactive-debugging"},{"type":"extras","title":"Debugging Workflow - How to Debug Reactor Workflows","doc":"1. **Start with synchronous execution** (`async?: false`)\n2. **Add debug steps** at key points in your workflow\n3. **Generate visual diagrams** to understand data flow\n4. **Set up telemetry monitoring** for performance insights\n5. **Use error analysis** to understand failure modes\n6. **Add interactive breakpoints** for complex issues","ref":"debugging-workflows.html#debugging-workflow"},{"type":"extras","title":"Common Debugging Scenarios - How to Debug Reactor Workflows","doc":"","ref":"debugging-workflows.html#common-debugging-scenarios"},{"type":"extras","title":"Step Not Executing - How to Debug Reactor Workflows","doc":"```elixir\n# Check if dependencies are met\ndebug :before_problem_step do\n  argument :data, result(:previous_step)\nend\n\nstep :problem_step do\n  argument :data, result(:previous_step)\n  # ...\nend\n```","ref":"debugging-workflows.html#step-not-executing"},{"type":"extras","title":"Unexpected Results - How to Debug Reactor Workflows","doc":"```elixir\n# Compare expected vs actual\ndebug :validate_result do\n  argument :result, result(:calculation_step)\nend\n\nstep :verify_result do\n  argument :result, result(:calculation_step)\n  run fn %{result: result}, _context ->\n    expected = 42\n    if result == expected do\n      {:ok, result}\n    else\n      {:error, \"Expected #{expected}, got #{result}\"}\n    end\n  end\nend\n```","ref":"debugging-workflows.html#unexpected-results"},{"type":"extras","title":"Related Guides - How to Debug Reactor Workflows","doc":"- [Testing Strategies](testing-strategies.md) - Test your debugging assumptions\n- [Performance Optimization](performance-optimization.md) - Optimize identified bottlenecks\n- [Error Handling Tutorial](../tutorials/02-error-handling.md) - Handle errors gracefully\n\nThis comprehensive debugging approach helps you quickly identify and resolve issues in your reactor workflows.","ref":"debugging-workflows.html#related-guides"},{"type":"extras","title":"How to Optimize Reactor Performance","doc":"# How to Optimize Reactor Performance","ref":"performance-optimization.html"},{"type":"extras","title":"Problem - How to Optimize Reactor Performance","doc":"You need to improve the performance of your reactor workflows, optimize concurrency, handle large-scale processing efficiently, and scale your system to handle high throughput with minimal resource usage.","ref":"performance-optimization.html#problem"},{"type":"extras","title":"Solution Overview - How to Optimize Reactor Performance","doc":"This guide covers systematic performance optimization techniques for Reactor workflows, including concurrency tuning, resource management, memory optimization, and monitoring strategies. You'll learn to identify bottlenecks and apply targeted optimizations.","ref":"performance-optimization.html#solution-overview"},{"type":"extras","title":"Prerequisites - How to Optimize Reactor Performance","doc":"- Understanding of Reactor basics (inputs, steps, arguments)\n- Familiarity with Elixir concurrency concepts (processes, tasks, supervisors)\n- Experience with async workflows and map steps\n- Basic knowledge of system performance concepts","ref":"performance-optimization.html#prerequisites"},{"type":"extras","title":"Understanding Performance Characteristics - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#understanding-performance-characteristics"},{"type":"extras","title":"Reactor's Performance Model - How to Optimize Reactor Performance","doc":"Reactor is designed for high-throughput, concurrent execution with these characteristics:\n\n**Strengths:**\n- **Automatic parallelisation**: Independent steps run concurrently by default\n- **Resource pooling**: Shared concurrency pools across reactor instances\n- **Lazy evaluation**: Only processes what's needed when it's needed\n- **Efficient batching**: Map steps process collections in configurable chunks\n\n**Bottlenecks to watch for:**\n- **Dependency chains**: Sequential dependencies limit parallelisation\n- **Resource contention**: Too many concurrent operations overwhelming external systems\n- **Memory usage**: Large intermediate results consuming memory\n- **CPU vs I/O mixing**: Synchronous CPU work blocking async I/O operations","ref":"performance-optimization.html#reactor-s-performance-model"},{"type":"extras","title":"Concurrency Optimization - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#concurrency-optimization"},{"type":"extras","title":"1. Configuring Basic Concurrency Limits - How to Optimize Reactor Performance","doc":"Control how many steps can run concurrently within a single reactor:\n\n```elixir\ndefmodule DataProcessor do\n  use Reactor\n\n  input :user_ids\n\n  # Multiple independent steps that can run in parallel\n  step :fetch_profiles do\n    argument :user_ids, input(:user_ids)\n    run fn %{user_ids: ids}, _context ->\n      profiles = fetch_user_profiles(ids)\n      {:ok, profiles}\n    end\n  end\n\n  step :fetch_preferences do\n    argument :user_ids, input(:user_ids)\n    run fn %{user_ids: ids}, _context ->\n      preferences = fetch_user_preferences(ids)\n      {:ok, preferences}\n    end\n  end\n\n  step :fetch_activity do\n    argument :user_ids, input(:user_ids)\n    run fn %{user_ids: ids}, _context ->\n      activity = fetch_user_activity(ids)\n      {:ok, activity}\n    end\n  end\n\n  # This step waits for all the above to complete\n  step :combine_data do\n    argument :profiles, result(:fetch_profiles)\n    argument :preferences, result(:fetch_preferences)\n    argument :activity, result(:fetch_activity)\n    \n    run fn args, _context ->\n      combined = combine_user_data(args)\n      {:ok, combined}\n    end\n  end\n\n  return :combine_data\nend\n\n# Control concurrency when running the reactor\n{:ok, result} = Reactor.run(\n  DataProcessor,\n  %{user_ids: [1, 2, 3, 4, 5]},\n  %{},\n  max_concurrency: 10  # At most 10 steps running concurrently\n)\n```","ref":"performance-optimization.html#1-configuring-basic-concurrency-limits"},{"type":"extras","title":"2. Tuning Concurrency Settings - How to Optimize Reactor Performance","doc":"Optimal concurrency settings depend on your specific workload and system characteristics. Start with these guidelines, then experiment and measure:\n\n```elixir\n# Starting points for different workload types:\n\n# For I/O-bound workloads (API calls, database queries)\nmax_concurrency: System.schedulers_online() * 4  # 4x CPU cores\n\n# For CPU-bound workloads  \nmax_concurrency: System.schedulers_online()      # 1x CPU cores\n\n# For mixed workloads\nmax_concurrency: System.schedulers_online() * 2  # 2x CPU cores\n\n# For external service limits (rate limiting)\nmax_concurrency: 10  # Based on service constraints\n```\n\n**How to find your optimal settings:**\n\n1. **Start with the suggested baseline** for your workload type\n2. **Monitor system resources** - CPU usage, memory, network connections\n3. **Test with different values** - try 50%, 150%, 200% of your baseline\n4. **Measure end-to-end performance** - both throughput and latency\n5. **Watch for external bottlenecks** - database connection limits, API rate limits\n6. **Consider system stability** - avoid settings that cause resource exhaustion\n\n**Signs you need to adjust:**\n\n- **Too low**: CPU cores are idle, external services aren't being fully utilised\n- **Too high**: High memory usage, connection pool exhaustion, degraded response times\n- **External limits**: Rate limiting errors, connection timeouts, service overload responses","ref":"performance-optimization.html#2-tuning-concurrency-settings"},{"type":"extras","title":"3. Shared Concurrency Pools - How to Optimize Reactor Performance","doc":"Share concurrency pools across reactor instances to prevent resource competition:\n\n```elixir\n# Create a shared pool\n{:ok, pool_key} = Reactor.Executor.ConcurrencyTracker.allocate_pool(100)\n\n# Use the pool across multiple reactor runs\nopts = [concurrency_key: pool_key, max_concurrency: 100]\n\n# All these reactors share the same 100-task limit\nTask.async(fn -> Reactor.run(DataProcessor, inputs1, %{}, opts) end)\nTask.async(fn -> Reactor.run(DataProcessor, inputs2, %{}, opts) end) \nTask.async(fn -> Reactor.run(DataProcessor, inputs3, %{}, opts) end)\n```","ref":"performance-optimization.html#3-shared-concurrency-pools"},{"type":"extras","title":"Map Step Performance Optimization - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#map-step-performance-optimization"},{"type":"extras","title":"1. Batch Size Tuning - How to Optimize Reactor Performance","doc":"Configure batch sizes based on data characteristics and system capacity:\n\n```elixir\ndefmodule BatchProcessingReactor do\n  use Reactor\n\n  input :records\n  input :processing_type\n\n  # For small, fast operations - larger batches\n  map :process_lightweight_data do\n    argument :source, input(:records)\n    batch_size 1000  # Process 1000 items at once\n    allow_async? true\n    \n    step :validate do\n      argument :record, element(:source)\n      run fn %{record: record}, _context ->\n        # Fast validation logic\n        {:ok, validate_record(record)}\n      end\n    end\n    \n    return :validate\n  end\n\n  # For expensive operations - smaller batches  \n  map :process_heavy_computation do\n    argument :source, input(:records)\n    batch_size 10    # Process only 10 items at once\n    # Tune reactor-level max_concurrency for CPU-bound work\n    \n    step :complex_calculation do\n      argument :record, element(:source)\n      run fn %{record: record}, _context ->\n        # CPU-intensive processing\n        {:ok, expensive_computation(record)}\n      end\n    end\n    \n    return :complex_calculation\n  end\n\n  return template(\"Processing complete: {{ processed }} records\") do\n    assign :processed, result(:process_lightweight_data)\n  end\nend\n```","ref":"performance-optimization.html#1-batch-size-tuning"},{"type":"extras","title":"2. Understanding Map Step Memory Usage - How to Optimize Reactor Performance","doc":"**Critical memory considerations** when using map steps:\n\nMap steps have complex memory usage patterns that are a function of both **batch size** and **mapped result size**. Understanding this is crucial for processing large datasets efficiently.\n\n#### Memory Usage Components\n\n**1. Input Record Storage:**\nEach batch of input records is converted into individual steps and added to the reactor state. These steps contain the input record as a `value` argument, contributing directly to memory usage until the step runs.\n\n**2. Intermediate Results Storage:**\nA new map step is emitted which depends on the results of all batch steps. This means **all batch step results are stored in Reactor's intermediate results** until the map step completes.\n\n**3. Final Result Storage:**\nThe overall result of the map step (collection of all batch results) is likely to be depended upon by other steps, so remains in intermediate results storage.\n\n#### Memory Usage Formula\n\n```elixir\nTotal Memory ≈ (Batch Size × Input Record Size) + \n               (Batch Size × Output Result Size) + \n               (Final Collection Size)\n```\n\n#### Batch Size Guidelines by Data Characteristics\n\n```elixir\n# For small input records, small output results\nbatch_size 1000  # Safe - total memory stays manageable\n\n# For large input records OR large output results  \nbatch_size 10    # Conservative - limits memory multiplication\n\n# For very large transformations (e.g., image processing)\nbatch_size 1     # Process one at a time to avoid memory spikes\n\n# Consider your total dataset size:\n# 1M records × 100 batch size = 100K steps in memory at once\n# 1M records × 10 batch size = 100K steps total, 10K at once\n```\n\n#### Example: Memory-Aware Batch Configuration\n\n```elixir\n# For small records and lightweight transformations\ndefmodule LightweightProcessor do\n  use Reactor\n\n  input :records\n\n  map :process_records do\n    source input(:records)\n    batch_size 1000  # Safe for small records (~1KB each)\n    \n    step :transform_record do\n      argument :record, element(:process_records)\n      run fn %{record: record}, _context ->\n        # Lightweight transformation\n        {:ok, simple_transform(record)}\n      end\n    end\n    \n    return :transform_record\n  end\n\n  return :process_records\nend\n\n# For large records or heavy transformations  \ndefmodule HeavyProcessor do\n  use Reactor\n\n  input :records\n\n  map :process_records do\n    source input(:records)\n    batch_size 10   # Conservative for large records (>10KB each)\n    \n    step :transform_record do\n      argument :record, element(:process_records)\n      run fn %{record: record}, _context ->\n        # Heavy transformation that produces large results\n        {:ok, complex_transform(record)}\n      end\n    end\n    \n    return :transform_record\n  end\n\n  return :process_records\nend\n\n# For very large records (images, files, etc.)\ndefmodule SingleItemProcessor do\n  use Reactor\n\n  input :large_items\n\n  map :process_items do\n    source input(:large_items)\n    batch_size 1    # Process one large item at a time\n    \n    step :transform_item do\n      argument :item, element(:process_items)\n      run fn %{item: item}, _context ->\n        # Memory-intensive processing (e.g., image manipulation)\n        {:ok, process_large_item(item)}\n      end\n    end\n    \n    return :transform_item\n  end\n\n  return :process_items\nend\n```","ref":"performance-optimization.html#2-understanding-map-step-memory-usage"},{"type":"extras","title":"3. Memory-Efficient Streaming - How to Optimize Reactor Performance","doc":"Process large datasets without loading everything into memory:\n\n```elixir\ndefmodule StreamingProcessor do\n  use Reactor\n\n  input :file_path\n  input :output_path\n\n  # Stream file processing without loading entire file\n  map :process_file_stream do\n    argument :source, input(:file_path)\n    batch_size 100\n    strict_ordering? false  # Improves performance when order doesn't matter\n    \n    step :transform_line do\n      argument :line, element(:source)\n      run fn %{line: line}, _context ->\n        transformed = transform_data(line)\n        {:ok, transformed}\n      end\n    end\n    \n    return :transform_line\n  end\n\n  # Write results in batches  \n  step :write_output do\n    argument :processed_data, result(:process_file_stream)\n    argument :output_path, input(:output_path)\n    \n    run fn %{processed_data: data, output_path: path}, _context ->\n      File.stream!(path, [], :line)\n      |> Stream.chunk_every(1000)\n      |> Stream.each(&write_batch(&1))\n      |> Stream.run()\n      \n      {:ok, :written}\n    end\n  end\n\n  return :write_output\nend\n\n```","ref":"performance-optimization.html#3-memory-efficient-streaming"},{"type":"extras","title":"CPU vs I/O Optimization - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#cpu-vs-i-o-optimization"},{"type":"extras","title":"1. Optimizing CPU and I/O Operations - How to Optimize Reactor Performance","doc":"Balance concurrency based on operation type - I/O operations can handle high concurrency, while CPU operations should use more conservative limits:\n\n```elixir\ndefmodule OptimalWorkloadReactor do\n  use Reactor\n\n  input :user_ids\n\n  # I/O operations - keep async (default)\n  step :fetch_profiles do\n    argument :user_ids, input(:user_ids)\n    run fn %{user_ids: ids}, _context ->\n      profiles = fetch_user_profiles(ids)\n      {:ok, profiles}\n    end\n  end\n\n  step :fetch_preferences do  \n    argument :user_ids, input(:user_ids)\n    run fn %{user_ids: ids}, _context ->\n      preferences = fetch_user_preferences(ids)\n      {:ok, preferences}\n    end\n  end\n\n  # CPU operations - keep async but use lower concurrency limits\n  step :calculate_recommendations do\n    argument :profiles, result(:fetch_profiles)\n    argument :preferences, result(:fetch_preferences)\n    \n    run fn %{profiles: profiles, preferences: prefs}, _context ->\n      recommendations = calculate_complex_recommendations(profiles, prefs)\n      {:ok, recommendations}\n    end\n  end\n\n  # I/O operation - back to async\n  step :save_recommendations do\n    argument :recommendations, result(:calculate_recommendations)\n    run fn %{recommendations: recs}, _context ->\n      save_to_database(recs)\n      {:ok, :saved}\n    end\n  end\n\n  return :save_recommendations\nend\n\n# Run with different concurrency settings based on workload\n# For this mixed I/O + CPU reactor, use moderate concurrency\n{:ok, result} = Reactor.run(\n  OptimalWorkloadReactor,\n  %{user_ids: [1, 2, 3, 4, 5]},\n  %{},\n  max_concurrency: System.schedulers_online() * 2  # Balanced for mixed workload\n)\n```","ref":"performance-optimization.html#1-optimizing-cpu-and-i-o-operations"},{"type":"extras","title":"2. CPU-bound Map Operations - How to Optimize Reactor Performance","doc":"Handle CPU-intensive map operations efficiently:\n\n```elixir\ndefmodule CPUIntensiveProcessor do\n  use Reactor\n\n  input :datasets\n\n  # CPU-bound processing with controlled concurrency\n  map :process_datasets do\n    argument :source, input(:datasets)\n    batch_size 5        # Small batches for CPU work\n    # Tune reactor-level max_concurrency for CPU-bound work\n    \n    step :complex_analysis do\n      argument :dataset, element(:source)\n      run fn %{dataset: data}, _context ->\n        # CPU-intensive mathematical analysis\n        result = perform_statistical_analysis(data)\n        {:ok, result}\n      end\n    end\n    \n    return :complex_analysis\n  end\n\n  return :process_datasets\nend\n```","ref":"performance-optimization.html#2-cpu-bound-map-operations"},{"type":"extras","title":"Memory Management - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#memory-management"},{"type":"extras","title":"1. Controlling Intermediate Results - How to Optimize Reactor Performance","doc":"Minimize memory usage by avoiding unnecessary intermediate result storage. Remember that undoable steps will have their results stored in the reactor's undo stack, so non-undoable steps save memory:\n\n```elixir\ndefmodule MemoryEfficientReactor do\n  use Reactor\n\n  input :large_dataset\n\n  # This step's result won't be stored since no other steps depend on it\n  # Also, it's not undoable so won't be kept in the undo stack\n  step :validate_data do\n    argument :data, input(:large_dataset)\n    run fn %{data: data}, _context ->\n      # Large intermediate result that we don't want to keep\n      validated = validate_large_dataset(data) \n      \n      # Return only what's needed for next steps\n      summary = %{\n        total_records: length(validated),\n        valid_records: count_valid(validated),\n        errors: extract_errors(validated)\n      }\n      \n      {:ok, summary}\n    end\n    # No undo/4 callback defined = not undoable = no result stored in undo stack\n  end\n\n  # Use summary instead of full dataset\n  step :generate_report do\n    argument :summary, result(:validate_data)\n    run fn %{summary: summary}, _context ->\n      report = create_summary_report(summary)\n      {:ok, report}\n    end\n  end\n\n  return :generate_report\nend\n```","ref":"performance-optimization.html#1-controlling-intermediate-results"},{"type":"extras","title":"Performance Monitoring - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#performance-monitoring"},{"type":"extras","title":"1. Adding Telemetry for Performance Tracking - How to Optimize Reactor Performance","doc":"Monitor reactor performance with built-in telemetry:\n\n```elixir\ndefmodule MonitoredReactor do\n  use Reactor\n  \n  middleware Reactor.Middleware.Telemetry\n\n  input :data\n\n  step :slow_operation do\n    argument :data, input(:data)\n    run fn %{data: data}, _context ->\n      # Slow operation that we want to monitor\n      result = expensive_operation(data)\n      {:ok, result}\n    end\n  end\n\n  return :slow_operation\nend\n\n# Set up telemetry handlers\n:telemetry.attach_many(\n  \"reactor-performance\",\n  [\n    [:reactor, :run, :start],\n    [:reactor, :run, :stop],\n    [:reactor, :step, :run, :start], \n    [:reactor, :step, :run, :stop]\n  ],\n  fn event, measurements, metadata, _config ->\n    case event do\n      [:reactor, :run, :stop] ->\n        duration_ms = measurements.duration |> System.convert_time_unit(:native, :millisecond)\n        Logger.info(\"Reactor completed in #{duration_ms}ms\")\n        \n      [:reactor, :step, :run, :stop] ->\n        duration_ms = measurements.duration |> System.convert_time_unit(:native, :millisecond)\n        step_name = metadata.step.name\n        Logger.debug(\"Step #{step_name} completed in #{duration_ms}ms\")\n        \n      _ -> :ok\n    end\n  end,\n  nil\n)\n```","ref":"performance-optimization.html#1-adding-telemetry-for-performance-tracking"},{"type":"extras","title":"Performance Benchmarking - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#performance-benchmarking"},{"type":"extras","title":"1. Using Benchee for Reactor Performance Testing - How to Optimize Reactor Performance","doc":"Use [Benchee](https://hex.pm/packages/benchee) to create professional performance benchmarks with detailed reports:\n\n```elixir\n# Add to mix.exs dependencies\n{:benchee, \"~> 1.0\", only: :dev}\n\ndefmodule ReactorBenchmarks do\n  def run_concurrency_benchmark do\n    data = generate_test_data(1_000)\n    \n    Benchee.run(\n      %{\n        \"concurrency_1\" => fn ->\n          Reactor.run(ProcessingReactor, %{data: data}, %{}, max_concurrency: 1)\n        end,\n        \"concurrency_5\" => fn ->\n          Reactor.run(ProcessingReactor, %{data: data}, %{}, max_concurrency: 5)\n        end,\n        \"concurrency_10\" => fn ->\n          Reactor.run(ProcessingReactor, %{data: data}, %{}, max_concurrency: 10)\n        end,\n        \"concurrency_20\" => fn ->\n          Reactor.run(ProcessingReactor, %{data: data}, %{}, max_concurrency: 20)\n        end\n      },\n      warmup: 2,\n      time: 5,\n      memory_time: 2,\n      formatters: [\n        Benchee.Formatters.Console,\n        {Benchee.Formatters.HTML, file: \"benchmarks/concurrency_results.html\"}\n      ]\n    )\n  end\n\n  def run_batch_size_benchmark do\n    data = generate_test_data(1_000)\n    \n    # Use inputs to test different batch sizes\n    Benchee.run(\n      %{\n        \"map_processing\" => fn {batch_size, data} ->\n          reactor = create_map_reactor(batch_size)\n          Reactor.run(reactor, %{data: data})\n        end\n      },\n      inputs: %{\n        \"batch_10\" => {10, data},\n        \"batch_50\" => {50, data}, \n        \"batch_100\" => {100, data},\n        \"batch_500\" => {500, data}\n      },\n      warmup: 1,\n      time: 3,\n      memory_time: 1\n    )\n  end\n\n  def run_reactor_comparison_benchmark do\n    data = generate_test_data(500)\n    \n    Benchee.run(\n      %{\n        \"sync_reactor\" => fn ->\n          Reactor.run(SyncReactor, %{data: data}, %{}, async?: false)\n        end,\n        \"async_reactor\" => fn ->\n          Reactor.run(AsyncReactor, %{data: data}, %{}, max_concurrency: 10)\n        end,\n        \"optimized_reactor\" => fn ->\n          Reactor.run(OptimizedReactor, %{data: data}, %{}, max_concurrency: 5)\n        end\n      },\n      warmup: 2,\n      time: 5,\n      memory_time: 2,\n      formatters: [\n        Benchee.Formatters.Console,\n        {Benchee.Formatters.HTML, file: \"benchmarks/reactor_comparison.html\"},\n        {Benchee.Formatters.JSON, file: \"benchmarks/reactor_comparison.json\"}\n      ]\n    )\n  end\nend\n\n# Example output from Benchee:\n#\n# Name                    ips        average  deviation         median         99th %\n# concurrency_10      12.34 K       81.05 μs    ±15.23%       76.00 μs      145.67 μs\n# concurrency_5        8.91 K      112.23 μs    ±18.45%      108.00 μs      189.34 μs\n# concurrency_1        3.45 K      289.78 μs    ±12.67%      285.00 μs      387.23 μs\n#\n# Comparison:\n# concurrency_10      12.34 K\n# concurrency_5        8.91 K - 1.38x slower +31.18 μs\n# concurrency_1        3.45 K - 3.58x slower +208.73 μs\n```","ref":"performance-optimization.html#1-using-benchee-for-reactor-performance-testing"},{"type":"extras","title":"Common Performance Anti-patterns - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#common-performance-anti-patterns"},{"type":"extras","title":"1. Avoid: Blocking Async Steps - How to Optimize Reactor Performance","doc":"When a step blocks, identify the root cause and model dependencies properly:\n\n```elixir\n# ❌ BAD: Blocking because of missing dependency\nstep :process_data do\n  run fn args, _context ->\n    # Blocking while waiting for external resource\n    wait_for_resource_to_be_ready()  # This wastes concurrency slots\n    process_with_resource(args)\n  end\nend\n\n# ✅ GOOD: Model the dependency explicitly with separate steps\nstep :prepare_resource do\n  run fn _args, _context ->\n    # This step ensures the resource is ready\n    setup_resource()\n    {:ok, :resource_ready}\n  end\nend\n\nstep :process_data do\n  argument :resource_ready, result(:prepare_resource)\n  argument :data, input(:data)\n  run fn %{data: data}, _context ->\n    # No blocking - resource dependency ensures it's ready\n    process_with_resource(data)\n  end\nend\n\n# ✅ BETTER: Use recursive step for external resources\nstep :wait_for_external_service do\n  run fn args, context ->\n    case check_external_service() do\n      {:ok, result} -> \n        {:ok, result}\n      {:error, :not_ready} ->\n        # Emit the current step again - Reactor will retry when it can\n        current_step = context.current_step\n        {:ok, nil, [current_step]}\n    end\n  end\nend\n```","ref":"performance-optimization.html#1-avoid-blocking-async-steps"},{"type":"extras","title":"2. Avoid: Excessive Dependencies - How to Optimize Reactor Performance","doc":"Design workflows to maximise parallelisation:\n\n```elixir\n# ❌ BAD: Artificial sequential dependencies\nstep :step1, do: run(fn -> {:ok, data1} end)\nstep :step2 do\n  argument :data1, result(:step1)  # Unnecessary dependency\n  run fn %{data1: _} -> {:ok, data2} end  # Doesn't actually use data1\nend\nstep :step3 do\n  argument :data2, result(:step2)  # Creates sequential chain\n  run fn %{data2: _} -> {:ok, data3} end\nend\n\n# ✅ GOOD: Independent steps run in parallel\nstep :step1, do: run(fn -> {:ok, data1} end)\nstep :step2, do: run(fn -> {:ok, data2} end)  # No dependency on step1\nstep :step3, do: run(fn -> {:ok, data3} end)  # No dependency on step2\n\n# Combine results only when needed\nstep :combine_results do\n  argument :data1, result(:step1)\n  argument :data2, result(:step2)\n  argument :data3, result(:step3)\n  run fn args -> {:ok, combine(args)} end\nend\n```","ref":"performance-optimization.html#2-avoid-excessive-dependencies"},{"type":"extras","title":"Performance Troubleshooting - How to Optimize Reactor Performance","doc":"","ref":"performance-optimization.html#performance-troubleshooting"},{"type":"extras","title":"Common issues and solutions - How to Optimize Reactor Performance","doc":"**Problem**: Reactor runs slower than expected  \n**Diagnosis**: Check for:\n- Reactor concurrency limits that don't match your workload\n- Overly conservative concurrency limits\n- Sequential dependencies that prevent parallelisation\n\n**Problem**: High memory usage  \n**Diagnosis**: Check for:\n- Large intermediate results being stored unnecessarily\n- Map steps with batch sizes that are too large\n- Streams not being processed lazily\n\n**Problem**: External service rate limiting errors  \n**Solution**: Use compensation with exponential backoff and reduce `max_concurrency`:\n\n```elixir\nstep :api_call do\n  run fn args, _context ->\n    call_external_api(args)\n  end\n  \n  compensate fn _args, context ->\n    # Use current_try for exponential backoff\n    retry_count = context.current_try\n    delay_ms = :math.pow(2, retry_count) * 1000 |> round()\n    \n    Process.sleep(delay_ms)\n    :retry\n  end\nend\n\n# Also reduce concurrency to respect service limits\n{:ok, result} = Reactor.run(\n  APIReactor,\n  inputs,\n  %{},\n  max_concurrency: 5  # Lower concurrency for rate-limited APIs\n)\n```\n\n**Problem**: Inconsistent performance  \n**Solution**: Ensure proper resource isolation and monitoring:\n\n```elixir\n# Use dedicated concurrency pools for different workload types\nfast_pool = Reactor.Executor.ConcurrencyTracker.allocate_pool(100)\nslow_pool = Reactor.Executor.ConcurrencyTracker.allocate_pool(10)\n\n# Fast operations\nReactor.run(FastReactor, data, %{}, concurrency_key: fast_pool)\n\n# Slow operations  \nReactor.run(SlowReactor, data, %{}, concurrency_key: slow_pool)\n```","ref":"performance-optimization.html#common-issues-and-solutions"},{"type":"extras","title":"Summary - How to Optimize Reactor Performance","doc":"Optimizing Reactor performance requires understanding your workload characteristics and applying appropriate strategies:\n\n**For I/O-bound workloads:**\n- Use higher concurrency limits (4x CPU cores)\n- Keep steps async by default\n- Use shared concurrency pools to prevent resource competition\n\n**For CPU-bound workloads:**\n- Tune reactor concurrency limits to match CPU cores\n- Keep steps async but control resource usage via reactor max_concurrency\n- Use smaller batch sizes in map operations\n\n**For mixed workloads:**\n- Separate I/O and CPU operations\n- Use moderate concurrency limits (2x CPU cores)\n- Monitor performance with telemetry\n\n**For memory efficiency:**\n- Process data in streams when possible\n- Avoid storing large intermediate results\n- Use appropriate batch sizes for your data characteristics","ref":"performance-optimization.html#summary"},{"type":"extras","title":"Related Guides - How to Optimize Reactor Performance","doc":"- [Building Async Workflows](../tutorials/03-async-workflows.md) - Understanding Reactor's concurrency model\n- [Data Processing Pipelines](data-pipelines.md) - Efficient batch processing patterns\n- [Testing Strategies](testing-strategies.md) - Performance testing approaches\n- [Debugging Workflows](debugging-workflows.md) - Performance monitoring and profiling","ref":"performance-optimization.html#related-guides"},{"type":"DSL","title":" - dsl-reactor","doc":"The top-level reactor DSL","ref":"dsl-reactor.html#reactor-"},{"type":"DSL","title":"return - dsl-reactor","doc":"Specify which step result to return upon completion.","ref":"dsl-reactor.html#reactor-return"},{"type":"DSL","title":"around - dsl-reactor","doc":"Wrap a function around a group of steps.","ref":"dsl-reactor.html#reactor-around"},{"type":"DSL","title":"around.name - dsl-reactor","doc":"A unique name for the group of steps.","ref":"dsl-reactor.html#reactor-around-name"},{"type":"DSL","title":"around.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-around-description"},{"type":"DSL","title":"around.fun - dsl-reactor","doc":"The around function. See `Reactor.Step.Around` for more information.","ref":"dsl-reactor.html#reactor-around-fun"},{"type":"DSL","title":"around.allow_async? - dsl-reactor","doc":"Whether the emitted steps should be allowed to run asynchronously.","ref":"dsl-reactor.html#reactor-around-allow_async?"},{"type":"DSL","title":"around.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-around-argument"},{"type":"DSL","title":"around.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-around-argument-description"},{"type":"DSL","title":"around.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-around-argument-name"},{"type":"DSL","title":"around.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-around-argument-source"},{"type":"DSL","title":"around.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-around-argument-transform"},{"type":"DSL","title":"around.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-around-wait_for"},{"type":"DSL","title":"around.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-around-wait_for-names"},{"type":"DSL","title":"around.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-around-wait_for-description"},{"type":"DSL","title":"around.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-around-where"},{"type":"DSL","title":"around.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-around-where-predicate"},{"type":"DSL","title":"around.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-around-where-description"},{"type":"DSL","title":"around.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-around-guard"},{"type":"DSL","title":"around.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-around-guard-fun"},{"type":"DSL","title":"around.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-around-guard-description"},{"type":"DSL","title":"collect - dsl-reactor","doc":"A Reactor step which simply collects and returns it's arguments.\n\nArguments can optionally be transformed before returning.","ref":"dsl-reactor.html#reactor-collect"},{"type":"DSL","title":"collect.name - dsl-reactor","doc":"A unique name for the step. Used when choosing the return value of the Reactor and for arguments into other steps.","ref":"dsl-reactor.html#reactor-collect-name"},{"type":"DSL","title":"collect.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-collect-description"},{"type":"DSL","title":"collect.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the entire argument map before it is returned.","ref":"dsl-reactor.html#reactor-collect-transform"},{"type":"DSL","title":"collect.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-collect-argument"},{"type":"DSL","title":"collect.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-collect-argument-description"},{"type":"DSL","title":"collect.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-collect-argument-name"},{"type":"DSL","title":"collect.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-collect-argument-source"},{"type":"DSL","title":"collect.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-collect-argument-transform"},{"type":"DSL","title":"collect.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-collect-wait_for"},{"type":"DSL","title":"collect.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-collect-wait_for-names"},{"type":"DSL","title":"collect.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-collect-wait_for-description"},{"type":"DSL","title":"collect.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-collect-where"},{"type":"DSL","title":"collect.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-collect-where-predicate"},{"type":"DSL","title":"collect.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-collect-where-description"},{"type":"DSL","title":"collect.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-collect-guard"},{"type":"DSL","title":"collect.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-collect-guard-fun"},{"type":"DSL","title":"collect.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-collect-guard-description"},{"type":"DSL","title":"compose - dsl-reactor","doc":"Compose another Reactor into this one.\n\nAllows place another Reactor into this one as if it were a single step.","ref":"dsl-reactor.html#reactor-compose"},{"type":"DSL","title":"compose.name - dsl-reactor","doc":"A unique name for the step. Allows the result of the composed reactor to be depended upon by steps in this reactor.","ref":"dsl-reactor.html#reactor-compose-name"},{"type":"DSL","title":"compose.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-compose-description"},{"type":"DSL","title":"compose.reactor - dsl-reactor","doc":"The reactor module or struct to compose upon.","ref":"dsl-reactor.html#reactor-compose-reactor"},{"type":"DSL","title":"compose.async? - dsl-reactor","doc":"Whether the composed steps should be run asynchronously.","ref":"dsl-reactor.html#reactor-compose-async?"},{"type":"DSL","title":"compose.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-compose-argument"},{"type":"DSL","title":"compose.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-compose-argument-description"},{"type":"DSL","title":"compose.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-compose-argument-name"},{"type":"DSL","title":"compose.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-compose-argument-source"},{"type":"DSL","title":"compose.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-compose-argument-transform"},{"type":"DSL","title":"compose.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-compose-wait_for"},{"type":"DSL","title":"compose.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-compose-wait_for-names"},{"type":"DSL","title":"compose.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-compose-wait_for-description"},{"type":"DSL","title":"compose.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-compose-where"},{"type":"DSL","title":"compose.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-compose-where-predicate"},{"type":"DSL","title":"compose.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-compose-where-description"},{"type":"DSL","title":"compose.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-compose-guard"},{"type":"DSL","title":"compose.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-compose-guard-fun"},{"type":"DSL","title":"compose.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-compose-guard-description"},{"type":"DSL","title":"debug - dsl-reactor","doc":"Inserts a step which will send debug information to the `Logger`.","ref":"dsl-reactor.html#reactor-debug"},{"type":"DSL","title":"debug.name - dsl-reactor","doc":"A unique identifier for the step.","ref":"dsl-reactor.html#reactor-debug-name"},{"type":"DSL","title":"debug.level - dsl-reactor","doc":"The log level to send the debug information to.","ref":"dsl-reactor.html#reactor-debug-level"},{"type":"DSL","title":"debug.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-debug-description"},{"type":"DSL","title":"debug.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-debug-argument"},{"type":"DSL","title":"debug.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-debug-argument-description"},{"type":"DSL","title":"debug.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-debug-argument-name"},{"type":"DSL","title":"debug.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-debug-argument-source"},{"type":"DSL","title":"debug.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-debug-argument-transform"},{"type":"DSL","title":"debug.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-debug-wait_for"},{"type":"DSL","title":"debug.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-debug-wait_for-names"},{"type":"DSL","title":"debug.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-debug-wait_for-description"},{"type":"DSL","title":"debug.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-debug-where"},{"type":"DSL","title":"debug.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-debug-where-predicate"},{"type":"DSL","title":"debug.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-debug-where-description"},{"type":"DSL","title":"debug.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-debug-guard"},{"type":"DSL","title":"debug.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-debug-guard-fun"},{"type":"DSL","title":"debug.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-debug-guard-description"},{"type":"DSL","title":"flunk - dsl-reactor","doc":"Creates a step which will always cause the Reactor to exit with an error.\n\nThis step will flunk with a `Reactor.Error.Invalid.ForcedFailureError` with it's message set to the provided message.\nAdditionally, any arguments to the step will be stored in the exception under the `arguments` key.","ref":"dsl-reactor.html#reactor-flunk"},{"type":"DSL","title":"flunk.name - dsl-reactor","doc":"A unique name for the step. Used when choosing the return value of the Reactor and for arguments into other steps.","ref":"dsl-reactor.html#reactor-flunk-name"},{"type":"DSL","title":"flunk.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-flunk-description"},{"type":"DSL","title":"flunk.message - dsl-reactor","doc":"The message to to attach to the exception.","ref":"dsl-reactor.html#reactor-flunk-message"},{"type":"DSL","title":"flunk.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-flunk-argument"},{"type":"DSL","title":"flunk.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-flunk-argument-description"},{"type":"DSL","title":"flunk.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-flunk-argument-name"},{"type":"DSL","title":"flunk.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-flunk-argument-source"},{"type":"DSL","title":"flunk.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-flunk-argument-transform"},{"type":"DSL","title":"flunk.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-flunk-wait_for"},{"type":"DSL","title":"flunk.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-flunk-wait_for-names"},{"type":"DSL","title":"flunk.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-flunk-wait_for-description"},{"type":"DSL","title":"flunk.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-flunk-where"},{"type":"DSL","title":"flunk.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-flunk-where-predicate"},{"type":"DSL","title":"flunk.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-flunk-where-description"},{"type":"DSL","title":"flunk.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-flunk-guard"},{"type":"DSL","title":"flunk.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-flunk-guard-fun"},{"type":"DSL","title":"flunk.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-flunk-guard-description"},{"type":"DSL","title":"group - dsl-reactor","doc":"Call functions before and after a group of steps.","ref":"dsl-reactor.html#reactor-group"},{"type":"DSL","title":"group.name - dsl-reactor","doc":"A unique name for the group of steps.","ref":"dsl-reactor.html#reactor-group-name"},{"type":"DSL","title":"group.before_all - dsl-reactor","doc":"The before function. See `Reactor.Step.Group` for more information.","ref":"dsl-reactor.html#reactor-group-before_all"},{"type":"DSL","title":"group.after_all - dsl-reactor","doc":"The after function. See `Reactor.Step.Group` for more information.","ref":"dsl-reactor.html#reactor-group-after_all"},{"type":"DSL","title":"group.allow_async? - dsl-reactor","doc":"Whether the emitted steps should be allowed to run asynchronously.","ref":"dsl-reactor.html#reactor-group-allow_async?"},{"type":"DSL","title":"group.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-group-description"},{"type":"DSL","title":"group.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-group-argument"},{"type":"DSL","title":"group.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-group-argument-description"},{"type":"DSL","title":"group.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-group-argument-name"},{"type":"DSL","title":"group.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-group-argument-source"},{"type":"DSL","title":"group.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-group-argument-transform"},{"type":"DSL","title":"group.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-group-wait_for"},{"type":"DSL","title":"group.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-group-wait_for-names"},{"type":"DSL","title":"group.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-group-wait_for-description"},{"type":"DSL","title":"group.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-group-where"},{"type":"DSL","title":"group.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-group-where-predicate"},{"type":"DSL","title":"group.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-group-where-description"},{"type":"DSL","title":"group.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-group-guard"},{"type":"DSL","title":"group.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-group-guard-fun"},{"type":"DSL","title":"group.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-group-guard-description"},{"type":"DSL","title":"input - dsl-reactor","doc":"Specifies an input to the Reactor.\n\nAn input is a value passed in to the Reactor when executing.\nIf a Reactor were a function, these would be it's arguments.\n\nInputs can be transformed with an arbitrary function before being passed\nto any steps.","ref":"dsl-reactor.html#reactor-input"},{"type":"DSL","title":"input.name - dsl-reactor","doc":"A unique name for this input. Used to allow steps to depend on it.","ref":"dsl-reactor.html#reactor-input-name"},{"type":"DSL","title":"input.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the input before it is passed to any steps.","ref":"dsl-reactor.html#reactor-input-transform"},{"type":"DSL","title":"input.description - dsl-reactor","doc":"An optional description for the input.","ref":"dsl-reactor.html#reactor-input-description"},{"type":"DSL","title":"map - dsl-reactor","doc":"Execute nested steps for every item of an iterator.\n\nAllows you to \"map over\" a collection applying a some steps to each element,\nreturning a list of new values.  The input collection must be bounded - ie you\ncannot pass infinite streams into this step or it will just loop forever - and\nbecause it has to keep the results from each batch will eventually just use up\nall available RAM.\n\nUnder the hood we use `Iter` to handle lazy iteration of the collection.  This\nmeans that you can pass an `Iter.t` or any value for which `Iter.IntoIterable`\nis implemented.\n\n> #### A note on ordering {: .tip}\n>\n> If your application doesn't need the results back in the same order that they\n> were provided then setting `strict_ordering?` to `false` will increase\n> performance - especially on large input sets.","ref":"dsl-reactor.html#reactor-map"},{"type":"DSL","title":"map.name - dsl-reactor","doc":"A unique name for the step.","ref":"dsl-reactor.html#reactor-map-name"},{"type":"DSL","title":"map.allow_async? - dsl-reactor","doc":"Whether the emitted steps should be allowed to run asynchronously.","ref":"dsl-reactor.html#reactor-map-allow_async?"},{"type":"DSL","title":"map.batch_size - dsl-reactor","doc":"The number of items to consume off the source when emitting steps.","ref":"dsl-reactor.html#reactor-map-batch_size"},{"type":"DSL","title":"map.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-map-description"},{"type":"DSL","title":"map.source - dsl-reactor","doc":"The iterator or enumerable to use as the source of the iteration.","ref":"dsl-reactor.html#reactor-map-source"},{"type":"DSL","title":"map.return - dsl-reactor","doc":"The name of the nested step to use as the return value.","ref":"dsl-reactor.html#reactor-map-return"},{"type":"DSL","title":"map.strict_ordering? - dsl-reactor","doc":"Whether the mapped values must be returned in the same order that they were provided.","ref":"dsl-reactor.html#reactor-map-strict_ordering?"},{"type":"DSL","title":"map.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-map-argument"},{"type":"DSL","title":"map.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-map-argument-description"},{"type":"DSL","title":"map.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-map-argument-name"},{"type":"DSL","title":"map.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-map-argument-source"},{"type":"DSL","title":"map.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-map-argument-transform"},{"type":"DSL","title":"map.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-map-wait_for"},{"type":"DSL","title":"map.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-map-wait_for-names"},{"type":"DSL","title":"map.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-map-wait_for-description"},{"type":"DSL","title":"map.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-map-where"},{"type":"DSL","title":"map.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-map-where-predicate"},{"type":"DSL","title":"map.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-map-where-description"},{"type":"DSL","title":"map.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-map-guard"},{"type":"DSL","title":"map.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-map-guard-fun"},{"type":"DSL","title":"map.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-map-guard-description"},{"type":"DSL","title":"recurse - dsl-reactor","doc":"Execute a reactor recursively until an exit condition is met or maximum iterations are reached.\n\nAllows the output of one iteration to become the input of the next. The reactor must have\ninput/output compatibility to ensure that the outputs from iteration i can be fed into\niteration i+1.","ref":"dsl-reactor.html#reactor-recurse"},{"type":"DSL","title":"recurse.name - dsl-reactor","doc":"A unique name for the step. Allows the result of the recursed reactor to be depended upon by steps in this reactor.","ref":"dsl-reactor.html#reactor-recurse-name"},{"type":"DSL","title":"recurse.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-recurse-description"},{"type":"DSL","title":"recurse.reactor - dsl-reactor","doc":"The reactor module or struct to recurse on.","ref":"dsl-reactor.html#reactor-recurse-reactor"},{"type":"DSL","title":"recurse.async? - dsl-reactor","doc":"Whether the recursed steps should be run asynchronously.","ref":"dsl-reactor.html#reactor-recurse-async?"},{"type":"DSL","title":"recurse.max_iterations - dsl-reactor","doc":"The maximum number of iterations to execute. If not specified, will rely solely on exit_condition.","ref":"dsl-reactor.html#reactor-recurse-max_iterations"},{"type":"DSL","title":"recurse.exit_condition - dsl-reactor","doc":"A function that takes the result of an iteration and returns a boolean. When true, the recursion will stop. If not provided, will rely solely on max_iterations.","ref":"dsl-reactor.html#reactor-recurse-exit_condition"},{"type":"DSL","title":"recurse.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-recurse-argument"},{"type":"DSL","title":"recurse.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-recurse-argument-description"},{"type":"DSL","title":"recurse.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-recurse-argument-name"},{"type":"DSL","title":"recurse.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-recurse-argument-source"},{"type":"DSL","title":"recurse.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-recurse-argument-transform"},{"type":"DSL","title":"recurse.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-recurse-wait_for"},{"type":"DSL","title":"recurse.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-recurse-wait_for-names"},{"type":"DSL","title":"recurse.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-recurse-wait_for-description"},{"type":"DSL","title":"recurse.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-recurse-where"},{"type":"DSL","title":"recurse.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-recurse-where-predicate"},{"type":"DSL","title":"recurse.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-recurse-where-description"},{"type":"DSL","title":"recurse.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-recurse-guard"},{"type":"DSL","title":"recurse.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-recurse-guard-fun"},{"type":"DSL","title":"recurse.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-recurse-guard-description"},{"type":"DSL","title":"step - dsl-reactor","doc":"Specifies a Reactor step.\n\nSteps are the unit of work in a Reactor.  Reactor will calculate the\ndependencies graph between the steps and execute as many as it can in each\niteration.\n\nSee the `Reactor.Step` behaviour for more information.","ref":"dsl-reactor.html#reactor-step"},{"type":"DSL","title":"step.name - dsl-reactor","doc":"A unique name for the step. Used when choosing the return value of the Reactor and for arguments into other steps.","ref":"dsl-reactor.html#reactor-step-name"},{"type":"DSL","title":"step.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-step-description"},{"type":"DSL","title":"step.impl - dsl-reactor","doc":"A module that implements the `Reactor.Step` behaviour that provides the implementation.","ref":"dsl-reactor.html#reactor-step-impl"},{"type":"DSL","title":"step.run - dsl-reactor","doc":"Provide an anonymous function which implements a `run/1-2` callback. Cannot be provided at the same time as the `impl` argument.","ref":"dsl-reactor.html#reactor-step-run"},{"type":"DSL","title":"step.undo - dsl-reactor","doc":"Provide an anonymous function which implements a `undo/1-3` callback. Cannot be provided at the same time as the `impl` argument.","ref":"dsl-reactor.html#reactor-step-undo"},{"type":"DSL","title":"step.compensate - dsl-reactor","doc":"Provide an anonymous function which implements a `compensate/1-3` callback. Cannot be provided at the same time as the `impl` argument.","ref":"dsl-reactor.html#reactor-step-compensate"},{"type":"DSL","title":"step.max_retries - dsl-reactor","doc":"The maximum number of times that the step can be retried before failing. Only used when the result of the `compensate` callback is `:retry`.","ref":"dsl-reactor.html#reactor-step-max_retries"},{"type":"DSL","title":"step.async? - dsl-reactor","doc":"When set to true the step will be executed asynchronously via Reactor's `TaskSupervisor`.","ref":"dsl-reactor.html#reactor-step-async?"},{"type":"DSL","title":"step.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the entire argument map before it is passed to the step.","ref":"dsl-reactor.html#reactor-step-transform"},{"type":"DSL","title":"step.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-step-argument"},{"type":"DSL","title":"step.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-step-argument-description"},{"type":"DSL","title":"step.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-step-argument-name"},{"type":"DSL","title":"step.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-step-argument-source"},{"type":"DSL","title":"step.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-step-argument-transform"},{"type":"DSL","title":"step.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-step-wait_for"},{"type":"DSL","title":"step.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-step-wait_for-names"},{"type":"DSL","title":"step.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-step-wait_for-description"},{"type":"DSL","title":"step.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-step-where"},{"type":"DSL","title":"step.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-step-where-predicate"},{"type":"DSL","title":"step.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-step-where-description"},{"type":"DSL","title":"step.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-step-guard"},{"type":"DSL","title":"step.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-step-guard-fun"},{"type":"DSL","title":"step.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-step-guard-description"},{"type":"DSL","title":"switch - dsl-reactor","doc":"Use a predicate to determine which steps should be executed.","ref":"dsl-reactor.html#reactor-switch"},{"type":"DSL","title":"switch.name - dsl-reactor","doc":"A unique name for the switch.","ref":"dsl-reactor.html#reactor-switch-name"},{"type":"DSL","title":"switch.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-switch-description"},{"type":"DSL","title":"switch.allow_async? - dsl-reactor","doc":"Whether the emitted steps should be allowed to run asynchronously.","ref":"dsl-reactor.html#reactor-switch-allow_async?"},{"type":"DSL","title":"switch.on - dsl-reactor","doc":"The value to match against.","ref":"dsl-reactor.html#reactor-switch-on"},{"type":"DSL","title":"switch.matches? - dsl-reactor","doc":"A group of steps to run when the predicate matches.","ref":"dsl-reactor.html#reactor-switch-matches?"},{"type":"DSL","title":"switch.matches?.predicate - dsl-reactor","doc":"A one-arity function which is used to match the switch input. If the switch returns a truthy value, then the nested steps will be run.","ref":"dsl-reactor.html#reactor-switch-matches?-predicate"},{"type":"DSL","title":"switch.matches?.allow_async? - dsl-reactor","doc":"Whether the emitted steps should be allowed to run asynchronously.","ref":"dsl-reactor.html#reactor-switch-matches?-allow_async?"},{"type":"DSL","title":"switch.matches?.return - dsl-reactor","doc":"Specify which step result to return upon completion.","ref":"dsl-reactor.html#reactor-switch-matches?-return"},{"type":"DSL","title":"switch.default - dsl-reactor","doc":"If none of the `matches?` branches match the input, then the `default`\nsteps will be run if provided.","ref":"dsl-reactor.html#reactor-switch-default"},{"type":"DSL","title":"switch.default.return - dsl-reactor","doc":"Specify which step result to return upon completion.","ref":"dsl-reactor.html#reactor-switch-default-return"},{"type":"DSL","title":"template - dsl-reactor","doc":"A step which passes all it's arguments as assigns into an `EEx` template and returns the result.","ref":"dsl-reactor.html#reactor-template"},{"type":"DSL","title":"template.name - dsl-reactor","doc":"A unique name for the step. Used when choosing the return value of the Reactor and for arguments into other steps.","ref":"dsl-reactor.html#reactor-template-name"},{"type":"DSL","title":"template.description - dsl-reactor","doc":"An optional description for the step.","ref":"dsl-reactor.html#reactor-template-description"},{"type":"DSL","title":"template.template - dsl-reactor","doc":"An `EEx` template.","ref":"dsl-reactor.html#reactor-template-template"},{"type":"DSL","title":"template.argument - dsl-reactor","doc":"Specifies an argument to a Reactor step.\n\nEach argument is a value which is either the result of another step, or an input value.\n\nIndividual arguments can be transformed with an arbitrary function before\nbeing passed to any steps.","ref":"dsl-reactor.html#reactor-template-argument"},{"type":"DSL","title":"template.argument.description - dsl-reactor","doc":"An optional description for the argument.","ref":"dsl-reactor.html#reactor-template-argument-description"},{"type":"DSL","title":"template.argument.name - dsl-reactor","doc":"The name of the argument which will be used as the key in the `arguments` map passed to the implementation.","ref":"dsl-reactor.html#reactor-template-argument-name"},{"type":"DSL","title":"template.argument.source - dsl-reactor","doc":"What to use as the source of the argument. See `Reactor.Dsl.Argument` for more information.","ref":"dsl-reactor.html#reactor-template-argument-source"},{"type":"DSL","title":"template.argument.transform - dsl-reactor","doc":"An optional transformation function which can be used to modify the argument before it is passed to the step.","ref":"dsl-reactor.html#reactor-template-argument-transform"},{"type":"DSL","title":"template.wait_for - dsl-reactor","doc":"Wait for the named step to complete before allowing this one to start.\n\nDesugars to `argument :_, result(step_to_wait_for)`","ref":"dsl-reactor.html#reactor-template-wait_for"},{"type":"DSL","title":"template.wait_for.names - dsl-reactor","doc":"The name of the step to wait for.","ref":"dsl-reactor.html#reactor-template-wait_for-names"},{"type":"DSL","title":"template.wait_for.description - dsl-reactor","doc":"An optional description.","ref":"dsl-reactor.html#reactor-template-wait_for-description"},{"type":"DSL","title":"template.where - dsl-reactor","doc":"Only execute the surrounding step if the predicate function returns true.\n\nThis is a simple version of `guard` which provides more flexibility at the cost of complexity.","ref":"dsl-reactor.html#reactor-template-where"},{"type":"DSL","title":"template.where.predicate - dsl-reactor","doc":"Provide a function which takes the step arguments and optionally the context and returns a boolean value.","ref":"dsl-reactor.html#reactor-template-where-predicate"},{"type":"DSL","title":"template.where.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-template-where-description"},{"type":"DSL","title":"template.guard - dsl-reactor","doc":"Provides a flexible method for conditionally executing a step, or replacing it's result.\n\nExpects a two arity function which takes the step's arguments and context and returns one of the following:\n\n- `:cont` - the guard has passed.\n- `{:halt, result}` - the guard has failed - instead of executing the step use the provided result.","ref":"dsl-reactor.html#reactor-template-guard"},{"type":"DSL","title":"template.guard.fun - dsl-reactor","doc":"The guard function.","ref":"dsl-reactor.html#reactor-template-guard-fun"},{"type":"DSL","title":"template.guard.description - dsl-reactor","doc":"An optional description of the guard.","ref":"dsl-reactor.html#reactor-template-guard-description"},{"type":"DSL","title":"middlewares - dsl-reactor","doc":"Middleware to be added to the Reactor","ref":"dsl-reactor.html#reactor-middlewares"},{"type":"DSL","title":"middlewares.middleware - dsl-reactor","doc":"Name a middleware to be added to the Reactor.","ref":"dsl-reactor.html#reactor-middlewares-middleware"},{"type":"DSL","title":"middlewares.middleware.module - dsl-reactor","doc":"The middleware to be added to the Reactor.","ref":"dsl-reactor.html#reactor-middlewares-middleware-module"},{"type":"DSL","title":"middlewares.middleware.description - dsl-reactor","doc":"An optional description for the middleware.","ref":"dsl-reactor.html#reactor-middlewares-middleware-description"},{"type":"extras","title":"Reactor Glossary","doc":"# Reactor Glossary\n\nThis glossary defines key terms, concepts, and technical vocabulary used throughout Reactor documentation.","ref":"glossary.html"},{"type":"extras","title":"Core Reactor Concepts - Reactor Glossary","doc":"**Argument** - A dependency declaration in a step that specifies what data the step needs and where it comes from (inputs, results from other steps, or static values).\n\n**Argument Transformation** - A function applied to an argument value before it's passed to a step, automatically extracted as separate transform steps during planning.\n\n**Compensation** - Error handling mechanism where a step defines how to handle its own failures, returning `:retry`, `:ok`, `{:continue, value}`, or `{:error, reason}`.\n\n**Compose** - A DSL step type that embeds another reactor as a single step, allowing hierarchical workflow composition.\n\n**Context** - Runtime execution environment shared across steps, containing user data, step metadata, retry information, and concurrency details.\n\n**Dependency Graph** - A directed acyclic graph (DAG) where vertices represent steps and edges represent dependencies, used to determine execution order.\n\n**DSL (Domain Specific Language)** - Declarative syntax built with Spark for defining reactors with inputs, steps, and their relationships.\n\n**Dynamic Step Creation** - The ability for steps to emit new steps during execution using `{:ok, result, new_steps}` return format.\n\n**Input** - A named parameter that a reactor accepts when executed, similar to function arguments.\n\n**Intermediate Results** - Storage for step outputs that are needed by dependent steps or the reactor's return value.\n\n**Reactor** - A workflow definition containing inputs, steps, dependencies, and execution logic for orchestrating complex business processes.\n\n**Result** - The output value from a successfully executed step, accessible to dependent steps via `result(:step_name)` syntax.\n\n**Saga Pattern** - Transaction-like coordination pattern across multiple resources without requiring distributed transactions, using compensation for failure handling.\n\n**Step** - A unit of work in a reactor with a unique name, dependencies, implementation, and optional error handling callbacks.\n\n**Subpath Access** - Ability to extract nested values from step results using syntax like `result(:step, [:key, :subkey])`.\n\n**Undo** - Rollback mechanism called when a step succeeded but a later step failed, used to maintain system consistency.","ref":"glossary.html#core-reactor-concepts"},{"type":"extras","title":"Technical Architecture Terms - Reactor Glossary","doc":"**Concurrency Pool** - Shared resource allocation system that limits concurrent step execution across reactor hierarchies to prevent resource exhaustion.\n\n**Concurrency Tracker** - ETS-based global system managing concurrency pools with atomic operations and process monitoring.\n\n**Cycle Detection** - Validation process ensuring the dependency graph contains no circular dependencies that would create deadlocks.\n\n**DAG (Directed Acyclic Graph)** - Mathematical structure representing step dependencies where edges flow from prerequisite steps to dependent steps.\n\n**Deadlock Prevention** - Automatic detection and mitigation of resource contention in nested reactor scenarios.\n\n**Dependency Resolution** - Process of determining when steps can execute based on availability of their required arguments.\n\n**Executor** - Core engine that coordinates reactor execution, managing concurrency, error handling, and step lifecycle.\n\n**libgraph** - External library used by Reactor for efficient DAG operations and dependency graph management.\n\n**Planner** - Component that converts step definitions into an executable dependency graph with cycle detection and ready state identification.\n\n**Ready Steps** - Steps with all dependencies satisfied that can begin execution immediately.\n\n**State Management** - System for tracking reactor execution state, intermediate results, undo stack, and concurrency information.\n\n**Step Runner** - Component handling individual step execution including argument collection, guard evaluation, and result processing.","ref":"glossary.html#technical-architecture-terms"},{"type":"extras","title":"DSL-Specific Terminology - Reactor Glossary","doc":"**Around Step** - DSL step type that wraps other steps with setup and teardown logic, similar to middleware.\n\n**Collect Step** - DSL step type that gathers multiple arguments into a single result, optionally applying transformations.\n\n**Debug Step** - Built-in step type that logs information to help with workflow debugging and monitoring.\n\n**Element** - Reference to current item in map step iterations using `element(:map_step_name)` syntax.\n\n**Flunk Step** - DSL step type that always fails with a specified error message, used for testing and controlled failures.\n\n**Group Step** - DSL step type that runs before/after functions around a collection of steps.\n\n**Guard** - Conditional execution mechanism that can halt step execution or provide alternative results based on runtime conditions.\n\n**Map Step** - DSL step type for processing collections by running nested steps for each item in the source data.\n\n**Middleware** - Cross-cutting concern components that can intercept and modify reactor execution at various points.\n\n**Recurse Step** - DSL step type for executing a reactor recursively until exit conditions are met or maximum iterations reached.\n\n**Switch Step** - DSL step type providing conditional logic with multiple branches based on predicate matching.\n\n**Template Step** - DSL step type that processes EEx templates with step arguments as assigns.\n\n**Wait For** - Dependency declaration that waits for step completion without using its result data.\n\n**Where Clause** - Simple conditional execution predicate that determines whether a step should run.","ref":"glossary.html#dsl-specific-terminology"},{"type":"extras","title":"Error Handling Concepts - Reactor Glossary","doc":"**Compensation Logic** - Step-level error handling that decides whether to retry, continue with alternative values, or fail.\n\n**Error Classification** - Categorisation of errors using splode library into Invalid, Internal, Unknown, and Validation types.\n\n**Exponential Backoff** - Retry strategy where delay increases exponentially with each retry attempt.\n\n**Max Retries** - Configuration limiting how many times a step can be retried through compensation.\n\n**Retry Logic** - Automatic re-execution of failed steps when compensation returns `:retry`.\n\n**Three-Tier Error Handling** - Reactor's approach using compensation (retry), undo (rollback), and global rollback levels.\n\n**Undo Stack** - Last-in-first-out collection of successfully completed undoable steps for rollback purposes.","ref":"glossary.html#error-handling-concepts"},{"type":"extras","title":"Concurrency and Execution Terms - Reactor Glossary","doc":"**Async by Default** - Reactor's philosophy of running steps asynchronously unless explicitly configured otherwise.\n\n**Concurrency Key** - Identifier for shared concurrency pools allowing resource coordination across reactor hierarchies.\n\n**Execution Loop** - Main algorithm prioritising async task completion, new task starts, and sync execution.\n\n**Max Concurrency** - Configuration parameter limiting simultaneous step execution (defaults to system CPU count).\n\n**Process Supervision** - Management of async step execution under Task.Supervisor with proper cleanup.\n\n**Resource Management** - System for allocation, tracking, and cleanup of concurrency slots and process resources.\n\n**Shared Pools** - Concurrency allocation mechanism allowing multiple reactors to coordinate resource usage.\n\n**Synchronous Execution** - Sequential step execution mode used for testing, resource constraints, or explicit requirements.\n\n**Task Supervision** - OTP pattern for managing async step execution with failure isolation and cleanup.","ref":"glossary.html#concurrency-and-execution-terms"},{"type":"extras","title":"Integration and Ecosystem Terms - Reactor Glossary","doc":"**Ash.Reactor** - Framework-specific extension providing deep integration with Ash resources, actions, and transactions. Available at [hex.pm/packages/ash](https://hex.pm/packages/ash).\n\n**Builder API** - Programmatic interface for constructing reactors at runtime using `Reactor.Builder` functions.\n\n**Ecosystem Extensions** - Additional packages that extend DSL capabilities:\n- [reactor_file](https://hex.pm/packages/reactor_file) - File system operations (copying, moving, permissions, I/O)\n- [reactor_process](https://hex.pm/packages/reactor_process) - Supervisor and process management operations  \n- [reactor_req](https://hex.pm/packages/reactor_req) - HTTP client steps with DSL for all HTTP methods\n\n**Framework Independence** - Design principle allowing Reactor to work with any Elixir application without framework dependencies.\n\n**Spark** - Foundation library providing DSL infrastructure shared across the Ash ecosystem.","ref":"glossary.html#integration-and-ecosystem-terms"},{"type":"extras","title":"Performance and Scalability Terms - Reactor Glossary","doc":"**Batch Size** - Configuration for map steps controlling how many items are processed together.\n\n**Bounded Resources** - Concurrency limits and resource constraints preventing system exhaustion.\n\n**Memory Optimisation** - Strategies for efficient intermediate result storage and cleanup.\n\n**Strict Ordering** - Map step configuration ensuring results are returned in input order (vs performance-optimised unordered).","ref":"glossary.html#performance-and-scalability-terms"},{"type":"extras","title":"Testing and Development Terms - Reactor Glossary","doc":"**Deterministic Execution** - Predictable step execution order achieved by disabling async mode for testing.\n\n**Integration Testing** - Testing complete reactor workflows end-to-end with real or mocked dependencies.\n\n**Mimic** - Recommended mocking library for testing step behaviour and error scenarios.\n\n**Unit Testing** - Testing individual step modules in isolation from reactor execution.","ref":"glossary.html#testing-and-development-terms"},{"type":"extras","title":"Core Reactor Concepts","doc":"# Core Reactor Concepts\n\nThis guide explains the fundamental concepts that make Reactor work, providing deep insight into how the framework approaches workflow orchestration, error handling, and concurrent execution.","ref":"concepts.html"},{"type":"extras","title":"Sagas and Compensation Patterns - Core Reactor Concepts","doc":"","ref":"concepts.html#sagas-and-compensation-patterns"},{"type":"extras","title":"What is the Saga Pattern? - Core Reactor Concepts","doc":"The Saga pattern provides transaction-like semantics across multiple distinct resources without requiring distributed transactions. Instead of traditional ACID properties, sagas use compensation to handle failures.\n\n```mermaid\ngraph TD\n    subgraph \"Forward Execution (Happy Path)\"\n        A[Step A: Reserve Inventory] --> B[Step B: Authorize Payment]\n        B --> C[Step C: Create Order]\n        C --> D[Step D: Send Confirmation]\n        D --> Success[Transaction Complete]\n    end\n    \n    subgraph \"Compensation Chain (Error Path)\"\n        FailPoint[Failure Point] --> CompD[Compensate D: Cancel Confirmation]\n        CompD --> CompC[Compensate C: Cancel Order]\n        CompC --> CompB[Compensate B: Void Payment]\n        CompB --> CompA[Compensate A: Release Inventory]\n        CompA --> Compensated[Transaction Compensated]\n    end\n    \n    %% Error flows\n    A -.->|Error| CompA\n    B -.->|Error| CompB\n    C -.->|Error| CompC\n    D -.->|Error| CompD\n    \n    style Success fill:#c8e6c9\n    style Compensated fill:#fff3e0\n    style FailPoint fill:#ffcdd2\n```\n\nIn Reactor, each step can optionally implement compensation logic that defines how to handle failures specific to that step's operation:\n\n```elixir\ndefmodule MyStep do\n  use Reactor.Step\n\n  def run(_arguments, _context, _step) do\n    # Main step logic that might fail\n    {:ok, result}\n  end\n\n  def compensate(reason, _arguments, _context, _step) do\n    # Handle the failure - can retry, provide fallback, or continue rollback\n    case reason do\n      %RetryableError{} -> :retry\n      %FallbackError{value: value} -> {:continue, value}\n      _other -> :ok  # Continue with rollback\n    end\n  end\nend\n```","ref":"concepts.html#what-is-the-saga-pattern"},{"type":"extras","title":"How Compensation Works - Core Reactor Concepts","doc":"When a step fails, Reactor automatically calls its `compensate/4` callback if it exists. The compensation function can return:\n\n- `:ok` - Accept the failure and continue rollback\n- `{:continue, value}` - Provide a replacement result and continue execution\n- `:retry` - Retry the failed step (up to `max_retries`)\n- `{:error, reason}` - Compensation itself failed\n\nThis approach gives each step control over how its failures are handled, enabling sophisticated error recovery strategies.","ref":"concepts.html#how-compensation-works"},{"type":"extras","title":"Directed Acyclic Graphs (DAGs) - Core Reactor Concepts","doc":"","ref":"concepts.html#directed-acyclic-graphs-dags"},{"type":"extras","title":"Dependency Graph Construction - Core Reactor Concepts","doc":"Reactor builds a directed acyclic graph where:\n- **Vertices** represent steps\n- **Edges** represent dependencies between steps\n- **Edge Labels** describe the dependency relationship\n\nDependencies are created automatically when step arguments reference results from other steps:\n\n```elixir\ndefmodule UserRegistrationReactor do\n  use Reactor\n\n  input :email\n  input :password\n\n  step :validate_email do\n    argument :email, input(:email)\n    run fn %{email: email}, _context ->\n      # Validation logic\n      {:ok, validated_email}\n    end\n  end\n\n  step :hash_password do\n    argument :password, input(:password)\n    # This step can run concurrently with :validate_email\n  end\n\n  step :create_user do\n    argument :email, result(:validate_email)     # Creates dependency edge\n    argument :password, result(:hash_password)   # Creates dependency edge\n    # This step waits for both validate_email and hash_password\n  end\nend\n```","ref":"concepts.html#dependency-graph-construction"},{"type":"extras","title":"Execution Order Determination - Core Reactor Concepts","doc":"The planner uses the DAG to determine which steps can run:\n\n1. **Ready Steps**: Steps with no unresolved dependencies (no in-edges)\n2. **Blocked Steps**: Steps waiting for dependency results\n3. **Concurrent Execution**: Multiple ready steps can run simultaneously\n\nThis dependency-driven approach maximises concurrency while ensuring correctness.","ref":"concepts.html#execution-order-determination"},{"type":"extras","title":"Cycle Detection - Core Reactor Concepts","doc":"Reactor validates that the dependency graph is acyclic during planning. Circular dependencies would create deadlocks, so they're detected and reported as errors before execution begins.\n\nThe one exception is when a step emits a new step with its own name during execution - this allows for iterative patterns where a step can retry itself or implement loops by re-emitting itself with updated context or arguments.","ref":"concepts.html#cycle-detection"},{"type":"extras","title":"Dependency Resolution - Core Reactor Concepts","doc":"","ref":"concepts.html#dependency-resolution"},{"type":"extras","title":"Argument Types - Core Reactor Concepts","doc":"Reactor supports three types of step arguments:\n\n1. **Input Arguments**: `input(:name)` - Values provided when starting the reactor\n2. **Result Arguments**: `result(:step_name)` - Outputs from other steps  \n3. **Value Arguments**: `value(x)` - Static values defined at build time","ref":"concepts.html#argument-types"},{"type":"extras","title":"Subpath Access - Core Reactor Concepts","doc":"You can access nested values from step results using subpaths:\n\n```elixir\nstep :process_user do\n  argument :user_id, result(:create_user, :id)        # Extract :id field\n  argument :profile, result(:create_user, [:profile, :data])  # Nested access\nend\n```","ref":"concepts.html#subpath-access"},{"type":"extras","title":"Argument Transformation - Core Reactor Concepts","doc":"Arguments can be transformed before being passed to steps:\n\n```elixir\nstep :send_email do\n  argument :email, result(:validate_email) do\n    transform &String.downcase/1\n  end\nend\n```","ref":"concepts.html#argument-transformation"},{"type":"extras","title":"Resolution Process - Core Reactor Concepts","doc":"During execution, Reactor resolves arguments by:\n\n1. **Input Arguments**: Looking up values in `reactor.context.private.inputs`\n2. **Result Arguments**: Looking up values in `reactor.intermediate_results`\n3. **Value Arguments**: Using the literal value provided\n4. **Applying Transformations**: If specified, transforming the resolved value\n\nNote that argument transformations are actually extracted into separate transform steps during the planning phase and inserted into the dependency graph. This means transforms become their own steps with proper dependency tracking, rather than being applied inline during argument resolution.","ref":"concepts.html#resolution-process"},{"type":"extras","title":"Concurrency Model - Core Reactor Concepts","doc":"","ref":"concepts.html#concurrency-model"},{"type":"extras","title":"Async by Default - Core Reactor Concepts","doc":"Steps run asynchronously by default to maximise throughput. This means:\n\n- Multiple steps can execute simultaneously\n- The reactor doesn't wait for one step to complete before starting another\n- Dependencies still ensure correct execution order","ref":"concepts.html#async-by-default"},{"type":"extras","title":"Concurrency Limits - Core Reactor Concepts","doc":"Reactor uses a concurrency tracker to prevent resource exhaustion:\n\n- **Default Limit**: `System.schedulers_online/0` (number of CPU cores)\n- **Shared Pools**: Multiple reactors can share concurrency limits using `concurrency_key`\n- **Custom Limits**: Configure per-reactor or globally","ref":"concepts.html#concurrency-limits"},{"type":"extras","title":"Task Supervision - Core Reactor Concepts","doc":"Async steps run as supervised tasks:\n\n- Each step runs under `Task.Supervisor`\n- Failed tasks don't crash the reactor\n- Proper cleanup on reactor termination","ref":"concepts.html#task-supervision"},{"type":"extras","title":"Synchronous Execution - Core Reactor Concepts","doc":"Some scenarios require synchronous execution:\n\n- **Explicit**: Set `async? false` on a step\n- **Resource Constraints**: When the concurrency pool is exhausted\n- **Reactor-Level**: Pass `async? false` when running the entire reactor (often used in tests for deterministic execution)","ref":"concepts.html#synchronous-execution"},{"type":"extras","title":"Error Handling Philosophy - Core Reactor Concepts","doc":"","ref":"concepts.html#error-handling-philosophy"},{"type":"extras","title":"Three-Level Error Handling - Core Reactor Concepts","doc":"Reactor implements a sophisticated three-level error handling strategy:\n\n#### 1. Compensation (Step-Level)\n\nWhen a step fails, its compensation function decides how to handle the failure:\n\n```elixir\ndef compensate(reason, arguments, context, step) do\n  case reason do\n    %NetworkError{} -> :retry  # Network issue, try again\n    %ValidationError{} -> :ok  # Invalid data, don't retry\n    %ServiceUnavailable{fallback: value} -> {:continue, value}  # Use fallback\n  end\nend\n```\n\n#### 2. Undo (Rollback Successful Work)\n\nWhen compensation can't resolve the error, Reactor undoes previously successful steps:\n\n```elixir\ndefmodule CreateUserStep do\n  def run(arguments, context, step) do\n    # Create user in database\n    {:ok, user}\n  end\n\n  def undo(user, arguments, context, step) do\n    # Remove the created user\n    Database.delete_user(user.id)\n    :ok\n  end\nend\n```\n\n#### 3. Global Rollback\n\nIf undo operations fail, the reactor enters a failed state with all error details preserved.","ref":"concepts.html#three-level-error-handling"},{"type":"extras","title":"Undo Stack Management - Core Reactor Concepts","doc":"Reactor maintains an undo stack of successfully completed undoable steps:\n\n- Steps with `undo/4` callbacks are automatically tracked\n- Steps are undone in reverse order (last-in-first-out)\n- Non-undoable steps are not tracked, saving memory","ref":"concepts.html#undo-stack-management"},{"type":"extras","title":"Error Classification - Core Reactor Concepts","doc":"Reactor uses the `splode` library to classify errors:\n\n- **Invalid**: Client errors (bad input, validation failures)\n- **Internal**: Server errors (bugs, unexpected conditions)  \n- **Unknown**: Unclassified errors\n- **Validation**: Schema or data validation errors\n\nThis classification helps determine appropriate error handling strategies.","ref":"concepts.html#error-classification"},{"type":"extras","title":"Retry Logic - Core Reactor Concepts","doc":"Steps can be retried automatically:\n\n- **Compensation Return**: Return `:retry` from compensation\n- **Retry Limits**: Configure `max_retries` per step (default: 5)\n- **Exponential Backoff**: Use `context.current_try` for delay calculation\n\n```elixir\ndef compensate(reason, arguments, context, step) do\n  if context.current_try < step.max_retries do\n    # Wait longer each retry\n    delay = :math.pow(2, context.current_try) * 1000\n    Process.sleep(delay)\n    :retry\n  else\n    :ok  # Give up after max retries\n  end\nend\n```","ref":"concepts.html#retry-logic"},{"type":"extras","title":"Integration Patterns - Core Reactor Concepts","doc":"","ref":"concepts.html#integration-patterns"},{"type":"extras","title":"Dynamic Step Creation - Core Reactor Concepts","doc":"Steps can create new steps during execution:\n\n```elixir\ndef run(arguments, context, step) do\n  # Process data and determine what steps to create\n  new_steps = build_dynamic_steps(arguments.data)\n  {:ok, result, new_steps}\nend\n```","ref":"concepts.html#dynamic-step-creation"},{"type":"extras","title":"Middleware Integration - Core Reactor Concepts","doc":"Reactor supports middleware for cross-cutting concerns:\n\n```elixir\nreactor \"instrumented_workflow\" do\n  middlewares do\n    middleware Reactor.Middleware.Telemetry\n  end\n  \n  # Steps automatically get telemetry events\nend\n```","ref":"concepts.html#middleware-integration"},{"type":"extras","title":"Composition with Sub-Reactors - Core Reactor Concepts","doc":"Complex workflows can be composed from smaller reactors:\n\n```elixir\nstep :user_onboarding do\n  argument :user_data, input(:user)\n  compose UserOnboardingReactor\nend\n```","ref":"concepts.html#composition-with-sub-reactors"},{"type":"extras","title":"Design Principles - Core Reactor Concepts","doc":"These concepts work together based on several key principles:\n\n1. **Dependency-Driven Execution**: Let data dependencies determine execution order\n2. **Local Error Handling**: Each step knows best how to handle its own failures\n3. **Automatic Rollback**: The system handles cleanup when things go wrong\n4. **Concurrent by Default**: Maximise throughput while ensuring correctness\n5. **Composable Workflows**: Build complex systems from simple, reusable components\n\nUnderstanding these concepts provides the foundation for designing effective workflows with Reactor, enabling you to build robust, concurrent, and fault-tolerant systems.","ref":"concepts.html#design-principles"},{"type":"extras","title":"Reactor Architecture","doc":"# Reactor Architecture\n\nThis guide explains how Reactor works internally, providing insight into its components, execution flow, and design patterns. Understanding the architecture helps you make informed decisions about workflow design, performance optimization, and debugging.","ref":"architecture.html"},{"type":"extras","title":"Core Components - Reactor Architecture","doc":"","ref":"architecture.html#core-components"},{"type":"extras","title":"Architecture Overview - Reactor Architecture","doc":"Reactor's architecture follows a clear separation of concerns between planning, execution, and state management:\n\n```mermaid\ngraph TB\n    subgraph \"Reactor Framework\"\n        DSL[DSL Definition] --> Parser[DSL Parser]\n        Parser --> Builder[Reactor Builder]\n        Builder --> Planner[Execution Planner]\n        Planner --> Executor[Concurrent Executor]\n        \n        subgraph \"Core Components\"\n            Steps[Step Definitions]\n            Args[Argument Dependencies]\n            Guards[Guard Conditions]\n            Middleware[Middleware Stack]\n        end\n        \n        Steps --> Builder\n        Args --> Builder\n        Guards --> Builder\n        Middleware --> Executor\n        \n        Executor --> Results[Execution Results]\n    end\n    \n    subgraph \"External World\"\n        Inputs[Input Data] --> DSL\n        Results --> Outputs[Output Data]\n        \n        APIs[External APIs]\n        DBs[Databases]\n        Files[File Systems]\n        \n        Executor <--> APIs\n        Executor <--> DBs\n        Executor <--> Files\n    end\n    \n    style DSL fill:#e1f5fe\n    style Executor fill:#fff3e0\n    style Results fill:#c8e6c9\n```","ref":"architecture.html#architecture-overview"},{"type":"extras","title":"The Reactor Struct - Reactor Architecture","doc":"At the heart of Reactor is the `Reactor` struct, which contains all the information needed for workflow execution:\n\n```elixir\ndefstruct context: %{},\n          id: nil,\n          input_descriptions: %{},\n          inputs: [],\n          intermediate_results: %{},\n          middleware: [],\n          plan: nil,\n          return: nil,\n          state: :pending,\n          steps: [],\n          undo: []\n```\n\n**Key Components:**\n- **Context**: Runtime context shared across all steps\n- **State**: Reactor lifecycle (`:pending`, `:executing`, `:halted`, `:failed`, `:successful`)\n- **Intermediate Results**: Step outputs stored for dependency resolution\n- **Undo Stack**: Successful undoable steps tracked for rollback\n- **Plan**: The dependency graph once constructed","ref":"architecture.html#the-reactor-struct"},{"type":"extras","title":"DSL Implementation - Reactor Architecture","doc":"Reactor's DSL is built with Spark, providing a declarative way to define workflows:\n\n#### DSL Processing Flow\n\n1. **Entity Collection**: DSL entities (steps, inputs, compose) are parsed during compilation\n2. **Transformation**: The DSL transformer validates and processes entities:\n   - Ensures step names are unique\n   - Sets default return value to the last step\n   - Plans the reactor at compile-time\n3. **Code Generation**: Creates a `reactor/0` function returning the planned struct\n4. **Runtime Conversion**: `Reactor.Info.to_struct/1` builds the executable reactor\n\n#### DSL Architecture\n\n```\nDSL Definition → Spark Entities → Transformer → Validation → Runtime Struct\n```\n\nThe transformer ensures reactors are validated and optimised at compile-time, catching errors early and enabling efficient execution.","ref":"architecture.html#dsl-implementation"},{"type":"extras","title":"Executor Components - Reactor Architecture","doc":"The executor system coordinates all aspects of reactor execution through several specialised components:\n\n#### Main Executor (`Reactor.Executor`)\n\nImplements the core execution algorithm with sophisticated async/sync coordination:\n\n**Execution Loop Priority:**\n1. Process completed async tasks first (maximum concurrency)\n2. Start new async steps (within concurrency limits)\n3. Run one sync step if available\n4. Handle deadlock prevention for shared pools\n5. Check completion conditions\n\n**Key Features:**\n- **Concurrency Management**: Respects `max_concurrency` and shared pools\n- **Deadlock Prevention**: Runs async steps synchronously when pools exhausted\n- **Timeout Handling**: Respects execution timeouts\n- **Error Orchestration**: Coordinates undo operations on failure\n\n#### Step Runner (`Reactor.Executor.StepRunner`)\n\nHandles individual step execution with comprehensive error handling:\n\n**Execution Pipeline:**\n```\nArgument Collection → Context Building → Guard Evaluation → Step Run → Result Processing\n```\n\n**Advanced Features:**\n- **Subpath Resolution**: Extracts nested values from step results using `result(:step, [:key, :subkey])`\n- **Process Metadata**: Maintains execution context in process dictionary for middleware\n- **Retry Management**: Tracks retry counts and respects step limits\n- **Compensation Handling**: Attempts error recovery when steps support it\n\n#### State Management (`Reactor.Executor.State`)\n\nRuntime-only execution state separate from reactor definition:\n\n```elixir\n%State{\n  async?: true,\n  concurrency_key: nil,\n  current_tasks: %{},     # task -> step mapping\n  errors: [],             # accumulated errors\n  max_concurrency: nil,   # concurrency limit\n  retries: %{},          # per-step retry counts\n  skipped: MapSet.new()   # steps skipped by guards\n}\n```\n\nThis separation allows the reactor definition to remain immutable while execution state is managed separately.\n\n#### Concurrency Tracker (`Reactor.Executor.ConcurrencyTracker`)\n\nETS-based global concurrency pool manager enabling resource sharing:\n\n**Architecture:**\n- **Pool Allocation**: Creates concurrency pools with configurable limits\n- **Atomic Operations**: Uses ETS `select_replace` for thread-safe resource tracking\n- **Process Monitoring**: Automatically cleans up pools when allocator processes die\n- **Deadlock Prevention**: Enables nested reactor execution without resource exhaustion\n\n**Pool Sharing Benefits:**\n- Multiple reactors can share concurrency limits\n- Prevents resource exhaustion in nested scenarios\n- Enables efficient resource utilisation across reactor hierarchy","ref":"architecture.html#executor-components"},{"type":"extras","title":"Planner (Reactor.Planner) - Reactor Architecture","doc":"Converts step definitions into an executable dependency graph:\n\n**Graph Construction Process:**\n1. **Vertex Creation**: Each step becomes a graph vertex\n2. **Dependency Analysis**: Examines step arguments to identify dependencies\n3. **Edge Creation**: Adds directed edges from dependency steps to dependent steps\n4. **Cycle Detection**: Validates the graph is acyclic\n5. **Ready State**: Identifies steps with no dependencies (ready to run)\n\n**Dependency Types:**\n- `input(:name)` - No dependencies (ready immediately)\n- `result(:step)` - Depends on named step\n- `value(x)` - No dependencies (static values)\n\nThe planner uses the [`libgraph`](https://hex.pm/packages/libgraph) library for efficient DAG operations and ensures all dependencies exist before execution begins.","ref":"architecture.html#planner-reactor-planner"},{"type":"extras","title":"Execution Flow - Reactor Architecture","doc":"","ref":"architecture.html#execution-flow"},{"type":"extras","title":"Initialization Phase - Reactor Architecture","doc":"**Setup Sequence:**\n```\nInputs + Context + Options → State Creation → Pool Allocation → Hook Initialization\n```\n\n1. **State Creation**: `Reactor.Executor.Init` builds execution state from options\n2. **Pool Management**: Allocates or reuses concurrency pools based on `concurrency_key`\n3. **Context Merging**: Combines user context with reactor context\n4. **Middleware Setup**: Initialises hooks for telemetry and other cross-cutting concerns","ref":"architecture.html#initialization-phase"},{"type":"extras","title":"Main Execution Loop - Reactor Architecture","doc":"The executor follows a sophisticated priority-based algorithm designed for maximum concurrency:\n\n```elixir\n# Simplified execution loop structure\ndef execute_loop(reactor, state) do\n  state\n  |> handle_completed_async_tasks()\n  |> start_new_async_tasks()\n  |> run_one_sync_step_if_available()\n  |> handle_deadlock_prevention()\n  |> check_completion_conditions()\n  |> continue_or_finish()\nend\n```\n\n**Execution Priorities:**\n1. **Async Task Completion**: Highest priority for maximum throughput\n2. **New Async Tasks**: Started when concurrency slots available\n3. **Sync Tasks**: Fill gaps when async slots unavailable\n4. **Deadlock Prevention**: Forces async tasks to run sync in nested scenarios","ref":"architecture.html#main-execution-loop"},{"type":"extras","title":"Step Execution Mechanics - Reactor Architecture","doc":"Each step follows a detailed execution pipeline:\n\n#### Argument Resolution\n\n```elixir\n# Argument types and their resolution\ninput(:name)                    # → reactor.context.private.inputs\nresult(:step)                   # → reactor.intermediate_results\nresult(:step, :key)             # → get_in(intermediate_results, [:step, :key])\nresult(:step, [:key, :subkey])  # → get_in(intermediate_results, [:step, :key, :subkey])\nvalue(literal)                  # → literal value\n```\n\n**Resolution Process:**\n- Validates all required arguments are available\n- Handles nested data extraction for subpaths\n- Applies transformations (extracted as separate steps during planning)\n- Builds argument map for step invocation\n\n#### Step Invocation\n\n```elixir\n# Step execution with comprehensive error handling\ncase step.run(arguments, context, step) do\n  {:ok, result} -> handle_success(result)\n  {:ok, result, new_steps} -> handle_success_with_new_steps(result, new_steps)\n  {:error, reason} -> attempt_compensation(reason)\n  :retry -> schedule_retry()\n  {:halt, reason} -> initiate_graceful_shutdown(reason)\nend\n```\n\n**Context Building:**\nEach step receives enhanced context:\n- **Reactor Context**: Shared data and configuration\n- **Step Context**: Current step, retry count, concurrency key\n- **Process Context**: Middleware-managed async execution context","ref":"architecture.html#step-execution-mechanics"},{"type":"extras","title":"Result Processing - Reactor Architecture","doc":"#### Success Path\n\n**Result Storage Strategy:**\n- Results stored in `intermediate_results` only when needed by:\n  - Dependent steps (determined by dependency graph)\n  - Reactor return value\n- Memory optimisation: unused results are discarded immediately\n\n**Undo Stack Management:**\n- Steps with `undo/4` callbacks are added to the undo stack with their results\n- Stack maintained in reverse chronological order (LIFO)\n- Non-undoable steps save memory by not storing results\n\n**Graph Updates:**\n- Completed steps are removed from the dependency graph\n- Dependent steps become ready when all dependencies satisfied\n- New steps (from `{:ok, result, new_steps}`) are planned and integrated\n\n#### Error Path\n\n**Error Handling Cascade:**\n1. **Compensation Attempt**: Call step's `compensate/4` if available\n2. **Undo Initiation**: If compensation fails, begin undo cascade\n3. **Stack Processing**: Undo steps in reverse order\n4. **Error Collection**: Accumulate all errors for final reporting\n5. **Resource Cleanup**: Release concurrency slots and clean up tasks\n\n**Compensation Options:**\n```elixir\ndef compensate(reason, arguments, context, step) do\n  case reason do\n    %RetryableError{} -> :retry              # Try again\n    %FallbackError{value: v} -> {:continue, v} # Use fallback\n    _other -> :ok                            # Accept failure\n  end\nend\n```","ref":"architecture.html#result-processing"},{"type":"extras","title":"State Management - Reactor Architecture","doc":"","ref":"architecture.html#state-management"},{"type":"extras","title":"Context Evolution - Reactor Architecture","doc":"**Multi-layered Context System:**\n```elixir\n# Base reactor context\n%{\n  private: %{\n    inputs: %{user_id: 123},           # User inputs\n    composed_reactors: #{},            # Composition tracking\n    replace_arguments: nil             # Argument transformation\n  }\n}\n\n# Enhanced step context (merged at runtime)\n%{\n  current_step: step,\n  current_try: 0,\n  retries_remaining: 5,\n  concurrency_key: #Reference<...>\n}\n```\n\n**Context Sharing:**\n- **Reactor Level**: Shared across all steps in the reactor\n- **Step Level**: Step-specific context merged with reactor context\n- **Process Level**: Middleware-managed context for async execution","ref":"architecture.html#context-evolution"},{"type":"extras","title":"Memory Management - Reactor Architecture","doc":"**Growth Patterns:**\n- **Constant**: Step definitions, dependency graph structure\n- **Linear**: Intermediate results, undo stack\n- **Temporary**: Task mappings, retry counters\n\n**Optimisation Strategies:**\n- Results stored only when required by dependencies or return value\n- Non-undoable steps excluded from undo stack\n- Completed steps removed from graph for garbage collection\n- Task mappings cleaned immediately upon completion","ref":"architecture.html#memory-management"},{"type":"extras","title":"Concurrency State - Reactor Architecture","doc":"**Resource Tracking:**\n```elixir\n# Active task management\ncurrent_tasks: %{\n  #PID<0.123.0> => %Step{name: :fetch_user},\n  #PID<0.124.0> => %Step{name: :process_data}\n}\n\n# Concurrency pool state (in ETS)\n{pool_key, current_count, max_count}\n```\n\n**Thread Safety:**\n- All state mutations happen in main executor process\n- Concurrency pool uses atomic ETS operations\n- Task results collected via message passing\n- No shared mutable state between async tasks","ref":"architecture.html#concurrency-state"},{"type":"extras","title":"Performance Characteristics - Reactor Architecture","doc":"","ref":"architecture.html#performance-characteristics"},{"type":"extras","title":"Concurrency Model - Reactor Architecture","doc":"**Bottleneck Analysis:**\n1. **Pool Exhaustion**: Limited by `max_concurrency` setting\n2. **Dependency Chains**: Sequential dependencies limit parallelisation\n3. **Resource Contention**: Shared pools across reactor hierarchy\n\n**Optimisation Features:**\n- **Automatic Sizing**: Default pool size based on `System.schedulers_online/0`\n- **Pool Sharing**: Multiple reactors share resources efficiently\n- **Eager Execution**: Tasks start immediately when dependencies satisfied\n- **Priority Scheduling**: Async completion processed first","ref":"architecture.html#concurrency-model"},{"type":"extras","title":"Memory Efficiency - Reactor Architecture","doc":"**Access Patterns:**\n- **Intermediate Results**: O(1) lookup by step name\n- **Undo Stack**: O(n) traversal during error handling only\n- **Task Management**: O(1) task-to-step mapping\n- **Dependency Resolution**: O(1) argument lookup\n\n**Scalability Characteristics:**\n- **Linear Growth**: Memory scales with active results and undo stack size\n- **Bounded Resources**: Concurrency limits prevent resource exhaustion\n- **Efficient Cleanup**: Completed steps removed immediately\n- **Network Effects**: Shared pools improve overall system utilisation","ref":"architecture.html#memory-efficiency"},{"type":"extras","title":"Graph Operations - Reactor Architecture","doc":"**Complexity Analysis:**\n- **Construction**: O(V + E) where V = steps, E = dependencies\n- **Planning**: O(V + E) cycle detection using efficient graph algorithms\n- **Execution**: O(1) ready step identification per iteration\n- **Updates**: O(1) step completion, O(k) for k new dependencies\n\n**Performance Optimisations:**\n- Compile-time planning for DSL-defined reactors\n- Incremental planning for dynamically added steps\n- Efficient graph library with optimised operations\n- Vertex removal enables immediate garbage collection","ref":"architecture.html#graph-operations"},{"type":"extras","title":"Integration Patterns - Reactor Architecture","doc":"","ref":"architecture.html#integration-patterns"},{"type":"extras","title":"Middleware Architecture - Reactor Architecture","doc":"**Hook Points:**\n- **Reactor Level**: `before_execute`, `after_execute`, `on_error`\n- **Step Level**: `before_step`, `after_step`, `on_step_error`\n- **Context Management**: Automatic context propagation to async tasks\n\n**Implementation:**\n```elixir\nreactor \"instrumented_workflow\" do\n  middlewares do\n    middleware Reactor.Middleware.Telemetry\n  end\nend\n```","ref":"architecture.html#middleware-architecture"},{"type":"extras","title":"Composition Model - Reactor Architecture","doc":"**Sub-Reactor Integration:**\n- Sub-reactors run with shared concurrency pools\n- Context propagated from parent to child\n- Error handling respects parent reactor's compensation strategy\n- Results flow back to parent reactor seamlessly","ref":"architecture.html#composition-model"},{"type":"extras","title":"Dynamic Workflow Capabilities - Reactor Architecture","doc":"**Runtime Step Creation:**\n```elixir\ndef run(arguments, context, step) do\n  new_steps = build_steps_based_on_data(arguments.data)\n  {:ok, result, new_steps}\nend\n```\n\n**Integration Process:**\n1. New steps planned and validated\n2. Dependencies integrated into existing graph\n3. Ready steps scheduled immediately\n4. Execution continues seamlessly\n\nThis architecture enables Reactor to efficiently orchestrate complex, concurrent workflows while maintaining transactional semantics and providing robust error handling capabilities. The design prioritises both performance and reliability, making it suitable for mission-critical applications requiring sophisticated workflow orchestration.","ref":"architecture.html#dynamic-workflow-capabilities"},{"type":"extras","title":"Design Decisions","doc":"# Design Decisions\n\nThis guide explains the reasoning behind Reactor's key design choices, helping you understand why the framework works the way it does and how these decisions benefit different use cases.","ref":"design-decisions.html"},{"type":"extras","title":"Why Dependency-Based Execution? - Design Decisions","doc":"","ref":"design-decisions.html#why-dependency-based-execution"},{"type":"extras","title":"The Decision - Design Decisions","doc":"Reactor uses argument dependencies to determine execution order rather than requiring explicit step sequencing. When a step declares `argument :user_id, result(:fetch_user)`, it automatically creates a dependency that ensures `fetch_user` runs before the current step.","ref":"design-decisions.html#the-decision"},{"type":"extras","title":"Reasoning Behind the Choice - Design Decisions","doc":"**Traditional Sequential Approach:**\n```elixir\n# Sequential workflow - explicit ordering\nstep_1 → step_2 → step_3 → step_4\n```\n\n**Reactor's Dependency Approach:**\n```elixir\n# Dependency-driven workflow - automatic ordering\ndefmodule UserWorkflow do\n  use Reactor\n  \n  step :fetch_user do\n    argument :user_id, input(:user_id)\n  end\n  \n  step :fetch_profile do\n    argument :user_id, input(:user_id)  # Can run concurrently with fetch_user\n  end\n  \n  step :merge_data do\n    argument :user, result(:fetch_user)      # Waits for fetch_user\n    argument :profile, result(:fetch_profile) # Waits for fetch_profile\n  end\nend\n```","ref":"design-decisions.html#reasoning-behind-the-choice"},{"type":"extras","title":"Benefits and Trade-offs - Design Decisions","doc":"**Benefits:**\n- **Automatic Concurrency**: Steps with no shared dependencies run simultaneously\n- **Natural Data Flow**: Dependencies express actual data relationships\n- **Dynamic Workflows**: Steps can emit new steps with automatically resolved dependencies\n- **Composition**: Sub-workflows integrate seamlessly without manual coordination\n\n**Trade-offs:**\n- **Learning Curve**: Developers must think in terms of data dependencies vs linear execution\n- **Debugging Complexity**: Execution order determined at runtime, may require visualisation tools\n- **Graph Overhead**: Dependency analysis and memory overhead for graph structures","ref":"design-decisions.html#benefits-and-trade-offs"},{"type":"extras","title":"Saga Pattern with Three-Tier Error Handling - Design Decisions","doc":"","ref":"design-decisions.html#saga-pattern-with-three-tier-error-handling"},{"type":"extras","title":"The Decision - Design Decisions","doc":"Reactor implements the saga pattern with a three-tier error handling approach: retry → compensation → undo, rather than traditional distributed transactions or simple error propagation.","ref":"design-decisions.html#the-decision-1"},{"type":"extras","title":"Why Saga Pattern? - Design Decisions","doc":"**Distributed Transaction Limitations:**\n- Require all participants to support two-phase commit\n- Poor availability during failures\n- Don't work across different system types (databases, APIs, file systems)\n\n**Saga Pattern Benefits:**\n- Works with any system type\n- Graceful degradation during failures\n- Business logic integrated into error handling\n- Natural fit for microservice architectures","ref":"design-decisions.html#why-saga-pattern"},{"type":"extras","title":"Three-Tier Error Handling - Design Decisions","doc":"**Level 1: Retry Logic**\n```elixir\ndef compensate(%NetworkTimeoutError{}, _arguments, context, _step) do\n  # Exponential backoff - Reactor handles max_retries automatically\n  delay = :math.pow(2, context.current_try) * 1000\n  Process.sleep(delay)\n  :retry\nend\n```\n\n**Level 2: Compensation (Smart Recovery)**\n```elixir\ndef compensate(%PaymentGatewayError{backup_gateway: backup}, arguments, _context, _step) do\n  # Try backup payment gateway\n  case BackupGateway.charge(arguments.amount, arguments.card) do\n    {:ok, charge} -> {:continue, charge}\n    {:error, _} -> :ok  # Proceed to undo\n  end\nend\n```\n\n**Level 3: Undo (Rollback)**\n```elixir\ndef undo(created_user, _arguments, _context, _step) do\n  # Remove the user that was successfully created\n  UserDatabase.delete(created_user.id)\n  :ok\nend\n```","ref":"design-decisions.html#three-tier-error-handling"},{"type":"extras","title":"Design Philosophy - Design Decisions","doc":"**Step-Level Autonomy with Global Coordination:**\n- Each step decides how to handle its own failures (local knowledge)\n- Global system coordinates rollback when needed (system consistency)\n- Business logic stays close to the step that understands it\n\n**Multiple Recovery Paths:**\n- Temporary issues handled by retry\n- Business logic errors handled by compensation\n- System consistency maintained by undo\n- Clear separation between technical failures and business logic failures","ref":"design-decisions.html#design-philosophy"},{"type":"extras","title":"DSL vs Programmatic APIs - Design Decisions","doc":"","ref":"design-decisions.html#dsl-vs-programmatic-apis"},{"type":"extras","title":"The Decision - Design Decisions","doc":"Reactor provides both a declarative DSL (using Spark) and a programmatic Builder API, rather than choosing just one approach.","ref":"design-decisions.html#the-decision-2"},{"type":"extras","title":"Rationale for Dual Approach - Design Decisions","doc":"**DSL Strengths - Static Workflows:**\n```elixir\ndefmodule OrderProcessing do\n  use Reactor\n  \n  input :order_id\n  input :payment_info\n  \n  step :validate_order do\n    argument :order_id, input(:order_id)\n    run &OrderValidator.validate/2\n  end\n  \n  step :process_payment do\n    argument :order, result(:validate_order)\n    argument :payment_info, input(:payment_info)\n    run &PaymentProcessor.charge/2\n    compensate &PaymentProcessor.refund/4\n  end\nend\n```\n\n**Builder API Strengths - Dynamic Workflows:**\n```elixir\ndef build_dynamic_workflow(config) do\n  reactor = Reactor.Builder.new()\n  \n  # Add steps based on runtime configuration\n  Enum.reduce(config.processing_steps, reactor, fn step_config, reactor ->\n    {:ok, reactor} = Reactor.Builder.add_step(reactor, step_config.name, step_config.impl)\n    reactor\n  end)\nend\n```","ref":"design-decisions.html#rationale-for-dual-approach"},{"type":"extras","title":"When to Use Each - Design Decisions","doc":"**Use DSL When:**\n- Workflow structure is known at compile time\n- You want declarative, self-documenting code\n- Static analysis and validation is important\n\n**Use Builder API When:**\n- Workflow structure depends on runtime data\n- Building workflows programmatically\n- Integrating with code generation tools\n\n**Both approaches produce identical runtime structures and can be composed together.**","ref":"design-decisions.html#when-to-use-each"},{"type":"extras","title":"Concurrency by Default - Design Decisions","doc":"","ref":"design-decisions.html#concurrency-by-default"},{"type":"extras","title":"The Decision - Design Decisions","doc":"Steps run asynchronously by default (`async?: true`) with sophisticated concurrency management and deadlock prevention.","ref":"design-decisions.html#the-decision-3"},{"type":"extras","title":"Why Async by Default? - Design Decisions","doc":"**Modern Computing Reality:**\n- Multi-core processors are standard\n- I/O operations dominate many workflows\n- Network latency is orders of magnitude higher than CPU operations\n\n**Performance Benefits:**\n```elixir\n# Without concurrency: 300ms total\nfetch_user(100ms) → fetch_preferences(100ms) → fetch_history(100ms)\n\n# With concurrency: ~100ms total  \nfetch_user(100ms) ┐\nfetch_preferences(100ms) ├→ merge_results\nfetch_history(100ms) ┘\n```","ref":"design-decisions.html#why-async-by-default"},{"type":"extras","title":"Sophisticated Concurrency Management - Design Decisions","doc":"**Deadlock Prevention:**\n```elixir\n# Automatic handling of nested reactor scenarios\ndefmodule ParentReactor do\n  use Reactor\n\n  compose :process_batch, ChildReactor do\n    # When concurrency pool exhausted, composed reactor runs synchronously\n    # Automatically detects and prevents deadlock\n  end\nend\n\n# Run with shared concurrency pool\nReactor.run(ParentReactor, %{}, max_concurrency: 8)\n```\n\n**Resource Management:**\n- Shared concurrency pools across reactor hierarchies\n- Automatic resource acquisition and release\n- Process monitoring and cleanup","ref":"design-decisions.html#sophisticated-concurrency-management"},{"type":"extras","title":"Design Philosophy - Design Decisions","doc":"**Performance by Default:**\n- Optimal performance without developer intervention\n- Scales automatically with system capabilities\n- Simple mental model: \"declare dependencies, get concurrency\"\n\n**Safety and Reliability:**\n- Automatic deadlock detection and prevention\n- Resource limits prevent system exhaustion\n- Explicit opt-out for special cases (`async?: false`)","ref":"design-decisions.html#design-philosophy-1"},{"type":"extras","title":"Integration with Spark/Ash Ecosystem - Design Decisions","doc":"","ref":"design-decisions.html#integration-with-spark-ash-ecosystem"},{"type":"extras","title":"The Decision - Design Decisions","doc":"Reactor is built on the Spark library and designed to integrate seamlessly with the Ash ecosystem, while still being a standalone workflow engine that works with any Elixir application.","ref":"design-decisions.html#the-decision-4"},{"type":"extras","title":"Why Build on Spark? - Design Decisions","doc":"**Shared DSL Infrastructure:**\n```elixir\n# Consistent DSL patterns across Ash ecosystem\ndefmodule MyResource do\n  use Ash.Resource\n  \n  actions do\n    read :list_users\n    create :create_user\n  end\nend\n\ndefmodule MyReactor do\n  use Reactor  # Same DSL foundation\n  \n  step :list_users\n  step :create_user\nend\n```","ref":"design-decisions.html#why-build-on-spark"},{"type":"extras","title":"Ecosystem Integration Benefits - Design Decisions","doc":"**Complementary Capabilities:**\n- **Ash**: Data layer with resources, actions, policies\n- **Reactor**: Workflow orchestration and business process management\n- **Phoenix**: Web interface and real-time communication\n\n**Shared Patterns:**\n- Compatible concurrency models\n- Similar error handling approaches\n- Consistent configuration and tooling\n\n**Natural Composition:**\n```elixir\ndefmodule OrderWorkflow do\n  use Reactor\n  \n  step :create_order do\n    run fn args, _context ->\n      # Integrate with Ash actions\n      args.order_params\n      |> Order.create()\n      |> case do\n        {:ok, order} -> {:ok, order}\n        {:error, changeset} -> {:error, changeset}\n      end\n    end\n  end\nend\n```","ref":"design-decisions.html#ecosystem-integration-benefits"},{"type":"extras","title":"Architectural Benefits - Design Decisions","doc":"**Learning Transfer:**\n- Developers familiar with Ash can quickly learn Reactor\n- Shared concepts, patterns, and tooling\n- Reduced learning curve\n\n**Future Evolution:**\n- Can evolve together with Ash ecosystem\n- Shared infrastructure improvements\n- Coordinated feature development","ref":"design-decisions.html#architectural-benefits"},{"type":"extras","title":"Design Constraints and Benefits - Design Decisions","doc":"**Constraints Accepted:**\n- Dependency on Spark\n- Need to maintain compatibility with Ash patterns\n\n**Benefits Gained:**\n- Massive reduction in DSL implementation complexity\n- Automatic compatibility with existing Ash tooling\n- Strong ecosystem alignment and developer experience","ref":"design-decisions.html#design-constraints-and-benefits"},{"type":"extras","title":"Summary - Design Decisions","doc":"These design decisions work together to create a framework that:\n\n1. **Maximises Performance**: Through dependency-based concurrency and async-by-default execution\n2. **Enables Complex Workflows**: Through saga patterns and sophisticated error handling\n3. **Provides Developer Choice**: Through dual DSL/programmatic APIs\n4. **Integrates Naturally**: Through Spark/Ash ecosystem alignment\n5. **Scales Gracefully**: Through composable patterns and resource management\n\nUnderstanding these decisions helps you make better choices when designing workflows and contributes to effective use of Reactor in complex applications.","ref":"design-decisions.html#summary"},{"type":"extras","title":"Reactor in the Elixir Ecosystem","doc":"# Reactor in the Elixir Ecosystem\n\nReactor is a framework-independent, dynamic, concurrent, dependency-resolving saga orchestrator for Elixir. While it's part of the ash-project organization and has excellent Ash framework integration, Reactor is designed to work with any Elixir application and can orchestrate workflows across diverse systems and frameworks.","ref":"ecosystem.html"},{"type":"extras","title":"Framework Independence - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#framework-independence"},{"type":"extras","title":"Core Design Philosophy - Reactor in the Elixir Ecosystem","doc":"Reactor's architecture is deliberately framework-agnostic:\n\n- **No Framework Dependencies**: Reactor's core has no hard dependencies on web frameworks, data layers, or specific libraries\n- **Generic Step Interface**: The `Reactor.Step` behaviour works with any Elixir code\n- **Composable Architecture**: Steps can integrate with any Elixir libraries or external systems\n- **Universal Patterns**: Dependency resolution, error handling, and compensation work regardless of the underlying technology stack\n\n```elixir\n# Reactor works with any Elixir code\ndefmodule GenericWorkflowReactor do\n  use Reactor\n\n  input :data\n\n  step :process_with_genserver do\n    argument :data, input(:data)\n    run fn args, _context ->\n      GenServer.call(MyService, {:process, args.data})\n    end\n  end\n\n  step :call_external_api do\n    argument :processed_data, result(:process_with_genserver)\n    run fn args, _context ->\n      HTTPoison.post(\"https://api.example.com\", args.processed_data)\n    end\n  end\n\n  step :store_in_ets do\n    argument :api_response, result(:call_external_api)\n    run fn args, _context ->\n      :ets.insert(:my_table, {:result, args.api_response})\n      {:ok, :stored}\n    end\n  end\n\n  return :store_in_ets\nend\n```","ref":"ecosystem.html#core-design-philosophy"},{"type":"extras","title":"Reactor Ecosystem Packages - Reactor in the Elixir Ecosystem","doc":"The Reactor ecosystem includes several specialized packages that extend its capabilities:","ref":"ecosystem.html#reactor-ecosystem-packages"},{"type":"extras","title":"Core Ecosystem Packages - Reactor in the Elixir Ecosystem","doc":"**[reactor_file](https://hex.pm/packages/reactor_file)**\n- File system operations within reactor workflows\n- File copying, moving, and manipulation steps\n- Directory creation/removal, permissions, and file I/O operations\n\n**[reactor_process](https://hex.pm/packages/reactor_process)**\n- Supervisor and process management operations\n- Start, restart, terminate, and delete child processes\n- Process supervision and lifecycle management within workflows\n\n**[reactor_req](https://hex.pm/packages/reactor_req)**\n- HTTP client steps for making web requests\n- DSL support for GET, POST, PUT, DELETE, and other HTTP methods\n- Integration with the Req library for structured HTTP operations","ref":"ecosystem.html#core-ecosystem-packages"},{"type":"extras","title":"Extension Pattern - Reactor in the Elixir Ecosystem","doc":"Ecosystem packages provide additional DSL entities that become available when you include them as extensions:\n\n```elixir\n# Example ecosystem integration\ndefmodule MyWorkflowReactor do\n  use Reactor, extensions: [Reactor.Req, Reactor.File, Reactor.Process]\n  \n  input :file_path\n  input :upload_url\n  \n  # File operations from reactor_file (DSL entities)\n  file_read :process_file do\n    path input(:file_path)\n  end\n  \n  # HTTP requests from reactor_req (DSL entities)\n  req_post :upload_data do\n    url input(:upload_url)\n    body result(:process_file)\n  end\n  \n  # Process management from reactor_process (DSL entities)\n  start_child :start_worker do\n    supervisor value(MyApp.WorkerSupervisor) \n    child_spec template({MyApp.Worker, result(:process_file)})\n  end\n  \n  return :upload_data\nend\n```","ref":"ecosystem.html#extension-pattern"},{"type":"extras","title":"Ash Framework Integration - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#ash-framework-integration"},{"type":"extras","title":"Ash.Reactor Extension - Reactor in the Elixir Ecosystem","doc":"While Reactor is framework-independent, the **[Ash.Reactor](https://hexdocs.pm/ash/Ash.Reactor.html)** extension provides deep integration with Ash resources and actions:\n\n```elixir\ndefmodule UserOnboardingReactor do\n  # Using the Ash.Reactor extension\n  use Ash.Reactor\n\n  input :user_params\n\n  # Ash-specific create action step\n  create :user, MyApp.User, :create do\n    inputs %{\n      email: input(:user_params, [:email]),\n      name: input(:user_params, [:name])\n    }\n  end\n\n  # Ash-specific update action step  \n  update :activate_user, MyApp.User, :activate do\n    record result(:user)\n    inputs %{activated_at: value(DateTime.utc_now())}\n  end\n\n  # Ash action with automatic undo support\n  action :send_welcome_email, MyApp.Notifications, :send_welcome do\n    inputs %{user_id: result(:user, :id)}\n    undo :outside_transaction\n  end\n\n  return :activate_user\nend\n```","ref":"ecosystem.html#ash-reactor-extension"},{"type":"extras","title":"Ash-Specific Capabilities - Reactor in the Elixir Ecosystem","doc":"**Resource Action Integration:**\n- Direct integration with Ash resource actions (create, read, update, destroy, and generic actions)\n- Reactor modules can serve as the `run` implementation for Ash generic actions\n- Bidirectional integration: Reactor can orchestrate Ash actions, and Ash actions can use Reactor for implementation\n- Automatic changeset handling and validation\n- Built-in support for Ash's authorization policies\n\n**Transaction Management:**\n- Automatic database transaction coordination\n- Rollback capabilities integrated with Ash's transaction handling\n- Undo strategies that respect Ash's data layer behaviour\n\n**Notification Handling:**\n- Proper integration with Ash's notification system","ref":"ecosystem.html#ash-specific-capabilities"},{"type":"extras","title":"When to Use Ash.Reactor vs Core Reactor - Reactor in the Elixir Ecosystem","doc":"**Use Ash.Reactor when:**\n- Working primarily with Ash resources and actions\n- Need automatic changeset validation and error handling\n- Require database transaction coordination\n- Want built-in authorization policy integration\n\n**Use Core Reactor when:**\n- Working with non-Ash systems and libraries\n- Building framework-agnostic workflows\n- Integrating with existing Elixir applications\n- Need maximum flexibility and minimal dependencies","ref":"ecosystem.html#when-to-use-ash-reactor-vs-core-reactor"},{"type":"extras","title":"Relationship to Other Elixir Tools - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#relationship-to-other-elixir-tools"},{"type":"extras","title":"vs. GenServer and OTP Patterns - Reactor in the Elixir Ecosystem","doc":"**Reactor Strengths:**\n- Declarative workflow definition with automatic dependency resolution\n- Built-in compensation patterns for error recovery\n- Multi-system coordination and state management\n\n**OTP Strengths:**\n- Long-running stateful processes\n- Real-time systems requiring immediate response\n- Performance-critical operations with direct supervision tree integration","ref":"ecosystem.html#vs-genserver-and-otp-patterns"},{"type":"extras","title":"vs. Broadway for Data Processing - Reactor in the Elixir Ecosystem","doc":"**Reactor Use Cases:**\n- Complex workflows with interdependent steps and compensation needs\n- Multi-system coordination requiring state management\n- Business processes needing human intervention or approval\n- Processing individual messages or batches within Broadway pipelines\n\n**Broadway Use Cases:**\n- High-throughput stream processing with simple transforms\n- Message queue consumption with batching and linear processing","ref":"ecosystem.html#vs-broadway-for-data-processing"},{"type":"extras","title":"vs. Task and Async Patterns - Reactor in the Elixir Ecosystem","doc":"**Reactor Advantages:**\n- Automatic dependency resolution and sophisticated error handling\n- Declarative composition with state management across operations\n- Shared concurrency pools that prevent resource exhaustion\n\n**Task/Async Advantages:**\n- Minimal overhead with direct process control\n- Simpler model for basic parallelisation and performance-critical paths","ref":"ecosystem.html#vs-task-and-async-patterns"},{"type":"extras","title":"Integration Examples - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#integration-examples"},{"type":"extras","title":"Multi-Framework Workflow - Reactor in the Elixir Ecosystem","doc":"```elixir\ndefmodule E2EProcessingReactor do\n  use Reactor\n\n  input :upload_params\n\n  # Phoenix file upload handling\n  step :handle_upload do\n    argument :params, input(:upload_params)\n    run &MyAppWeb.UploadController.process_upload/2\n  end\n\n  # Ash resource operations (if using Ash.Reactor)\n  step :store_metadata do\n    argument :file_info, result(:handle_upload)\n    run fn args, _context ->\n      MyApp.FileMetadata.create(args.file_info)\n    end\n  end\n\n  # External service integration\n  step :process_with_ai do\n    argument :file_path, result(:handle_upload, [:path])\n    run fn args, _context ->\n      HTTPoison.post(\"https://ai-service.com/process\", %{\n        file_url: args.file_path\n      })\n    end\n  end\n\n  # Background job scheduling\n  step :schedule_cleanup do\n    argument :file_path, result(:handle_upload, [:path])\n    run fn args, _context ->\n      Oban.insert(CleanupWorker.new(%{file_path: args.file_path}))\n    end\n  end\n\n  return :store_metadata\nend\n```","ref":"ecosystem.html#multi-framework-workflow"},{"type":"extras","title":"Bidirectional Ash Integration - Reactor in the Elixir Ecosystem","doc":"```elixir\n# Reactor as Ash action implementation\ndefmodule MyApp.UserRegistrationReactor do\n  use Ash.Reactor\n\n  input :email\n  input :password\n  input :name\n\n  create :user, MyApp.User, :create do\n    inputs %{\n      email: input(:email),\n      name: input(:name)\n    }\n  end\n\n  action :send_welcome_email, MyApp.Notifications, :send_welcome do\n    inputs %{user_id: result(:user, :id)}\n  end\n\n  return :user\nend\n\n# Ash resource using Reactor as action implementation\ndefmodule MyApp.User do\n  use Ash.Resource\n\n  actions do\n    create :create do\n      argument :email, :string, allow_nil?: false\n      argument :password, :string, allow_nil?: false\n      argument :name, :string, allow_nil?: false\n    end\n\n    # Reactor module as generic action implementation\n    action :complete_registration, :struct do\n      argument :email, :string, allow_nil?: false\n      argument :password, :string, allow_nil?: false  \n      argument :name, :string, allow_nil?: false\n      \n      run MyApp.UserRegistrationReactor\n    end\n  end\nend\n```","ref":"ecosystem.html#bidirectional-ash-integration"},{"type":"extras","title":"Ecosystem Package Integration - Reactor in the Elixir Ecosystem","doc":"```elixir\ndefmodule DocumentProcessingReactor do\n  use Reactor, extensions: [Reactor.Req, Reactor.File, Reactor.Process]\n\n  input :document_url\n  input :processing_config\n\n  # HTTP download with reactor_req\n  req_get :download_document do\n    url input(:document_url)\n  end\n\n  # File processing with reactor_file\n  file_write :save_to_temp do\n    path template(\"/tmp/{{uuid}}.pdf\")\n    content result(:download_document, [:body])\n  end\n\n  # External processing with reactor_process\n  start_child :convert_to_text do\n    supervisor value(MyApp.ProcessSupervisor)\n    child_spec template({PDFConverter, [result(:save_to_temp, [:path])]})\n  end\n\n  # Cleanup with compensation\n  file_delete :cleanup_temp_file do\n    path result(:save_to_temp, [:path])\n    wait_for :convert_to_text\n  end\n\n  return :convert_to_text\nend\n```","ref":"ecosystem.html#ecosystem-package-integration"},{"type":"extras","title":"When to Choose Reactor - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#when-to-choose-reactor"},{"type":"extras","title":"Decision Framework - Reactor in the Elixir Ecosystem","doc":"**Choose Reactor when your workflow has:**\n\n1. **Complex Dependencies**: Multi-step processes where operations depend on results from previous steps\n2. **Error Recovery Needs**: Sophisticated compensation, rollback, or cleanup requirements\n3. **Cross-System Coordination**: Integration spanning multiple services, APIs, or external systems\n4. **State Management Requirements**: Need to track intermediate results and context across operations\n5. **Declarative Benefits**: Complex logic that benefits from clear, maintainable specifications\n\n**Specific use cases include:**\n- Financial transactions requiring compensation (payment processing, refunds)\n- Multi-stage business processes (user onboarding, order fulfilment)\n- Data pipelines with validation and transformation steps\n- API orchestration across microservices\n- Resource provisioning with cleanup on failure\n\n**Choose alternatives when:**\n- **High Performance**: Maximum throughput with minimal overhead is critical\n- **Simple Operations**: Straightforward CRUD or linear processing without dependencies\n- **Real-Time Constraints**: Immediate response requirements (use OTP patterns)\n- **Stream Processing**: High-throughput data transformation (use Broadway)","ref":"ecosystem.html#decision-framework"},{"type":"extras","title":"Community and Ecosystem - Reactor in the Elixir Ecosystem","doc":"","ref":"ecosystem.html#community-and-ecosystem"},{"type":"extras","title":"Learning Resources - Reactor in the Elixir Ecosystem","doc":"**Framework-Independent Learning:**\n- Core Reactor patterns apply across any Elixir application\n- Ecosystem packages provide specialized capabilities\n- Community patterns and best practices work universally\n\n**Ash-Specific Learning:**\n- [Ash Reactor guide](https://hexdocs.pm/ash/reactor.html) for framework integration\n- [Ash community Discord](https://discord.gg/3hA2j4Jt) for specific integration questions\n- Combined Ash + Reactor tutorials and examples","ref":"ecosystem.html#learning-resources"},{"type":"extras","title":"Future Evolution - Reactor in the Elixir Ecosystem","doc":"Reactor's framework-independent design makes it adaptable to new technologies and integration requirements. The ecosystem could potentially grow in many directions:\n\n**Possible Integration Areas:**\n- Database-specific integrations beyond Ash\n- Cloud service orchestration packages\n- Monitoring and observability extensions\n- Testing and development tools\n\n**Framework Integration Opportunities:**\n- Phoenix LiveView workflow coordination\n- Nerves IoT device orchestration  \n- Machine learning pipeline management\n- Blockchain and distributed system coordination\n\n**Contributing to the Ecosystem:**\nWhat integrations would be valuable for your use cases? The Reactor community welcomes contributions that extend its capabilities while maintaining its core strengths in workflow orchestration, dependency resolution, and error handling. Whether used standalone or with framework-specific extensions like Ash.Reactor, it provides a solid foundation for building reliable, maintainable business workflows.","ref":"ecosystem.html#future-evolution"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}